#include <pspsdk.h>
#include <pspkernel.h>
#include <pspsysmem_kernel.h>
#include <pspsysevent.h>
#include <psploadcore.h>
#include <pspiofilemgr_kernel.h>
#include <pspsyscon.h>
#include <pspnand_driver.h>
#include <pspidstorage.h>
#include <pspdisplay_kernel.h>
#include <systemctrl.h>
#include <pspthreadman_kernel.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "idsregeneration.h"
#include "aes.h"

#include "hwconfig.h"
#include "magicgate.h"
#include "serial.h"
#include "usb.h"
#include "unk140.h"
#include "version.h"
#include "lcd.h"
#include "unk52.h"
#include "region.h"
#include "mi.h"


PSP_MODULE_INFO("pspIdsRegeneration_Driver", 0x1007, 1, 0);

enum
{
	TA79v1,
	TA79v2,
	TA79v3,
	TA81,
	TA82,
	TA85,
	TA85v2,
	TA86,
	TA88,
	TAUN,
	DEVKIT
};

enum 
{
	JAPAN = 3,
	AMERICA = 4,
	EUROPE = 5
};

typedef struct U64
{
	u32 low;
	u32 high;
} U64;

int g_tachyon, g_baryon, g_pommel, g_mb, g_region;
u64 g_fuseid;
u8 setParam[8];

int psptype, isslim, is79;

u32 table[256];

static void gen_table(void)
{
    unsigned int i, j, x;
 
    for (i=0; i<256; i++)
    {
        x = i;
        for (j=0; j<8; j++)
        {
            if (x & 1)
                x = (x>>1) ^ 0xEDB88320;
            else
                x = (x>>1);
        }

        table[i] = x;
    }
}

static u32 gen_hash(unsigned int seed, unsigned char buf[], int size)
{
    unsigned int hash = ~seed;
    unsigned char index;
    int i;

    for (i=0; i!=size; i++)
    {
        index = buf[i] ^ hash;
        hash>>= 8;
        hash ^= table[index];
    }

    return(~hash);
}



int idsRegenerationSetup(int tachyon, int baryon, int pommel, int mb, u64 fuseid, int region, u8 *setparam)
{
	int k1 = pspSdkSetK1(0);

	gen_table();

	if (region < JAPAN || region > EUROPE)
	{
		pspSdkSetK1(k1);
		return -1;
	}

	g_tachyon = tachyon;
	g_baryon = baryon;
	g_pommel = pommel;
	g_mb = mb;
	g_fuseid = fuseid;
	g_region = region;

	if (mb < TA82)
	{
		psptype = 1;
	}
	else if (mb == TA82 || mb == TA86)
	{
		psptype = 2;
	}
	else
	{
		psptype = 3;
	}

	isslim = (psptype == 3);
	is79 = (mb <= TA79v3);

	if (setparam)
		memcpy(setParam, setparam, 8);
	else
		memset(setParam, 0, 8);

	pspSdkSetK1(k1);
	return 0;
}

int idsRegenerationGetIndex(IdsIndex *index, int *n)
{
	int k1 = pspSdkSetK1(0);

	if (psptype == 1)
	{
		if (g_region == JAPAN)
		{
			index[0].keyfirst = 0x120;
			index[0].keyend = 0x13F;
			
			index[1].keyfirst = 0x10;
			index[1].keyend = 0x2F;

			index[2].keyfirst = 0x0F;
			index[2].keyend = 0x0F;

			index[3].keyfirst = 0x50;
			index[3].keyend = 0x50;

			index[4].keyfirst = 0x45;
			index[4].keyend = 0x46;

			index[5].keyfirst = 0x04;
			index[5].keyend = 0x06;

			index[6].keyfirst = 0x41;
			index[6].keyend = 0x43;

			index[7].keyfirst = 0x140;
			index[7].keyend = 0x140;

			index[8].keyfirst = 0x44;
			index[8].keyend = 0x44;

			index[9].keyfirst = 0x40;
			index[9].keyend = 0x40;

			index[10].keyfirst = 0x30;
			index[10].keyend = 0x3F;

			index[11].keyfirst = 0x100;
			index[11].keyend = 0x11F;

			*n = 12;
		}
		else if (g_region == AMERICA)
		{
			// Got from american
			index[0].keyfirst = 0x120;
			index[0].keyend = 0x13F;

			index[1].keyfirst = 0x10;
			index[1].keyend = 0x2F;

			index[2].keyfirst = 0x140;
			index[2].keyend = 0x140;

			index[3].keyfirst = 0x100;
			index[3].keyend = 0x11F;

			index[4].keyfirst = 0x0F;
			index[4].keyend = 0x0F;

			index[5].keyfirst = 0x50;
			index[5].keyend = 0x50;

			index[6].keyfirst = 0x45;
			index[6].keyend = 0x46;

			index[7].keyfirst = 0x04;
			index[7].keyend = 0x06;

			index[8].keyfirst = 0x41;
			index[8].keyend = 0x44;

			index[9].keyfirst = 0x40;
			index[9].keyend = 0x40;

			index[10].keyfirst = 0x30;
			index[10].keyend = 0x3F;

			*n = 11;
		}
	
		else // region 5 euro
		{
			index[0].keyfirst = 0x120;
			index[0].keyend = 0x13F;

			index[1].keyfirst = 0x10;
			index[1].keyend = 0x2F;

			index[2].keyfirst = 0x0F;
			index[2].keyend = 0x0F;

			index[3].keyfirst = 0x50;
			index[3].keyend = 0x50;

			index[4].keyfirst = 0x45;
			index[4].keyend = 0x46;

			index[5].keyfirst = 0x04;
			index[5].keyend = 0x06;

			index[6].keyfirst = 0x41;
			index[6].keyend = 0x44;

			index[7].keyfirst = 0x40;
			index[7].keyend = 0x40;

			index[8].keyfirst = 0x100;
			index[8].keyend = 0x11F;

			index[9].keyfirst = 0x140;
			index[9].keyend = 0x140;

			index[10].keyfirst = 0x30;
			index[10].keyend = 0x3F;

			*n = 11;
		}
	}

	else if (psptype == 2)
	{
		if (g_mb == TA82)
		{
			if (1)
			{
				// got from euro
				index[0].keyfirst = 0x120;
				index[0].keyend = 0x13F;
			
				index[1].keyfirst = 0x10;
				index[1].keyend = 0x2F;

				index[2].keyfirst = 0x100;
				index[2].keyend = 0x11F;

				index[3].keyfirst = 0x0F;
				index[3].keyend = 0x0F;

				index[4].keyfirst = 0x50;
				index[4].keyend = 0x50;

				index[5].keyfirst = 0x47;
				index[5].keyend = 0x47;

				index[6].keyfirst = 0x45;
				index[6].keyend = 0x46;

				index[7].keyfirst = 0x04;
				index[7].keyend = 0x06;

				index[8].keyfirst = 0x41;
				index[8].keyend = 0x43;

				index[9].keyfirst = 0x51;
				index[9].keyend = 0x51;

				index[10].keyfirst = 0x44;
				index[10].keyend = 0x44;

				index[11].keyfirst = 0x40;
				index[11].keyend = 0x40;

				index[12].keyfirst = 0x140;
				index[12].keyend = 0x140;

				index[13].keyfirst = 0x30;
				index[13].keyend = 0x3F;

				*n = 14;
			}
		}
		else if (g_mb == TA86)
		{
			if (g_region == EUROPE)
			{
				// got from euro 86
				index[0].keyfirst = 0x120;
				index[0].keyend = 0x13F;

				index[1].keyfirst = 0x10;
				index[1].keyend = 0x2F;

				index[2].keyfirst = 0x140;
				index[2].keyend = 0x140;

				index[3].keyfirst = 0x100;
				index[3].keyend = 0x11F;

				index[4].keyfirst = 0x0F;
				index[4].keyend = 0x0F;

				index[5].keyfirst = 0x50;
				index[5].keyend = 0x50;

				index[6].keyfirst = 0x47;
				index[6].keyend = 0x47;

				index[7].keyfirst = 0x45;
				index[7].keyend = 0x46;

				index[8].keyfirst = 0x04;
				index[8].keyend = 0x06;

				index[9].keyfirst = 0x08;
				index[9].keyend = 0x08;

				index[10].keyfirst = 0x41;
				index[10].keyend = 0x43;

				index[11].keyfirst = 0x51;
				index[11].keyend = 0x51;

				index[12].keyfirst = 0x44;
				index[12].keyend = 0x44;

				index[13].keyfirst = 0x07;
				index[13].keyend = 0x07;

				index[14].keyfirst = 0x40;
				index[14].keyend = 0x40;

				index[15].keyfirst = 0x30;
				index[15].keyend = 0x3F;

				*n = 16;
			}
			else
			{
				// got from jap 86
				index[0].keyfirst = 0x120;
				index[0].keyend = 0x13F;

				index[1].keyfirst = 0x10;
				index[1].keyend = 0x2F;

				index[2].keyfirst = 0x100;
				index[2].keyend = 0x11F;

				index[3].keyfirst = 0x0F;
				index[3].keyend = 0x0F;

				index[4].keyfirst = 0x50;
				index[4].keyend = 0x50;

				index[5].keyfirst = 0x47;
				index[5].keyend = 0x47;

				index[6].keyfirst = 0x45;
				index[6].keyend = 0x46;

				index[7].keyfirst = 0x04;
				index[7].keyend = 0x06;

				index[8].keyfirst = 0x08;
				index[8].keyend = 0x08;

				index[9].keyfirst = 0x41;
				index[9].keyend = 0x43;

				index[10].keyfirst = 0x51;
				index[10].keyend = 0x51;

				index[11].keyfirst = 0x44;
				index[11].keyend = 0x44;

				index[12].keyfirst = 0x07;
				index[12].keyend = 0x07;

				index[13].keyfirst = 0x40;
				index[13].keyend = 0x40;

				index[14].keyfirst = 0x140;
				index[14].keyend = 0x140;

				index[15].keyfirst = 0x30;
				index[15].keyend = 0x3F;

				*n = 16;
			}
		}
		else
		{
			pspSdkSetK1(k1);
			return -1;
		}
	}

	else if (psptype == 3)
	{
		if (g_mb == TA85 || g_mb == TA85v2)
		{
			if (g_region == JAPAN)
			{
				index[0].keyfirst = 0x120;
				index[0].keyend = 0x13F;

				index[1].keyfirst = 0x0F;
				index[1].keyend = 0x0F;

				index[2].keyfirst = 0x50;
				index[2].keyend = 0x50;

				index[3].keyfirst = 0x52;
				index[3].keyend = 0x53;

				index[4].keyfirst = 0x47;
				index[4].keyend = 0x47;

				index[5].keyfirst = 0x45;
				index[5].keyend = 0x46;

				index[6].keyfirst = 0x04;
				index[6].keyend = 0x06;

				index[7].keyfirst = 0x08;
				index[7].keyend = 0x08;

				index[8].keyfirst = 0x41;
				index[8].keyend = 0x43;

				index[9].keyfirst = 0x51;
				index[9].keyend = 0x51;

				index[10].keyfirst = 0x54;
				index[10].keyend = 0x54;

				index[11].keyfirst = 0x140;
				index[11].keyend = 0x140;

				index[12].keyfirst = 0x44;
				index[12].keyend = 0x44;

				index[13].keyfirst = 0x07;
				index[13].keyend = 0x07;

				index[14].keyfirst = 0x40;
				index[14].keyend = 0x40;

				index[15].keyfirst = 0x100;
				index[15].keyend = 0x11F;

				index[16].keyfirst = 0x30;
				index[16].keyend = 0x3F;

				index[17].keyfirst = 0x10;
				index[17].keyend = 0x2F;

				*n = 18;
			}
			else // america & euro
			{
				// got from american
				index[0].keyfirst = 0x0F;
				index[0].keyend = 0x0F;

				index[1].keyfirst = 0x50;
				index[1].keyend = 0x50;

				index[2].keyfirst = 0x52;
				index[2].keyend = 0x53;

				index[3].keyfirst = 0x47;
				index[3].keyend = 0x47;

				index[4].keyfirst = 0x45;
				index[4].keyend = 0x46;

				index[5].keyfirst = 0x04;
				index[5].keyend = 0x06;

				index[6].keyfirst = 0x08;
				index[6].keyend = 0x08;

				index[7].keyfirst = 0x41;
				index[7].keyend = 0x43;

				index[8].keyfirst = 0x51;
				index[8].keyend = 0x51;

				index[9].keyfirst = 0x54;
				index[9].keyend = 0x54;

				index[10].keyfirst = 0x140;
				index[10].keyend = 0x140;

				index[11].keyfirst = 0x44;
				index[11].keyend = 0x44;

				index[12].keyfirst = 0x07;
				index[12].keyend = 0x07;

				index[13].keyfirst = 0x40;
				index[13].keyend = 0x40;

				index[14].keyfirst = 0x120;
				index[14].keyend = 0x13F;

				index[15].keyfirst = 0x100;
				index[15].keyend = 0x11F;

				index[16].keyfirst = 0x30;
				index[16].keyend = 0x3F;

				index[17].keyfirst = 0x10;
				index[17].keyend = 0x2F;

				*n = 18;
			}
		}
		else // TA88
		{
			if (1)
			{
				// got from america
				index[0].keyfirst = 0x120;
				index[0].keyend = 0x13F;

				index[1].keyfirst = 0x100;
				index[1].keyend = 0x11F;

				index[2].keyfirst = 0x10;
				index[2].keyend = 0x2F;

				index[3].keyfirst = 0x0F;
				index[3].keyend = 0x0F;

				index[4].keyfirst = 0x50;
				index[4].keyend = 0x50;

				index[5].keyfirst = 0x52;
				index[5].keyend = 0x53;

				index[6].keyfirst = 0x47;
				index[6].keyend = 0x47;

				index[7].keyfirst = 0x45;
				index[7].keyend = 0x46;

				index[8].keyfirst = 0x04;
				index[8].keyend = 0x06;

				index[9].keyfirst = 0x08;
				index[9].keyend = 0x08;

				index[10].keyfirst = 0x41;
				index[10].keyend = 0x43;

				index[11].keyfirst = 0x51;
				index[11].keyend = 0x51;

				index[12].keyfirst = 0x54;
				index[12].keyend = 0x54;

				index[13].keyfirst = 0x140;
				index[13].keyend = 0x140;

				index[14].keyfirst = 0x44;
				index[14].keyend = 0x44;

				index[15].keyfirst = 0x07;
				index[15].keyend = 0x07;

				index[16].keyfirst = 0x40;
				index[16].keyend = 0x40;

				index[17].keyfirst = 0x30;
				index[17].keyend = 0x3F;

				*n = 18;
			}
		}
	}
	else
	{
		pspSdkSetK1(k1);
		return -1;
	}

	pspSdkSetK1(k1);
	return 0;
}

int idsRegenerationGetHwConfigKeys(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	memset(buf, 0, 512*3);

	if (psptype == 1)
	{
		memcpy(buf, baryon_type1, sizeof(baryon_type1));
		memcpy(buf+512, clockgen_type1, sizeof(clockgen_type1));
		memcpy(buf+1024, ddr_type1, sizeof(ddr_type1));
	}
	else if (psptype == 2)
	{
		memcpy(buf, baryon_type2, sizeof(baryon_type2));
		memcpy(buf+512, clockgen_type2, sizeof(clockgen_type2));
		memcpy(buf+1024, ddr_type2, sizeof(ddr_type2));
	}
	else if (psptype == 3)
	{
		memcpy(buf, baryon_type3, sizeof(baryon_type3));
		memcpy(buf+512, clockgen_type3, sizeof(clockgen_type3));
		memcpy(buf+1024, ddr_type3, sizeof(ddr_type3));
	}
	else
	{
		pspSdkSetK1(k1);
		return -1;
	}

	pspSdkSetK1(k1);
	return 1;
}

int idsRegenerationGetMGKeys(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	memset(buf, 0, 512*0x20);

	if (psptype == 1)
	{
		memcpy(buf, mg_81_AM_25A00D7C2958D, sizeof(mg_81_AM_25A00D7C2958D));
	}
	else if (psptype == 2)
	{
		memcpy(buf, mg_86_EU_PINK, sizeof(mg_86_EU_PINK));
	}
	else if (psptype == 3)
	{
		memcpy(buf, mg_85v2_EU_PLATA, sizeof(mg_85v2_EU_PLATA));
	}
	else
	{
		pspSdkSetK1(k1);
		return -1;
	}

	pspSdkSetK1(k1);
	return 1;
}

int idsRegenerationGetFactoryBadBlocksKey(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	memset(buf, 0, 512);

	pspSdkSetK1(k1);
	return 1;
}

int idsRegenerationGetSerialKey(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	memset(buf, 0, 512);

	if (psptype == 1)
	{
		memcpy(buf, serial_type1, sizeof(serial_type1));		
	}
	else if (psptype == 2)
	{
		memcpy(buf, serial_type2, sizeof(serial_type2));
	}
	else if (psptype == 3)
	{
		memcpy(buf, serial_type3, sizeof(serial_type3));
	}
	else
	{
		pspSdkSetK1(k1);
		return -1;
	}

	buf[0x1F] = g_region;

	pspSdkSetK1(k1);
	return 1;
}

int idsRegenerationGetWlanKey(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	memset(buf, 0, 512);

	if (g_region == JAPAN)
	{
		buf[0] = 3;
	}
	else if (g_region == AMERICA)
	{
		buf[0] = 0;
	}
	else if (g_region == EUROPE)
	{
		buf[0] = 2;
	}
	else
	{
		pspSdkSetK1(k1);
		return -1;
	}

	if (!isslim)
	{
		buf[2] = 1;
	}
	else
	{
		buf[1] = 0x10;
		buf[2] = 1;
		buf[3] = 1;
		buf[4] = 2;
	}

	pspSdkSetK1(k1);
	return 1;
}

int idsRegenerationGetAudioVolumeSetupKey(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	memset(buf, 0, 512);

	pspSdkSetK1(k1);
	return 1;
}

int idsRegenerationGetUsbKeys(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	memset(buf, 0, 512*3);

	if (!isslim)
	{
		memcpy(buf, usb41_1000, sizeof(usb41_1000));
		memcpy(buf+1024, usb43_1000, sizeof(usb43_1000));
	}
	else
	{
		memcpy(buf, usb41_2000, sizeof(usb41_2000));
		memcpy(buf+1024, usb43_2000, sizeof(usb43_2000));
	}

	pspSdkSetK1(k1);
	return 1;
}

int idsRegenerationGetUnkKey140(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	memset(buf, 0, 512);

	if (psptype == 1)
	{
		memcpy(buf, key140_type1, sizeof(key140_type1));
	}

	else if (psptype == 2)
	{
		memcpy(buf, key140_type2, sizeof(key140_type2));
	}

	else if (psptype == 3)
	{
		memcpy(buf, key140_type3, sizeof(key140_type3));
	}
	else
	{
		pspSdkSetK1(k1);
		return -1;
	}

	pspSdkSetK1(k1);
	return 1;
}

int idsRegenerationGetMGKey40(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	memset(buf, 0, 512);
	
	if (psptype == 1)
	{
		memcpy(buf, mg_81_AM_25A00D7C2958D+0x88, 8);
	}
	else if (psptype == 2)
	{
		memcpy(buf, mg_86_EU_PINK+0x88, 8);
	}
	else if (psptype == 3)
	{
		memcpy(buf, mg_85v2_EU_PLATA+0x88, 8);
	}
	else
	{
		pspSdkSetK1(k1);
		return -1;
	}

	pspSdkSetK1(k1);
	return 1;
}

int idsRegenerationGetUnkKeys3X(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	memset(buf, 0, 512*0x10);

	pspSdkSetK1(k1);
	return 1;
}

int idsRegenerationGetParentalLockKey(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	if (psptype == 1)
	{
		pspSdkSetK1(k1);
		return 0;
	}

	memset(buf, 0, 512);
	buf[0] = 9;

	pspSdkSetK1(k1);
	return 1;
}

u8 g_plain[0x800];

static void getUniqueId(u8 *id)
{
    U64 fuse;

	memcpy(&fuse, &g_fuseid, 8);
	
	u32 reg1 = fuse.low;
    u32 reg2 = fuse.high;

    id[0] = reg2 >> 24;
    id[1] = reg2 >> 16;
    id[2] = reg2 >>  8;
    id[3] = reg2 >>  0;

    id[4] = reg1 >> 24;
    id[5] = reg1 >> 16;
    id[6] = reg1 >>  8;
    id[7] = reg1 >>  0;
}

#define SCE_SHA1_BLOCK_SIZE		64
#define SCE_SHA1_DIGEST_SIZE	20

unsigned char intdigest[SCE_SHA1_DIGEST_SIZE];
unsigned char keybuf[SCE_SHA1_BLOCK_SIZE];

/*int WriteFile(char *file, void *buf, int size)
{
	SceUID fd = sceIoOpen(file, PSP_O_WRONLY | PSP_O_CREAT | PSP_O_TRUNC, 0777);
	sceIoWrite(fd, buf, size);
	sceIoClose(fd);

	return 0;
}*/

/* Function to compute the HMAC digest(RFC2202) */
static void genHmacSha1(unsigned char *secretkey, unsigned int keylen, unsigned char *plain, unsigned int datalen, unsigned char *digest)
{
	SceKernelUtilsSha1Context ctx;
	int     i;

	memset(keybuf, 0x00, sizeof(keybuf));
	if (keylen <= SCE_SHA1_BLOCK_SIZE) {

		memcpy(keybuf, secretkey, keylen);
	} else {

		sceKernelUtilsSha1Digest(secretkey, keylen, keybuf);
		keylen = SCE_SHA1_DIGEST_SIZE;
	}


	for (i=0; i<SCE_SHA1_BLOCK_SIZE; i++) {
		keybuf[i] ^= 0x36;
	}
	sceKernelUtilsSha1BlockInit(&ctx);
	sceKernelUtilsSha1BlockUpdate(&ctx, keybuf, SCE_SHA1_BLOCK_SIZE);


	sceKernelUtilsSha1BlockUpdate(&ctx, plain, datalen);


	sceKernelUtilsSha1BlockResult(&ctx, intdigest);



	for (i=0; i<SCE_SHA1_BLOCK_SIZE; i++) {
		keybuf[i] ^= (0x36 ^ 0x5C);
	}
	sceKernelUtilsSha1BlockInit(&ctx);
	sceKernelUtilsSha1BlockUpdate(&ctx, keybuf, SCE_SHA1_BLOCK_SIZE);


	sceKernelUtilsSha1BlockUpdate(&ctx, intdigest, SCE_SHA1_DIGEST_SIZE);


	sceKernelUtilsSha1BlockResult(&ctx, digest);

	return;
}

static int genSignedHash(int size, u8 *buf, u8 *digest)
{
    u8 pspid[8];
    int i, j;

    if (size > 0x100)
    {
        //Kprintf("buffer size error...\n");
        return -1;
    }
    
    memset(g_plain, 0, sizeof(g_plain));
    memcpy(g_plain, buf, size);

    getUniqueId(pspid);
    
    /*for(i = 0; i < size; i += 8)
    {
		for(j = 0; j < 8; j++)
			g_plain[i+j] ^= pspid[j];
    }*/
	for(i = 0; i < size; i ++)
    {
		for(j = 0; j < 8; j++)
			g_plain[(i*8)+j] ^= pspid[j];
    }

    genHmacSha1((u8 *)"specialweek", 11, g_plain, 0x100, digest);
    
    return 1;
}

int _idsRegenerationGenerateFactoryFirmwareKey(u8 *buf)
{
	u32 hash[5];
	u8 *vbuf;
	int size;

	memset(buf, 0, 512);

	if (psptype == 1)
	{
		return 0;
	}

	else if (psptype == 2)
	{
		if (g_mb == TA82)
		{
			strcpy((char *)buf, "2.60");
			vbuf = version260;
			size = sizeof(version260);
		}
		else
		{
			strcpy((char *)buf, "2.81");
			vbuf = version281;
			size = sizeof(version281);
		}
	}

	else if (psptype == 3)
	{
		if (g_mb == TA85)
		{
			strcpy((char *)buf, "3.60");
			vbuf = version360;
			size = sizeof(version360);
		}
		else if (g_mb == TA85v2)
		{
			strcpy((char *)buf, "3.72");
			vbuf = version372;
			size = sizeof(version372);
		}
		else
		{
			strcpy((char *)buf, "3.90");
			vbuf = version390;
			size = sizeof(version390);
		}
	}
	else
	{
		return -1;
	}

	memset(hash, 0, sizeof(hash));
	genSignedHash(size, vbuf, (void *)hash);

	memcpy(buf+8, &hash[0], 4);
	memcpy(buf+12, &hash[1], 4);

	return 1;
}

int idsRegenerationGenerateFactoryFirmwareKey(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	int res = sceKernelExtendKernelStack(0x4000, (void *)_idsRegenerationGenerateFactoryFirmwareKey, buf);

	pspSdkSetK1(k1);
	return res;
}

int idsRegenerationGetLCDKey(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	if (psptype == 1 || g_mb == TA82)
	{
		pspSdkSetK1(k1);
		return 0;
	}

	memset(buf, 0, 512);

	if (!isslim)
	{
		memcpy(buf, lcd_86, sizeof(lcd_86));
	}
	else
	{
		if (g_mb == TA85)
		{
			memcpy(buf, lcd_type73, sizeof(lcd_type73));
		}
		else if (g_mb == TA85v2)
		{
			memcpy(buf, lcd_type6C, sizeof(lcd_type6C));
		}
		else
		{
			memcpy(buf, lcd_type71, sizeof(lcd_type6C));
		}
	}

	pspSdkSetK1(k1);
	return 1;
}

int _idsRegenerationGenerateCallibrationKey(u8 *buf)
{
	//int i;
	//u8 setParam[8];
	//int res;

	if (psptype == 1 || g_mb == TA82)
	{
		return 0;
	}

	memset(buf, 0, 512);

	/*for(i = 0; i < 3; i++)
    {
		res = sceSysconReceiveSetParam(1, setParam);

		if (res == 0)
        {
			break;
        }
        else if (res == 0x80250011)
        {
			return res;
        }
    }

    if (i == 3)
		return -2;*/

	u32 *header = (u32*)(buf+0x00);
    header[0] = 0x41506144; // APaD
	header[1] = 1;          // type
	header[2] = 8;          // dataLen
	memcpy(buf+0x10, setParam, 8);
	header[3] = gen_hash(0xFFFFFFFF,buf+0x10, 0x1F0);	

	return 1;
}

int idsRegenerationGenerateCallibrationKey(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	int res = sceKernelExtendKernelStack(0x2000, (void *)_idsRegenerationGenerateCallibrationKey, buf);

	pspSdkSetK1(k1);
	return res;
}

int idsRegenerationGetUnkKeys5253(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	if (!isslim)
	{
		pspSdkSetK1(k1);
		return 0;
	}

	memset(buf, 0, 1024);

	if (g_mb == TA85)
	{
		if (g_region == EUROPE)
		{
			memcpy(buf, key52_85_EU, sizeof(key52_85_EU));
		}
		else if (g_region == AMERICA)
		{
			memcpy(buf, key52_85_AM, sizeof(key52_85_AM));
		}
		else
		{
			memcpy(buf, key52_85_JAP, sizeof(key52_85_JAP));
		}
	}
	else if (g_mb == TA85v2)
	{
		if (g_region == EUROPE)
		{
			memcpy(buf, key52_85v2_EU, sizeof(key52_85v2_EU));
		}
		else if (g_region == AMERICA)
		{
			memcpy(buf, key52_85v2_AM, sizeof(key52_85v2_AM));
		}
		else
		{
			memcpy(buf, key52_85v2_JAP, sizeof(key52_85v2_JAP));
		}
	}
	else
	{
		if (g_region == EUROPE)
		{
			memcpy(buf, key52_88_EU, sizeof(key52_88_EU));
		}
		else if (g_region == AMERICA)
		{
			memcpy(buf, key52_88_AM, sizeof(key52_88_AM));
		}
		else
		{
			memcpy(buf, key52_88_JAP, sizeof(key52_88_JAP));
		}
	}

	pspSdkSetK1(k1);
	return 1;
}

int idsRegenerationGetDefaultXMBColorKey(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	if (!isslim)
	{
		pspSdkSetK1(k1);
		return 0;
	}

	memset(buf, 0, 512);
	buf[0] = buf[2] = 2;

	pspSdkSetK1(k1);
	return 1;
}

typedef struct
{
	u8  buf1[8]; // 0
	u8  buf2[8]; // 8
	u8  buf3[8]; // 0x10
} SomeStructure;

u8 idskey0[16] = 
{
	0x47, 0x5E, 0x09, 0xF4, 0xA2, 0x37, 0xDA, 0x9B, 
	0xEF, 0xFF, 0x3B, 0xC0, 0x77, 0x14, 0x3D, 0x8A
};

void P1(u8 *buf)
{
	u8 m[0x40]; // sp
	int i, j;
	char rr[5] = { 0x01, 0x0F, 0x36, 0x78, 0x40 }; // sp+0x50
	u8 ss[5]; // sp+0x40
	
	memset(m, 0, 0x40);
	memcpy(m, buf, 0x3C);

	for (i = 0; i < 0x3C; i++)
	{
		for (j = 0; j < 5; j++)
		{
			u8 x;
			u32 y;
			char c;

			x = 0;			
			c = rr[j];
			y = m[i];
			
			while (c != 0)
			{
				if (c & 1)
				{
					x ^= y;
				}

				y = (y << 1);
					
				if ((y & 0x0100))
				{
					y ^= 0x11D;
				}

				c /= 2;
			}

			ss[j] = x;
		}

		for (j = 0; j < 5; j++)
		{
			m[i+j] ^= ss[j]; 
		}
	}

	for (i = 0x3C; i < 0x40; i++)
	{
		buf[i] = m[i];
		m[i] = 0;
	}
}

void GenerateSigncheck(SomeStructure *ss, int *b, u8 *out)
{
	AES_KEY ctx, ctx2; // sp+0x20
	int i, j;
	u8 sg_key1[0x10], sg_key2[0x10]; // sp, sp+0x10
		
	AES_set_encrypt_key(idskey0, 128, &ctx);
	AES_set_decrypt_key(idskey0, 128, &ctx2);

	for (i = 0; i < 16; i++)
	{
		sg_key1[i] = sg_key2[i] = ss->buf1[i % 8];
	}

	for (i = 0; i < 3; i++)
	{
		AES_encrypt(sg_key2, sg_key2, &ctx);
		AES_decrypt(sg_key1, sg_key1, &ctx2);
	}

	AES_set_encrypt_key(sg_key2, 128, &ctx);

	for (i = 0; i < 3; i++)
	{
		for (j = 0; j < 3; j++)
		{
			AES_encrypt(sg_key1, sg_key1, &ctx);
		}

		memcpy(out+(i*16), sg_key1, 0x10);		
	}

	memcpy(out+0x30, ss->buf1, 8);
	memcpy(out+0x38, b, 4);

	P1(out);
}

int EncryptRegion(u8 *scheck, u8 *in, u32 size, u8 *out)
{
	const int n = 3; // param t0
	int a;
	int i, j;
	u8  k1[0x10], k2[0x10]; // sp, sp+0x110
	u8 *r;
	AES_KEY ctx; // (fp=sp+0x10)
	int v = size / 16;
	
	a = (size & 0x0F);

	if (a != 0 || (size == 0 && (n & 1)))
	{
		if ((n & 2) == 0)
			return -1;
		
		v++;
		r = k1;
		
		for (i = 0; i < a; i++)
		{
			k1[i] = in[size-a+i];
		}

		k1[i++] = 0x80;

		for (; i < 0x10; i++)
		{
			k1[i] = 0;
		}

		a = 0x10;
	}
	else
	{
		r = (in+(v*16))-0x10;
	}

	AES_set_encrypt_key(scheck, 128, &ctx);

	if ((n & 1))
	{
		memset(out, 0, 16);
	}
	else if (v > 0)
	{
		AES_encrypt(out, out, &ctx);
	}

	for (i = 0; i < (v-1); i++)
	{
		for (j = 0; j < 16; j++)
		{
			out[j] ^= in[i*16+j];
		}

		AES_encrypt(out, out, &ctx);
	}	

	if (v > 0)
	{
		for (i = 0; i < 0x10; i++)
		{
			out[i] ^= r[i];
		}
	}

	if (!(n & 2))
	{
		return 0;
	}

	memset(k2, 0, 0x10);
	AES_encrypt(k2, k2, &ctx);

	a = a ? 2 : 1;
	
	for (i = 0; i < a; i++)
	{
		u32 x = ((char)k2[0] < 0);
		
		for (j = 0; j < 15; j++)
		{
			u32 t = ((k2[j+1] >> 7) | (k2[j] << 1));
			k2[j] = t;
		}

		k2[15] <<= 1;

		if (x == 1)
		{
			int t = (int)k2[15] ^ 0xFFFFFF87;
			k2[15] = t;
		}
	}

	for (i = 0; i < 0x10; i++)
	{
		out[i] ^= k2[i];		
	}

	AES_encrypt(out, out, &ctx);

	return 0;
}

typedef struct UmdIndex
{
  int t;
  int m;
} UmdIndex;

typedef struct Param3
{
  int unk1;        //+0x0
  u8 buf1[0x10]; //+0x4
  u8 buf2[0x10]; //+0x14
  UmdIndex outbuf[0x1000]; //+0x24 doesn't matter the size
}  Param3;


int sub_00003E38(u32 a0, u32 a1, u32 a2, u32 a3)
{
	if (a1 < a3) return 0;

	if (a3 < a1) return 2;

	if (a0 < a2) return 0;

	if (a2 >= a0)
	{
		return 1;
	} 
  
	return 2;
}

typedef struct UmdKey
{
	u32 t;
	u32 m;
	u8 key[0x10];
} UmdKey;

#define N_KEYS	772

UmdKey umdkeys[N_KEYS] =
{

// 0-78 -> TA-85 v2 euro
// 79-111 -> from TA86 euro
// 112-164 -> from TA-79 jap

{
	0x00000002, 0x80000000,
	{0x33, 0x50, 0x48, 0xCE, 0x08, 0x5E, 0xF3, 0xDD, 0xFE, 0x8D, 0x69, 0xE3, 0x81, 0xD7, 0x2B, 0xCF }
}, // 0

{
	0x0000000F, 0x80000000,
	{0x5D, 0xBA, 0xFE, 0xA7, 0xA9, 0xFB, 0x7D, 0x9B, 0x65, 0x10, 0x8B, 0x6D, 0xA7, 0xDE, 0x31, 0x19 }
}, // 1

{
	0x00000012, 0x80000000,
	{0xB7, 0x6D, 0xC0, 0xD3, 0xCC, 0xC5, 0x85, 0xFE, 0x85, 0xD6, 0x7A, 0xFE, 0x34, 0x2F, 0xD7, 0x14 }
}, // 2

{
	0x0000001F, 0x80000000,
	{0x47, 0xC8, 0x1A, 0xFF, 0x27, 0x10, 0x41, 0x64, 0x55, 0x24, 0xEE, 0xB8, 0xB3, 0x8A, 0xEE, 0x62 }
}, // 3

{
	0x00000022, 0x80000000,
	{0x67, 0x8B, 0x6A, 0x33, 0x98, 0x5C, 0x45, 0x19, 0x46, 0x07, 0xF0, 0x4E, 0x1D, 0x70, 0xEA, 0x1B }
}, // 4

{
	0x0000002F, 0x80000000,
	{0xCE, 0x5E, 0x6A, 0x10, 0x7C, 0x3F, 0x9A, 0x18, 0xCF, 0xF2, 0xB4, 0xA9, 0x16, 0xB0, 0x8E, 0xA6 }
}, // 5

{
	0x00000032, 0x80000000,
	{0xF5, 0x43, 0xE8, 0xE5, 0x39, 0x78, 0x5C, 0xEA, 0x9B, 0x91, 0x85, 0x1A, 0x30, 0x5E, 0x75, 0x1E }
}, // 6

{
	0x0000003F, 0x80000000,
	{0x49, 0xF1, 0xFA, 0x16, 0xBE, 0x5F, 0xC2, 0x00, 0xDC, 0x8D, 0x50, 0xD2, 0x6F, 0x1D, 0x5F, 0x48 }
}, // 7

{
	0x00000042, 0x80000000,
	{0x73, 0x97, 0x32, 0x8B, 0x36, 0x28, 0xBC, 0x51, 0x29, 0xEE, 0x7E, 0x82, 0x8D, 0x26, 0xD5, 0x24 }
}, // 8

{
	0x0000004F, 0x80000000,
	{0x65, 0xBE, 0x3D, 0x73, 0x26, 0x55, 0xE9, 0x05, 0x74, 0xF0, 0x06, 0x1A, 0xDA, 0x2C, 0xB8, 0x62 }
}, // 9

{
	0x1000000F, 0x80000000,
	{0x71, 0x16, 0x41, 0x21, 0xC3, 0xE0, 0x6A, 0xB5, 0x76, 0x57, 0x35, 0x90, 0xAB, 0x22, 0x3E, 0xFD }
}, // 10

{
	0x1000001F, 0x80000000,
	{0x42, 0x01, 0x82, 0x19, 0x24, 0xFF, 0x0C, 0x34, 0xEB, 0xB4, 0x3F, 0x5D, 0xCA, 0x80, 0x7E, 0x8D }
}, // 11

{
	0x1000002F, 0x80000000,
	{0x40, 0x78, 0x66, 0xEB, 0x9B, 0xA5, 0xE9, 0x29, 0x0B, 0x2F, 0x8E, 0x4D, 0x30, 0x11, 0x67, 0x54 }
}, // 12

{
	0x1000003F, 0x80000000,
	{0x27, 0xC7, 0x59, 0xF1, 0x48, 0xD8, 0x7E, 0xA8, 0xC0, 0xC2, 0x7F, 0x01, 0x3A, 0xE6, 0xEC, 0xD6 }
}, // 13

{
	0x1000004F, 0x80000000,
	{0x05, 0x8F, 0xE5, 0x35, 0xA0, 0x4C, 0x89, 0xB1, 0xC0, 0x12, 0xAD, 0xE7, 0xDA, 0x06, 0xD9, 0x22 }
}, // 14

{
	0x2000000F, 0x80000000,
	{0x1A, 0x4A, 0xA0, 0x55, 0x30, 0x8D, 0x2B, 0xBC, 0x63, 0x2B, 0x54, 0x68, 0xBB, 0x3A, 0x9D, 0x5B }
}, // 15

{
	0x00000001, 0x00000000,
	{0xFD, 0x61, 0xF4, 0xC2, 0x39, 0x3E, 0x1C, 0xA0, 0xC2, 0x7F, 0xF8, 0x55, 0x04, 0x2B, 0x8B, 0xDC }
}, // 16

{
	0x00000002, 0x00000000,
	{0xD0, 0x5E, 0x62, 0x43, 0xD6, 0x79, 0x5F, 0xCD, 0x98, 0x59, 0x12, 0xF0, 0x1C, 0xC4, 0xE2, 0x5F }
}, // 17

{
	0x00000004, 0x00000000,
	{0xDE, 0x08, 0x48, 0xC0, 0x29, 0xD2, 0x75, 0x5B, 0x69, 0x43, 0x42, 0x8A, 0xA5, 0x69, 0xB3, 0xF7 }
}, // 18

{
	0x00000008, 0x00000000,
	{0xD4, 0xF8, 0x16, 0x0E, 0xAF, 0x2D, 0x38, 0xEC, 0x1D, 0x2A, 0xDC, 0x71, 0x3C, 0xFE, 0x9E, 0x6C }
}, // 19

{
	0x00000010, 0x00000000,
	{0x72, 0x77, 0x23, 0x01, 0x6E, 0x91, 0x15, 0x43, 0x0F, 0xDD, 0x6B, 0xDF, 0x71, 0xE8, 0x03, 0x88 }
}, // 20

{
	0x00000020, 0x00000000,
	{0x30, 0x04, 0xF2, 0xAA, 0x88, 0xCE, 0x10, 0x65, 0xA5, 0x7D, 0xE5, 0x9F, 0x38, 0xDE, 0x1F, 0x34 }
}, // 21

{
	0x00000040, 0x00000000,
	{0xE1, 0x9D, 0x85, 0xD3, 0x5D, 0x21, 0x62, 0xB5, 0x62, 0x05, 0xC7, 0xFF, 0x92, 0xB8, 0xB6, 0x98 }
}, // 22

{
	0x00000080, 0x00000000,
	{0x26, 0xB7, 0x0A, 0x89, 0xD7, 0xB3, 0xD2, 0x5B, 0x30, 0x34, 0xE3, 0xC9, 0xB0, 0x38, 0xC5, 0x9A }
}, // 23

{
	0x00000101, 0x00000000,
	{0xE7, 0x87, 0xE2, 0xD5, 0x46, 0x8A, 0x2D, 0xDE, 0x0B, 0x32, 0x1E, 0x06, 0xE4, 0x98, 0x43, 0xEA }
}, // 24

{
	0x00000202, 0x00000000,
	{0xFA, 0x67, 0x31, 0x50, 0x1A, 0x19, 0xB3, 0x75, 0x74, 0x7A, 0x62, 0xB0, 0x0C, 0x09, 0xEC, 0x69 }
}, // 25

{
	0x00000404, 0x00000000,
	{0x80, 0xD4, 0xE0, 0x7D, 0x74, 0xE1, 0xA0, 0x2F, 0x21, 0x8E, 0x1B, 0x74, 0x88, 0xA4, 0x7C, 0x5D }
}, // 26

{
	0x00000808, 0x00000000,
	{0x67, 0x30, 0x2C, 0xF7, 0xCB, 0x97, 0x6E, 0xC3, 0xA8, 0x92, 0x00, 0x8D, 0xA3, 0x68, 0x86, 0x1B }
}, // 27

{
	0x00001010, 0x00000000,
	{0xC9, 0xF4, 0x1F, 0xD2, 0xA3, 0x43, 0xD8, 0x5F, 0xBA, 0x0A, 0xA2, 0x45, 0x5A, 0x5C, 0x0F, 0x8A }
}, // 28

{
	0x00002020, 0x00000000,
	{0xB3, 0x85, 0xBE, 0x2A, 0xE7, 0xDF, 0xE0, 0x90, 0x97, 0x53, 0x3E, 0xE8, 0xEC, 0xAC, 0xB2, 0xAB }
}, // 29

{
	0x00004041, 0x00000000,
	{0xCF, 0x99, 0x22, 0xD5, 0x37, 0x9A, 0xF1, 0xC5, 0x03, 0xAB, 0x43, 0x80, 0x5D, 0xEF, 0x3E, 0xA8 }
}, // 30

{
	0x00008082, 0x00000000,
	{0xC1, 0xFF, 0xFA, 0xCC, 0x92, 0x95, 0xDF, 0x19, 0xED, 0xBB, 0x7E, 0x8C, 0x5E, 0x7B, 0x4D, 0x2F }
}, // 31

{
	0x00010105, 0x00000000,
	{0xDA, 0xD4, 0xB7, 0x2E, 0x71, 0xC4, 0x09, 0xDE, 0x41, 0x2D, 0x1A, 0x4C, 0x06, 0x4E, 0x20, 0x32 }
}, // 32

{
	0x0002020A, 0x00000000,
	{0x32, 0x56, 0x45, 0xB7, 0x4F, 0x56, 0x52, 0xFA, 0x47, 0x21, 0xE5, 0x9F, 0x6D, 0xAA, 0xE2, 0xB3 }
}, // 33

{
	0x00040414, 0x00000000,
	{0xCB, 0xFE, 0xC2, 0x20, 0x5A, 0x12, 0xBE, 0x52, 0xCF, 0x99, 0x24, 0x34, 0x1D, 0x84, 0x5E, 0x5F }
}, // 34

{
	0x00080828, 0x00000000,
	{0xAC, 0x92, 0x61, 0xD1, 0xE9, 0x5B, 0xF1, 0xDE, 0x06, 0x2D, 0x28, 0x8E, 0x1B, 0x44, 0xD9, 0x97 }
}, // 35

{
	0x00101050, 0x00000000,
	{0x0C, 0x38, 0x63, 0x17, 0x5B, 0x01, 0x13, 0x80, 0x59, 0x6F, 0x9A, 0xAB, 0x29, 0xB8, 0x9B, 0xE1 }
}, // 36

{
	0x002020A0, 0x00000000,
	{0x4E, 0x1F, 0x91, 0xC1, 0x0A, 0xF0, 0xFB, 0x36, 0xEB, 0x72, 0x70, 0x9D, 0x08, 0x06, 0x4E, 0x51 }
}, // 37

{
	0x00404140, 0x00000000,
	{0xBB, 0x77, 0xAF, 0x29, 0xD0, 0x12, 0x34, 0x8E, 0x09, 0x36, 0xD6, 0x97, 0xF0, 0xD3, 0x45, 0xDE }
}, // 38

{
	0x00808280, 0x00000000,
	{0xFF, 0xE5, 0xCA, 0x92, 0x03, 0xC1, 0xAA, 0x92, 0x02, 0xD2, 0xB7, 0x40, 0x5C, 0x95, 0x1C, 0xE4 }
}, // 39

{
	0x01010500, 0x00000000,
	{0x68, 0x98, 0xD7, 0xD2, 0x67, 0x50, 0x17, 0xAE, 0xB3, 0x1B, 0x12, 0x36, 0x72, 0x43, 0x43, 0x15 }
}, // 40

{
	0x02020A00, 0x00000000,
	{0xEC, 0xA5, 0x90, 0x0A, 0x10, 0x5C, 0x90, 0xFB, 0xED, 0x28, 0xD7, 0xE9, 0x52, 0x28, 0x73, 0x3E }
}, // 41

{
	0x04041401, 0x00000000,
	{0x91, 0x56, 0x58, 0x7C, 0x59, 0xFA, 0x02, 0x0A, 0x0E, 0xE6, 0x28, 0xD5, 0x86, 0x31, 0xA8, 0xA4 }
}, // 42

{
	0x08082803, 0x00000000,
	{0xB3, 0x05, 0xFB, 0x2D, 0x15, 0xDD, 0x96, 0x8B, 0x7F, 0x9E, 0x49, 0x33, 0x17, 0xF5, 0xC0, 0x32 }
}, // 43

{
	0x10105006, 0x00000000,
	{0xF1, 0x58, 0x34, 0x46, 0x58, 0xAA, 0x06, 0x70, 0xDD, 0x07, 0x8F, 0x4C, 0x1A, 0x3F, 0x30, 0x83 }
}, // 44

{
	0x2020A00C, 0x00000000,
	{0x48, 0x44, 0x07, 0x5E, 0xD2, 0x9F, 0x0B, 0xD4, 0x11, 0x8C, 0xC7, 0x69, 0x59, 0x63, 0xFB, 0xB6 }
}, // 45

{
	0x40414018, 0x00000000,
	{0xC0, 0x85, 0x59, 0x52, 0xC3, 0x36, 0x9F, 0x94, 0xD3, 0x68, 0x2F, 0x8C, 0xD8, 0xB2, 0x9B, 0xAE }
}, // 46

{
	0x80828030, 0x00000000,
	{0x7D, 0xE3, 0x2D, 0x1C, 0x53, 0x05, 0x0D, 0x18, 0x47, 0x06, 0x42, 0x41, 0x58, 0x8E, 0xC3, 0xA3 }
}, // 47

{
	0x01050061, 0x00000001,
	{0xC0, 0x6A, 0xC0, 0x6B, 0xB5, 0xB5, 0x33, 0xB1, 0xD5, 0xF3, 0x3C, 0xD4, 0x66, 0xBA, 0x9E, 0x7F }
}, // 48

{
	0x020A00C3, 0x00000002,
	{0x34, 0xB2, 0xD0, 0x97, 0x53, 0xC1, 0x56, 0x22, 0x89, 0xAB, 0x08, 0x63, 0x24, 0x68, 0xD9, 0x24 }
}, // 49

{
	0x04140186, 0x00000004,
	{0x89, 0x39, 0x8A, 0x7B, 0x17, 0x55, 0x02, 0x26, 0x75, 0xEB, 0xDE, 0x94, 0x91, 0x9E, 0xB3, 0x89 }
}, // 50

{
	0x0828030C, 0x00000008,
	{0xD9, 0x2A, 0xDB, 0x67, 0xCC, 0x7C, 0xA2, 0x2E, 0x64, 0x74, 0x48, 0x30, 0xF4, 0xED, 0x78, 0xAA }
}, // 51

{
	0x10500618, 0x00000010,
	{0x2B, 0x71, 0xEE, 0x0A, 0xD1, 0x2F, 0xC1, 0xEB, 0xE0, 0x76, 0x88, 0xEB, 0x3F, 0xFE, 0xDF, 0x37 }
}, // 52

{
	0x20A00C30, 0x00000020,
	{0x57, 0xE0, 0x51, 0x3E, 0x3F, 0x2D, 0xA5, 0xE4, 0x54, 0x69, 0x2D, 0x97, 0x7E, 0x17, 0x17, 0x87 }
}, // 53

{
	0x41401860, 0x00000040,
	{0x12, 0x76, 0x66, 0x75, 0x4F, 0x6E, 0x64, 0xAF, 0xF9, 0xAE, 0xB0, 0x5B, 0x99, 0x9B, 0xB2, 0xB9 }
}, // 54

{
	0x828030C1, 0x00000080,
	{0xCE, 0x43, 0x97, 0x82, 0xF8, 0x97, 0xF6, 0xCA, 0x1B, 0xF9, 0x62, 0x61, 0xAE, 0xA2, 0x4A, 0x27 }
}, // 55

{
	0x05006183, 0x00000101,
	{0x6B, 0x1D, 0xE7, 0xAA, 0x8B, 0x3A, 0x65, 0xE1, 0xD7, 0x33, 0xEC, 0x5E, 0x31, 0x09, 0xF9, 0x40 }
}, // 56

{
	0x0A00C307, 0x00000202,
	{0xA7, 0x23, 0xAA, 0x3C, 0x8E, 0x81, 0xBC, 0x3C, 0x5C, 0x8F, 0xC8, 0x8E, 0x56, 0x3A, 0xE0, 0x5F }
}, // 57

{
	0x1401860F, 0x00000404,
	{0x51, 0x4A, 0x55, 0x9B, 0x2E, 0x81, 0x6F, 0x83, 0x60, 0x36, 0x88, 0xFA, 0x9F, 0xE9, 0xDB, 0x5E }
}, // 58

{
	0x28030C1E, 0x00000808,
	{0x87, 0xE2, 0x65, 0x13, 0xE4, 0xBF, 0x4C, 0xF5, 0x56, 0xAA, 0xC3, 0xD5, 0x59, 0x8F, 0x19, 0x19 }
}, // 59

{
	0x5006183D, 0x00001010,
	{0x6A, 0x56, 0x83, 0x2F, 0x78, 0xA7, 0xC5, 0xED, 0xA4, 0x90, 0x00, 0x93, 0x28, 0x7D, 0xDE, 0xE3 }
}, // 60

{
	0xA00C307B, 0x00002020,
	{0xC3, 0xFD, 0x72, 0xB7, 0x50, 0xCF, 0x2D, 0x3A, 0xDA, 0x00, 0x84, 0xEA, 0x7F, 0xF4, 0xB0, 0x21 }
}, // 61

{
	0x401860F7, 0x00004041,
	{0x13, 0xCD, 0xFB, 0x84, 0x6E, 0xE4, 0x7F, 0x01, 0x2B, 0x13, 0x97, 0xDF, 0xFD, 0xF7, 0xEE, 0x8D }
}, // 62

{
	0x8030C1EF, 0x00008082,
	{0xE0, 0xFB, 0x34, 0xF0, 0xE8, 0x1E, 0x63, 0xFB, 0x4D, 0x0C, 0x96, 0xC4, 0x46, 0x9A, 0x94, 0x4B }
}, // 63

{
	0x006183DF, 0x00010105,
	{0xF3, 0x0D, 0x9D, 0x0D, 0xB1, 0xB4, 0x57, 0x6C, 0x71, 0x2B, 0x4D, 0x73, 0x56, 0x3C, 0x0E, 0xB3 }
}, // 64

{
	0x00C307BE, 0x0002020A,
	{0xC5, 0xA2, 0x08, 0xBE, 0x15, 0x4B, 0x43, 0x39, 0xA9, 0xAF, 0x15, 0x8A, 0x74, 0xE4, 0xEA, 0x30 }
}, // 65

{
	0x01860F7C, 0x00040414,
	{0xBB, 0x30, 0x7C, 0xB3, 0xFA, 0x1E, 0xF6, 0xE5, 0xC2, 0xE7, 0x35, 0xAB, 0xD5, 0x7B, 0xE4, 0xF0 }
}, // 66

{
	0x030C1EF8, 0x00080828,
	{0x1E, 0xDE, 0x13, 0xE4, 0x4B, 0xB4, 0xDC, 0x6E, 0xB1, 0xBE, 0x62, 0x4B, 0xBE, 0x29, 0xC6, 0x3C }
}, // 67

{
	0x06183DF0, 0x00101050,
	{0x04, 0xF7, 0xEE, 0x30, 0x87, 0xC2, 0x47, 0x98, 0x66, 0x92, 0x61, 0x26, 0x5E, 0xDA, 0x04, 0x9E }
}, // 68

{
	0x0C307BE1, 0x002020A0,
	{0x19, 0x37, 0xCD, 0xFF, 0x8A, 0x7F, 0x5A, 0xF3, 0x66, 0x92, 0xE8, 0x8C, 0xBB, 0x59, 0x16, 0xA6 }
}, // 69

{
	0x1860F7C3, 0x00404140,
	{0x08, 0x6B, 0x77, 0x04, 0x34, 0xA6, 0xA5, 0x5C, 0x59, 0x6E, 0x16, 0xB2, 0x87, 0xD0, 0x86, 0x49 }
}, // 70

{
	0x30C1EF87, 0x00808280,
	{0x34, 0x76, 0x9A, 0x98, 0xAD, 0x11, 0x40, 0x48, 0x7A, 0x40, 0xBA, 0x28, 0x3F, 0xDA, 0xEF, 0x30 }
}, // 71

{
	0x6183DF0F, 0x01010500,
	{0x2E, 0x2D, 0x42, 0x52, 0xEA, 0x1A, 0x74, 0x3D, 0xA8, 0xF2, 0x5C, 0x4E, 0x3D, 0xC7, 0xB4, 0x7B }
}, // 72

{
	0xC307BE1E, 0x02020A00,
	{0x83, 0x3D, 0x43, 0xF6, 0xFE, 0x17, 0x28, 0xF6, 0x73, 0x9C, 0x38, 0x20, 0x2F, 0xAF, 0xF9, 0xF4 }
}, // 73

{
	0x860F7C3C, 0x04041401,
	{0x21, 0xF5, 0x15, 0xB7, 0x24, 0xD5, 0x48, 0x54, 0xAD, 0x43, 0xAD, 0x5D, 0xD9, 0x3B, 0x77, 0xE2 }
}, // 74

{
	0x0C1EF878, 0x08082803,
	{0xEF, 0xCD, 0xD4, 0xF6, 0xE1, 0x3B, 0x9E, 0xC6, 0x79, 0x1C, 0x54, 0xBB, 0xED, 0x85, 0x86, 0x7C }
}, // 75

{
	0x00000001, 0x70000000,
	{0x3A, 0xBA, 0x2E, 0x14, 0xC1, 0x07, 0x37, 0x9F, 0x5F, 0x03, 0x3F, 0x79, 0x4D, 0xE0, 0xCE, 0x7E }
}, // 76

{
	0x00000002, 0x80000000,
	{0x33, 0x50, 0x48, 0xCE, 0x08, 0x5E, 0xF3, 0xDD, 0xFE, 0x8D, 0x69, 0xE3, 0x81, 0xD7, 0x2B, 0xCF }
},// 77

{
	0x00000002, 0x80000000,
	{0x33, 0x50, 0x48, 0xCE, 0x08, 0x5E, 0xF3, 0xDD, 0xFE, 0x8D, 0x69, 0xE3, 0x81, 0xD7, 0x2B, 0xCF }
},// 78

{
	0x08082802, 0x00000000,
	{0xF3, 0xB6, 0xC9, 0x65, 0xE3, 0x0C, 0x67, 0x5E, 0xEA, 0xC2, 0xF7, 0x3C, 0x1A, 0x83, 0x68, 0x16 }
}, // 79

{
	0x10105004, 0x00000000,
	{0x4D, 0x1B, 0x1C, 0xD2, 0x97, 0x14, 0x7A, 0xD9, 0xC8, 0x4A, 0x59, 0xBB, 0xAF, 0xB6, 0xA8, 0x35 }
}, // 80

{
	0x2020A008, 0x00000000,
	{0x59, 0xB4, 0x08, 0x34, 0x13, 0x66, 0x9B, 0x5A, 0xBE, 0x52, 0x78, 0x92, 0xDC, 0x28, 0x8D, 0x92 }
}, // 81

{
	0x40414010, 0x00000000,
	{0xCF, 0xE9, 0xFA, 0x5D, 0x6D, 0xF2, 0xEB, 0x66, 0xD4, 0x7B, 0xEB, 0x94, 0xC9, 0x22, 0xEC, 0xC1 }
}, // 82

{
	0x80828020, 0x00000000,
	{0x86, 0x36, 0x60, 0x2C, 0x89, 0xCF, 0x84, 0x30, 0xB7, 0x68, 0xA6, 0x1B, 0xB4, 0x4D, 0x56, 0xA9 }
}, // 83

{
	0x01050041, 0x00000001,
	{0x4A, 0xB1, 0x9D, 0x12, 0x21, 0xEA, 0xF3, 0x8B, 0x38, 0x92, 0x29, 0x5A, 0x58, 0xA0, 0xBF, 0xE4 }
}, // 84

{
	0x020A0082, 0x00000002,
	{0x9C, 0xEF, 0x75, 0x5D, 0xBB, 0x2E, 0x10, 0x61, 0x69, 0x1C, 0x2D, 0xD2, 0x8A, 0xE0, 0xE0, 0x72 }
}, // 85

{
	0x04140104, 0x00000004,
	{0x9C, 0x8B, 0x06, 0xBB, 0x0A, 0xFD, 0x2B, 0x1B, 0xCD, 0xA0, 0x52, 0x23, 0x28, 0x69, 0x11, 0xEB }
}, // 86

{
	0x08280208, 0x00000008,
	{0x79, 0xFB, 0x15, 0xF5, 0xA5, 0x5A, 0xF0, 0xFA, 0x85, 0xBD, 0x10, 0xBC, 0x98, 0x3E, 0xBF, 0x93 }
}, // 87

{
	0x10500410, 0x00000010,
	{0x8F, 0x65, 0x93, 0x7E, 0x1C, 0x85, 0x54, 0x7B, 0x86, 0xB6, 0x5A, 0x46, 0xC7, 0x09, 0x56, 0xAB }
}, // 88

{
	0x20A00820, 0x00000020,
	{0xA6, 0x40, 0x07, 0x0A, 0x4B, 0xA2, 0x24, 0xB2, 0x39, 0x7A, 0xF6, 0xF3, 0xBC, 0x71, 0xDB, 0xB6 }
}, // 89

{
	0x41401040, 0x00000040,
	{0x0E, 0xAD, 0xCE, 0x00, 0xCB, 0x42, 0x37, 0xB8, 0xC6, 0x4D, 0x78, 0x22, 0x98, 0xB0, 0x71, 0x2B }
}, // 90

{
	0x82802081, 0x00000080,
	{0x95, 0x1C, 0xA2, 0x4C, 0x09, 0xD5, 0x7B, 0xE2, 0xED, 0x4F, 0x1E, 0x80, 0xA5, 0xFA, 0xF4, 0x14 }
}, // 91

{
	0x05004102, 0x00000101,
	{0xD7, 0x6E, 0x36, 0x28, 0x58, 0x4A, 0xDF, 0x77, 0xAA, 0x03, 0x44, 0x91, 0xB0, 0xFB, 0xD5, 0x92 }
}, // 92

{
	0x0A008204, 0x00000202,
	{0xB4, 0x65, 0xE9, 0xA0, 0xF6, 0xB2, 0xA2, 0x40, 0x47, 0xE9, 0x76, 0xAF, 0x8C, 0x20, 0x0B, 0xB9 }
}, // 93

{
	0x14010408, 0x00000404,
	{0xA3, 0x11, 0xFC, 0xD3, 0xFB, 0xF8, 0x83, 0x6E, 0xB1, 0x5A, 0x44, 0x55, 0x0C, 0xB8, 0x69, 0x7A }
}, // 94

{
	0x28020810, 0x00000808,
	{0xA1, 0x5D, 0xFA, 0x46, 0xC5, 0x01, 0xD2, 0x78, 0x1E, 0xAE, 0xC2, 0x99, 0xEA, 0xEB, 0xEF, 0x16 }
}, // 95

{
	0x50041020, 0x00001010,
	{0xCD, 0xF4, 0x87, 0x9D, 0x17, 0x84, 0x4A, 0x74, 0xCB, 0x7A, 0xD5, 0xFA, 0xD8, 0xC0, 0xEB, 0x9D }
}, // 96

{
	0xA0082041, 0x00002020,
	{0x4D, 0xD7, 0xCA, 0x76, 0x58, 0xC9, 0x8F, 0xC6, 0xA9, 0xB8, 0xA9, 0x79, 0xF8, 0xA4, 0xA4, 0x7C }
}, // 97

{
	0x40104083, 0x00004041,
	{0x0B, 0xBF, 0x19, 0x05, 0xC3, 0xDF, 0xEA, 0x3B, 0xE6, 0x39, 0x8A, 0x90, 0xE0, 0xD3, 0xF5, 0xB0 }
}, // 98

{
	0x80208106, 0x00008082,
	{0x21, 0x11, 0x96, 0x09, 0x14, 0x56, 0xE9, 0xBA, 0x01, 0x05, 0xAE, 0xD8, 0xA8, 0x8A, 0x42, 0xC7 }
}, // 99

{
	0x0041020C, 0x00010105,
	{0xBA, 0x95, 0x8D, 0x94, 0xA4, 0x20, 0x3B, 0x17, 0x92, 0xA5, 0x45, 0x3D, 0x34, 0xF9, 0x19, 0x5B }
}, // 100

{
	0x00820419, 0x0002020A,
	{0x5D, 0x4E, 0xCA, 0xCB, 0x6B, 0x94, 0x93, 0x13, 0x82, 0x89, 0xE9, 0x54, 0x15, 0xFD, 0x03, 0x9D }
}, // 101

{
	0x01040833, 0x00040414,
	{0xDA, 0x9A, 0x77, 0x99, 0x81, 0x2A, 0x6F, 0xAD, 0x0E, 0xA0, 0xDA, 0xBD, 0x55, 0xC0, 0x09, 0x1B }
}, // 102

{
	0x02081067, 0x00080828,
	{0x12, 0xC6, 0x2B, 0xF6, 0xC2, 0x4D, 0x48, 0x66, 0xE5, 0x91, 0xDF, 0xE4, 0x3A, 0x64, 0x93, 0xDB }
}, // 103

{
	0x041020CE, 0x00101050,
	{0x6D, 0x15, 0x86, 0x85, 0x75, 0xD7, 0x1D, 0x7F, 0x77, 0x13, 0x21, 0x0C, 0x61, 0xE1, 0x00, 0xDC }
}, // 104

{
	0x0820419D, 0x002020A0,
	{0xBF, 0x13, 0x73, 0x35, 0xB0, 0x4E, 0x82, 0xAE, 0x77, 0xA9, 0xF7, 0xBB, 0x00, 0x4F, 0xF0, 0x7C }
}, // 105

{
	0x1040833B, 0x00404140,
	{0x39, 0xCB, 0x08, 0xBE, 0x33, 0xE6, 0x5B, 0x1D, 0xDE, 0x38, 0xEB, 0xAB, 0x88, 0x35, 0xC6, 0x82 }
}, // 106

{
	0x20810676, 0x00808280,
	{0x86, 0x2E, 0x5A, 0x1B, 0x8A, 0x24, 0x12, 0x4A, 0x6A, 0x8C, 0x1A, 0xCF, 0x19, 0x22, 0xA4, 0x9C }
}, // 107

{
	0x41020CED, 0x01010500,
	{0x8C, 0x58, 0x04, 0x7D, 0x2A, 0x39, 0x5A, 0x4A, 0x09, 0x4B, 0x5B, 0x44, 0x9C, 0x64, 0xB4, 0x18 }
}, // 108

{
	0x820419DB, 0x02020A00,
	{0x71, 0xF3, 0x1D, 0xC3, 0xB4, 0x06, 0x7E, 0x83, 0x1A, 0x23, 0x35, 0xA7, 0xC3, 0x09, 0xE7, 0x58 }
}, // 109

{
	0x040833B6, 0x04041401,
	{0x1B, 0x2F, 0x60, 0x2A, 0x43, 0x59, 0x77, 0x1F, 0x55, 0x18, 0x77, 0x61, 0xCC, 0x83, 0x60, 0xAB }
}, // 110

{
	0x0810676C, 0x08082802,
	{0x47, 0xD0, 0x07, 0x7C, 0x09, 0x9B, 0xA2, 0x04, 0x30, 0x45, 0x51, 0x6D, 0x9F, 0x89, 0x7E, 0x56 }
}, // 111

{
	0x00000000, 0x80000000,
	{0x37, 0x2D, 0xD8, 0x7F, 0x71, 0xBA, 0x1B, 0xC9, 0xC8, 0x40, 0xC8, 0xFA, 0xA2, 0x48, 0x97, 0xFC }
}, // 112

{
	0x00000010, 0x80000000,
	{0x15, 0xFC, 0x30, 0xE8, 0x8E, 0xC9, 0xEB, 0xE0, 0x17, 0x5D, 0x1C, 0x05, 0x6A, 0xA4, 0xAA, 0x04 }
}, // 113

{
	0x00000020, 0x80000000,
	{0x0F, 0x42, 0xC0, 0x42, 0x18, 0xB0, 0x37, 0x42, 0x11, 0xC4, 0x9B, 0xE3, 0x20, 0x1C, 0x1E, 0x19 }
}, // 114

{
	0x00000030, 0x80000000,
	{0x79, 0x4C, 0xF7, 0x29, 0xD8, 0x29, 0x03, 0x02, 0x03, 0x2A, 0x9D, 0xFF, 0x1A, 0xE8, 0x38, 0xA0 }
}, // 115

{
	0x00000040, 0x80000000,
	{0x19, 0x45, 0x6F, 0x72, 0x9E, 0x95, 0xAD, 0x6B, 0xB4, 0xDA, 0x01, 0xFA, 0x78, 0xEA, 0x4C, 0x3E }
}, // 116

{
	0x00004040, 0x00000000,
	{0x3D, 0xB3, 0xE0, 0x5C, 0xC7, 0x05, 0x76, 0x6B, 0x45, 0xB4, 0x5F, 0x82, 0x36, 0x24, 0x02, 0xB1 }
}, // 117

{
	0x00008081, 0x00000000,
	{0x20, 0x33, 0x9F, 0x3E, 0x97, 0x20, 0xCE, 0x0A, 0xD7, 0x69, 0x29, 0xB2, 0xAE, 0xFE, 0x79, 0x8D }
}, // 118

{
	0x00010103, 0x00000000,
	{0x7C, 0xC9, 0xFC, 0x63, 0x9D, 0x2F, 0x58, 0xA5, 0xBC, 0xF3, 0x76, 0xCD, 0xB0, 0xBD, 0xF7, 0x1A }
}, // 119

{
	0x00020206, 0x00000000,
	{0x46, 0xCF, 0xEB, 0xF9, 0x48, 0x90, 0xDC, 0x1F, 0x34, 0x0F, 0xA8, 0x79, 0x5B, 0x0E, 0xAA, 0xB8 }
}, // 120

{
	0x0004040C, 0x00000000,
	{0x29, 0x89, 0x20, 0xB6, 0xC1, 0xBE, 0x2E, 0x1E, 0x70, 0x77, 0xF7, 0xDE, 0x64, 0xDC, 0x8F, 0x14 }
}, // 121

{
	0x00080818, 0x00000000,
	{0x42, 0x77, 0x79, 0x02, 0xC0, 0x56, 0xF0, 0x62, 0x29, 0xF6, 0x14, 0x71, 0xC1, 0xF4, 0x61, 0xE3 }
}, // 122

{
	0x00101030, 0x00000000,
	{0x1C, 0xD3, 0x50, 0xF6, 0xA0, 0xAE, 0x69, 0x50, 0xFB, 0xA4, 0x36, 0x5B, 0x55, 0x32, 0xEF, 0x6D }
}, // 123

{
	0x00202060, 0x00000000,
	{0x08, 0xA2, 0xAC, 0x6D, 0x5E, 0xD4, 0x21, 0x06, 0x80, 0xE6, 0x99, 0xE3, 0xC2, 0x6A, 0x3C, 0x1C }
}, // 124

{
	0x004040C0, 0x00000000,
	{0x15, 0xF9, 0xD8, 0xE1, 0x5C, 0x40, 0xDD, 0xE7, 0x55, 0x67, 0x49, 0xAD, 0xDB, 0x63, 0x71, 0x4D }
}, // 125

{
	0x00808180, 0x00000000,
	{0x90, 0x95, 0xF0, 0x3E, 0x16, 0xC9, 0xAD, 0x1A, 0xA0, 0xCA, 0x74, 0x33, 0x6D, 0x2D, 0x21, 0x52 }
}, // 126

{
	0x01010300, 0x00000000,
	{0x78, 0x2B, 0x36, 0x85, 0xDA, 0x7E, 0xB7, 0xE0, 0xC0, 0x73, 0xF7, 0x53, 0x54, 0x98, 0x4D, 0x9B }
}, // 127

{
	0x02020600, 0x00000000,
	{0x24, 0xF0, 0xFB, 0xCE, 0xAF, 0xBB, 0x32, 0x39, 0x12, 0xBB, 0xAA, 0xCC, 0x6A, 0x3C, 0x4E, 0xFC }
}, // 128

{
	0x04040C00, 0x00000000,
	{0xF7, 0x9B, 0x5A, 0xE2, 0xE4, 0x32, 0x52, 0xE5, 0x27, 0x26, 0x0B, 0x68, 0x42, 0x76, 0x22, 0x60 }
}, // 129

{
	0x08081801, 0x00000000,
	{0x95, 0x1F, 0xAC, 0x48, 0xF7, 0x81, 0xEB, 0x10, 0x60, 0x9F, 0xC7, 0x0D, 0xCF, 0xFD, 0xDB, 0x5A }
}, // 130

{
	0x10103002, 0x00000000,
	{0x55, 0xED, 0x1F, 0xB0, 0x6C, 0xD4, 0x9C, 0x97, 0x40, 0x53, 0x09, 0x8E, 0x80, 0xBC, 0x2A, 0xF9 }
}, // 131

{
	0x20206004, 0x00000000,
	{0x2E, 0x0B, 0xD9, 0xD7, 0x9D, 0x94, 0xAC, 0x9A, 0xB6, 0xC1, 0xC1, 0x36, 0x37, 0x62, 0xB9, 0xF5 }
}, // 132

{
	0x4040C008, 0x00000000,
	{0x89, 0x36, 0x39, 0xE8, 0x87, 0xFA, 0xB8, 0x0C, 0x26, 0xAD, 0x37, 0x78, 0x03, 0x89, 0x0A, 0x03 }
}, // 133

{
	0x80818010, 0x00000000,
	{0x62, 0x37, 0x2C, 0x0B, 0xD1, 0x97, 0x40, 0x62, 0xE8, 0x13, 0x45, 0x22, 0xDC, 0x2A, 0x79, 0x53 }
}, // 134

{
	0x01030020, 0x00000001,
	{0x5E, 0x23, 0x97, 0xEB, 0xE1, 0x08, 0x78, 0x7B, 0x81, 0x13, 0x6E, 0x72, 0x19, 0xB1, 0x96, 0xEA }
}, // 135

{
	0x02060041, 0x00000002,
	{0x0A, 0x9E, 0x4B, 0x56, 0x28, 0xB3, 0x00, 0x68, 0xCC, 0xEB, 0xC2, 0x8A, 0x13, 0x6A, 0x7B, 0x42 }
}, // 136

{
	0x040C0082, 0x00000004,
	{0xB4, 0xF6, 0x12, 0x5A, 0x8A, 0x6F, 0x99, 0xF3, 0x6A, 0xF8, 0xE0, 0xEC, 0x49, 0x36, 0xCE, 0xA9 }
}, // 137

{
	0x08180104, 0x00000008,
	{0x40, 0x5D, 0xB9, 0xB2, 0xAE, 0xC1, 0xB2, 0x44, 0x7C, 0xC7, 0xE1, 0xCD, 0x88, 0xCB, 0x34, 0x79 }
}, // 138

{
	0x10300208, 0x00000010,
	{0xAA, 0x13, 0x21, 0xD7, 0x85, 0x22, 0xD8, 0xC6, 0x3C, 0x8B, 0x1E, 0x03, 0x80, 0x47, 0xD7, 0x6C }
}, // 139

{
	0x20600410, 0x00000020,
	{0x35, 0x88, 0xC3, 0x64, 0x7A, 0xA5, 0x19, 0x32, 0x0E, 0xD8, 0xC0, 0x70, 0xCC, 0x62, 0x3D, 0x41 }
}, // 140

{
	0x40C00820, 0x00000040,
	{0xD4, 0x54, 0x5A, 0xE8, 0x84, 0x8F, 0x5F, 0x3F, 0xA5, 0x4D, 0xB6, 0x22, 0x82, 0x24, 0x5A, 0xB4 }
}, // 141

{
	0x81801041, 0x00000080,
	{0xD3, 0x27, 0x71, 0x3A, 0x72, 0xC8, 0x6D, 0x3D, 0x6D, 0xC6, 0x93, 0xD8, 0x70, 0x0E, 0xE2, 0xEC }
}, // 142

{
	0x03002082, 0x00000101,
	{0xE7, 0x5A, 0x26, 0xE5, 0x0F, 0x61, 0x00, 0x46, 0xE5, 0xEB, 0x81, 0x6A, 0xE1, 0x28, 0x4C, 0x67 }
}, // 143

{
	0x06004104, 0x00000202,
	{0xE9, 0xCE, 0x7F, 0xC8, 0x13, 0x73, 0x22, 0xD1, 0x81, 0xAB, 0xC5, 0x0A, 0x61, 0x5B, 0x3E, 0x1D }
}, // 144

{
	0x0C008208, 0x00000404,
	{0xD4, 0xCA, 0x58, 0x9D, 0x75, 0xC8, 0xE8, 0xAA, 0x9D, 0x8F, 0x11, 0xBD, 0x53, 0x6E, 0x1C, 0x81 }
}, // 145

{
	0x18010411, 0x00000808,
	{0x69, 0x82, 0xD4, 0xF3, 0x07, 0x61, 0x74, 0x7E, 0xE2, 0x90, 0xC4, 0x5A, 0x9C, 0x7D, 0x60, 0x32 }
}, // 146

{
	0x30020823, 0x00001010,
	{0xBB, 0x40, 0x82, 0x82, 0xA8, 0xB8, 0x92, 0xA1, 0xEB, 0xAB, 0x6C, 0x0C, 0x5A, 0xB2, 0x16, 0x27 }
}, // 147

{
	0x60041047, 0x00002020,
	{0xF6, 0x05, 0x0F, 0x92, 0xDE, 0x52, 0x81, 0x06, 0xDC, 0xC9, 0x18, 0xD2, 0xD0, 0x60, 0x62, 0x36 }
}, // 148

{
	0xC008208E, 0x00004040,
	{0x11, 0x71, 0xDB, 0xB8, 0x54, 0x5F, 0xE8, 0x4E, 0xE3, 0x3C, 0xDF, 0xF1, 0xD6, 0x5B, 0x74, 0x56 }
}, // 149

{
	0x8010411C, 0x00008081,
	{0xA7, 0xCC, 0x67, 0x0A, 0xD5, 0xFB, 0xF3, 0x1A, 0x8C, 0xC5, 0xE6, 0xA6, 0x07, 0x8A, 0x59, 0x49 }
}, // 150

{
	0x00208238, 0x00010103,
	{0x57, 0xDF, 0x9A, 0xB8, 0x91, 0x70, 0xD0, 0xA0, 0xB2, 0xF7, 0x71, 0xE0, 0x3E, 0x03, 0xD0, 0xFF }
}, // 151

{
	0x00410471, 0x00020206,
	{0xF9, 0xD0, 0x9A, 0xB8, 0x45, 0xB1, 0xE3, 0x10, 0x5A, 0x01, 0x59, 0x58, 0x4D, 0x25, 0x1E, 0x6E }
}, // 152

{
	0x008208E2, 0x0004040C,
	{0x7D, 0x11, 0xB2, 0x36, 0x78, 0xEB, 0x97, 0xA5, 0xAC, 0xA2, 0x07, 0x17, 0x6E, 0xAD, 0xEC, 0xCD }
}, // 153

{
	0x010411C4, 0x00080818,
	{0x47, 0x39, 0x98, 0x82, 0x08, 0x65, 0x76, 0xD9, 0xE8, 0x95, 0x4E, 0xEE, 0xB4, 0xE7, 0xDE, 0x72 }
}, // 154

{
	0x02082388, 0x00101030,
	{0xFC, 0x7E, 0x13, 0xBB, 0xF6, 0xAE, 0x1A, 0x25, 0x48, 0x9C, 0xF3, 0x6E, 0xCC, 0x66, 0x54, 0x54 }
}, // 155

{
	0x04104710, 0x00202060,
	{0xEC, 0xE0, 0x9D, 0xA8, 0xCD, 0x6B, 0x55, 0xDC, 0xA7, 0xC4, 0x30, 0xDE, 0xF4, 0xD8, 0x8A, 0x61 }
}, // 156

{
	0x08208E21, 0x004040C0,
	{0xE8, 0x43, 0x94, 0xED, 0xB7, 0xBD, 0x7C, 0x31, 0x17, 0xA1, 0xE0, 0x0E, 0x57, 0x50, 0xF5, 0xCC }
}, // 157

{
	0x10411C43, 0x00808180,
	{0x3B, 0x64, 0xC0, 0xD0, 0x04, 0x4A, 0xB0, 0xA8, 0x28, 0xAE, 0x3C, 0xC3, 0xC3, 0x3F, 0x15, 0xBE }
}, // 158

{
	0x20823887, 0x01010300,
	{0xDF, 0x6C, 0x94, 0xAE, 0x52, 0xF3, 0x73, 0x76, 0x90, 0x76, 0xDA, 0x27, 0xC4, 0x48, 0xF9, 0x70 }
}, // 159

{
	0x4104710F, 0x02020600,
	{0x89, 0x3F, 0xFD, 0x02, 0xD5, 0xDC, 0xA4, 0x70, 0x38, 0x09, 0xDA, 0xEB, 0x83, 0x28, 0x3C, 0x91 }
}, // 160

{
	0x8208E21F, 0x04040C00,
	{0xE4, 0xF6, 0x73, 0x10, 0xB8, 0x10, 0xA8, 0x5D, 0xF1, 0xE7, 0x7B, 0x77, 0x8D, 0xAC, 0x47, 0x22 }
}, // 161

{
	0x0411C43E, 0x08081801,
	{0x3F, 0x1D, 0x26, 0xD5, 0x36, 0xE9, 0x87, 0x2F, 0x98, 0xE3, 0x4D, 0x34, 0x78, 0xCD, 0x56, 0x67 }
}, // 162

{
	0x00000000, 0x80000000,
	{0x37, 0x2D, 0xD8, 0x7F, 0x71, 0xBA, 0x1B, 0xC9, 0xC8, 0x40, 0xC8, 0xFA, 0xA2, 0x48, 0x97, 0xFC }
},// 163

{
	0x00000000, 0x80000000,
	{0x37, 0x2D, 0xD8, 0x7F, 0x71, 0xBA, 0x1B, 0xC9, 0xC8, 0x40, 0xC8, 0xFA, 0xA2, 0x48, 0x97, 0xFC }
},// 164

{
	0x04040C01, 0x00000000,
	{0xB3, 0xF0, 0xE3, 0x57, 0x64, 0xB6, 0xD9, 0x86, 0x1D, 0x26, 0x33, 0x7D, 0xEB, 0x8A, 0xC0, 0x47 }
}, // 165

{
	0x08081803, 0x00000000,
	{0x87, 0x15, 0x11, 0x16, 0xDB, 0xBA, 0xC8, 0xFC, 0x6B, 0x8D, 0x3A, 0x57, 0xCF, 0x97, 0x68, 0x7F }
}, // 166

{
	0x10103006, 0x00000000,
	{0x92, 0xF9, 0xC1, 0x0C, 0xF5, 0x58, 0xD9, 0xA5, 0xCE, 0xC2, 0xD6, 0x6F, 0x62, 0x69, 0xF9, 0xE2 }
}, // 167

{
	0x2020600C, 0x00000000,
	{0x28, 0xB5, 0x2D, 0x97, 0xF2, 0xF1, 0x9D, 0x08, 0xFD, 0x1C, 0x2E, 0xC8, 0xF5, 0x9A, 0xAE, 0xA0 }
}, // 168

{
	0x4040C018, 0x00000000,
	{0x7A, 0x21, 0xF2, 0xB4, 0x44, 0x96, 0x3A, 0xE1, 0x9C, 0x8B, 0x96, 0x41, 0xE5, 0xC5, 0x35, 0x93 }
}, // 169

{
	0x80818030, 0x00000000,
	{0x99, 0xF4, 0x75, 0x24, 0x94, 0x61, 0x01, 0xCE, 0x15, 0x84, 0x76, 0x32, 0x14, 0xA4, 0x16, 0x67 }
}, // 170

{
	0x01030061, 0x00000001,
	{0x2E, 0x5C, 0x60, 0x0A, 0xF4, 0xC7, 0x64, 0x71, 0xC5, 0x53, 0x21, 0x01, 0xB9, 0x31, 0x38, 0x00 }
}, // 171

{
	0x020600C3, 0x00000002,
	{0x0E, 0x39, 0xFD, 0x1A, 0xAB, 0x39, 0xC7, 0xC9, 0x45, 0x09, 0x09, 0x00, 0x4D, 0xD2, 0x33, 0x7C }
}, // 172

{
	0x040C0186, 0x00000004,
	{0xC8, 0x5A, 0x27, 0x9F, 0x5A, 0x18, 0xCE, 0xCA, 0xE3, 0x3D, 0xC0, 0x24, 0x04, 0xDD, 0x4A, 0x6A }
}, // 173

{
	0x0818030C, 0x00000008,
	{0x7B, 0xBB, 0xD8, 0xB5, 0x15, 0xB6, 0xAC, 0x2F, 0xCC, 0x9A, 0x03, 0xEB, 0x5C, 0xD6, 0x9C, 0xE3 }
}, // 174

{
	0x10300618, 0x00000010,
	{0xF0, 0x1B, 0xF5, 0x6C, 0x47, 0x71, 0x02, 0xB5, 0x7E, 0x66, 0xA8, 0x7A, 0x64, 0x16, 0xAB, 0x21 }
}, // 175

{
	0x20600C30, 0x00000020,
	{0x9A, 0xB3, 0xD1, 0xAA, 0xAE, 0x49, 0x65, 0x6B, 0xD9, 0x9B, 0xDC, 0x10, 0xF3, 0x8D, 0x86, 0x3C }
}, // 176

{
	0x40C01860, 0x00000040,
	{0x02, 0x6F, 0xB2, 0x90, 0xA1, 0x56, 0x43, 0xA1, 0x21, 0x06, 0xE3, 0x3E, 0x13, 0x87, 0x56, 0x62 }
}, // 177

{
	0x818030C0, 0x00000080,
	{0xFB, 0x73, 0x8A, 0xCF, 0x4F, 0x96, 0xCC, 0x6B, 0x19, 0x09, 0xA7, 0xB2, 0x19, 0xF0, 0x0C, 0xED }
}, // 178

{
	0x03006180, 0x00000101,
	{0x9D, 0x9C, 0x28, 0xC2, 0x90, 0xEB, 0x87, 0x9B, 0xA2, 0x45, 0x4F, 0xF2, 0x93, 0x0B, 0xA4, 0x73 }
}, // 179

{
	0x0600C300, 0x00000202,
	{0x22, 0x86, 0x29, 0xF3, 0x76, 0x99, 0xBD, 0x2B, 0x90, 0xC3, 0xE4, 0x16, 0x8C, 0x8A, 0x21, 0x09 }
}, // 180

{
	0x0C018600, 0x00000404,
	{0xB6, 0xEC, 0xE1, 0x82, 0x6B, 0xFB, 0x71, 0x87, 0x66, 0x28, 0xB8, 0xFC, 0x21, 0x09, 0x3E, 0x9F }
}, // 181

{
	0x18030C01, 0x00000808,
	{0x4C, 0x5D, 0x9C, 0xAA, 0xFB, 0xE4, 0x22, 0x0A, 0xEB, 0x7A, 0x69, 0xD2, 0x46, 0x3F, 0xB3, 0x83 }
}, // 182

{
	0x30061803, 0x00001010,
	{0x24, 0xB2, 0xB3, 0xE2, 0x2D, 0xCB, 0x91, 0x05, 0x7D, 0x18, 0xCB, 0xED, 0x1C, 0x7D, 0x02, 0xE0 }
}, // 183

{
	0x600C3007, 0x00002020,
	{0x74, 0x53, 0xF2, 0xDC, 0xE7, 0xE7, 0xC4, 0x56, 0xDF, 0x1D, 0xB5, 0xD7, 0x4E, 0x2A, 0x30, 0x8C }
}, // 184

{
	0xC018600E, 0x00004040,
	{0xCC, 0x99, 0xA4, 0xB8, 0x97, 0x6C, 0xF2, 0x97, 0x1E, 0xB6, 0x29, 0xE9, 0x8F, 0x06, 0x29, 0x28 }
}, // 185

{
	0x8030C01C, 0x00008081,
	{0x9F, 0xED, 0xB7, 0xF9, 0x54, 0x9C, 0xD2, 0x7E, 0x90, 0xEF, 0x39, 0x06, 0x5B, 0xAD, 0xE1, 0xD6 }
}, // 186

{
	0x00618038, 0x00010103,
	{0x7A, 0xAF, 0x9F, 0xA1, 0x60, 0xA8, 0x46, 0xCD, 0x26, 0x82, 0xC9, 0xE0, 0xBB, 0xAC, 0xEB, 0x8A }
}, // 187

{
	0x00C30070, 0x00020206,
	{0x38, 0x8E, 0x03, 0xA9, 0xEC, 0xE9, 0x3B, 0x40, 0x7F, 0xDD, 0x61, 0x83, 0x3D, 0xA2, 0x84, 0x57 }
}, // 188

{
	0x018600E0, 0x0004040C,
	{0xB2, 0x02, 0x09, 0xF7, 0x65, 0x7D, 0xAA, 0x09, 0xFE, 0xDB, 0xEF, 0x59, 0x5E, 0xA8, 0xA9, 0x0B }
}, // 189

{
	0x030C01C1, 0x00080818,
	{0x48, 0xCA, 0x9E, 0x5A, 0xB6, 0x80, 0x7E, 0xCF, 0x7E, 0x2E, 0x39, 0xCD, 0x5E, 0xDB, 0x94, 0xC7 }
}, // 190

{
	0x06180382, 0x00101030,
	{0xDE, 0x98, 0xB5, 0x0E, 0xFF, 0x7D, 0x81, 0x4C, 0xF9, 0xD9, 0x64, 0x94, 0x9E, 0xAA, 0x7F, 0xE5 }
}, // 191

{
	0x0C300705, 0x00202060,
	{0x5E, 0x66, 0xD8, 0x6C, 0x67, 0x70, 0x77, 0x20, 0x02, 0xB0, 0xFE, 0xEE, 0xE6, 0xE5, 0xA8, 0xA1 }
}, // 192

{
	0x18600E0A, 0x004040C0,
	{0x76, 0x01, 0x20, 0x3F, 0x95, 0xE1, 0x3B, 0x4D, 0x02, 0x46, 0xE2, 0xC7, 0xA7, 0xB9, 0x5B, 0x9B }
}, // 193

{
	0x30C01C15, 0x00808180,
	{0xF1, 0xFA, 0x8B, 0xCA, 0x6D, 0x09, 0x2B, 0x2B, 0x8D, 0xA5, 0x39, 0x9D, 0x08, 0xF9, 0xA8, 0x0A }
}, // 194

{
	0x6180382B, 0x01010300,
	{0x13, 0x07, 0x91, 0x71, 0x70, 0xF2, 0xEE, 0x5A, 0x40, 0xEF, 0xF5, 0x85, 0x61, 0x4E, 0xCD, 0x8F }
}, // 195

{
	0xC3007056, 0x02020600,
	{0x61, 0x26, 0xC7, 0xBB, 0x02, 0xEE, 0xB4, 0xF9, 0xC0, 0x93, 0xB5, 0x82, 0xD3, 0x48, 0xD5, 0x42 }
}, // 196

{
	0x8600E0AD, 0x04040C01,
	{0x90, 0x57, 0xD7, 0xC5, 0x06, 0x2F, 0x28, 0x58, 0xA9, 0x46, 0x0C, 0x39, 0x8B, 0x54, 0xA6, 0x40 }
}, // 197

{
	0x0C01C15A, 0x08081803,
	{0x4A, 0xA8, 0x99, 0x98, 0x83, 0x76, 0x64, 0x3F, 0xF1, 0x42, 0x32, 0x36, 0x18, 0xDC, 0xCE, 0x5F }
}, // 198

{
	0x00000001, 0x80000000,
	{0xC4, 0x1E, 0xAA, 0xDB, 0x22, 0x92, 0xF3, 0x16, 0x95, 0xE6, 0xBC, 0x3B, 0x11, 0x1F, 0xC6, 0x24 }
}, // 199

{
	0x00000011, 0x80000000,
	{0x92, 0x19, 0xA4, 0x85, 0xE9, 0x8B, 0x61, 0x1C, 0xD1, 0xA9, 0x14, 0x95, 0x75, 0xF2, 0x15, 0x4B }
}, // 200

{
	0x00000021, 0x80000000,
	{0x99, 0x05, 0x7D, 0x00, 0xBE, 0xFD, 0x9E, 0x42, 0xE0, 0xC2, 0x1B, 0x94, 0xF3, 0x27, 0x42, 0xC8 }
}, // 201

{
	0x00000031, 0x80000000,
	{0xB2, 0x50, 0x95, 0x4D, 0x1C, 0xCC, 0x4B, 0x67, 0x0D, 0xD3, 0x6A, 0xEB, 0xC2, 0x1D, 0x48, 0x3B }
}, // 202

{
	0x00000041, 0x80000000,
	{0x02, 0xBD, 0x3D, 0xBC, 0x07, 0x23, 0x7A, 0x96, 0x87, 0x0A, 0x0F, 0x92, 0xB5, 0x0D, 0x7C, 0xAC }
}, // 203

{
	0x00010104, 0x00000000,
	{0xF8, 0xD5, 0x79, 0x30, 0xF4, 0xDC, 0xE6, 0x67, 0xDE, 0x30, 0xCB, 0x04, 0xB5, 0x1A, 0x26, 0x83 }
}, // 204

{
	0x00020208, 0x00000000,
	{0xDB, 0x1F, 0xB8, 0xED, 0xFE, 0xB9, 0x37, 0x3E, 0xFA, 0x25, 0x36, 0x96, 0x91, 0x79, 0xEF, 0xFF }
}, // 205

{
	0x00040410, 0x00000000,
	{0x33, 0x6F, 0xF8, 0x7D, 0x19, 0x10, 0x45, 0x4C, 0x65, 0x1C, 0x42, 0xDC, 0xB8, 0x08, 0x0E, 0xC4 }
}, // 206

{
	0x00080820, 0x00000000,
	{0x1A, 0x5B, 0xF3, 0xFB, 0x56, 0xBD, 0xF1, 0xF9, 0x7C, 0x9E, 0x47, 0xAB, 0x6F, 0x51, 0xF5, 0x9F }
}, // 207

{
	0x00101040, 0x00000000,
	{0xCD, 0x18, 0xA1, 0xD7, 0xFC, 0x08, 0xE8, 0x84, 0x95, 0x64, 0xDE, 0x9B, 0x77, 0xD8, 0xEE, 0x4E }
}, // 208

{
	0x00202080, 0x00000000,
	{0x2C, 0x6D, 0x7B, 0xAB, 0x33, 0xA5, 0x86, 0x13, 0xFD, 0x09, 0x27, 0x36, 0x65, 0xE4, 0x3A, 0x40 }
}, // 209

{
	0x00404100, 0x00000000,
	{0xE1, 0x1F, 0xEC, 0xE5, 0xA3, 0x7C, 0xF6, 0x40, 0xE2, 0x73, 0x1E, 0x65, 0x5C, 0x8F, 0x5B, 0x14 }
}, // 210

{
	0x00808200, 0x00000000,
	{0x14, 0x2A, 0x78, 0x81, 0x49, 0x55, 0x90, 0x10, 0x79, 0xEE, 0x35, 0xA8, 0x98, 0x07, 0xBB, 0xFE }
}, // 211

{
	0x01010400, 0x00000000,
	{0x6F, 0xFF, 0x23, 0x98, 0x8D, 0x66, 0x1A, 0x9F, 0x35, 0x06, 0xC1, 0x98, 0x33, 0x91, 0x6A, 0x8C }
}, // 212

{
	0x02020800, 0x00000000,
	{0x35, 0xCC, 0xC8, 0x78, 0x33, 0x48, 0x7F, 0x6C, 0x79, 0x26, 0xD1, 0x43, 0x19, 0xFE, 0x71, 0x21 }
}, // 213

{
	0x04041000, 0x00000000,
	{0x47, 0xEA, 0xC2, 0x31, 0xE3, 0x41, 0xEE, 0x79, 0xFB, 0x91, 0xB8, 0x05, 0x5F, 0x5F, 0x42, 0x73 }
}, // 214

{
	0x08082001, 0x00000000,
	{0xD6, 0x18, 0x65, 0xA6, 0x70, 0x1F, 0xF8, 0x47, 0xEE, 0xF7, 0x40, 0x54, 0xD4, 0xE0, 0x16, 0xBA }
}, // 215

{
	0x10104002, 0x00000000,
	{0x1E, 0x81, 0x1E, 0x10, 0x5A, 0xB0, 0xF1, 0x1C, 0x3F, 0xE3, 0xCD, 0x7B, 0x1B, 0x0B, 0xFC, 0x7E }
}, // 216

{
	0x20208004, 0x00000000,
	{0xC7, 0xD3, 0x91, 0xD1, 0x65, 0xBF, 0xA6, 0x2F, 0x91, 0x06, 0x27, 0x73, 0xE3, 0x98, 0x28, 0xDB }
}, // 217

{
	0x40410008, 0x00000000,
	{0x42, 0xFE, 0x10, 0x2D, 0xE0, 0x61, 0x75, 0xF4, 0xEA, 0x4B, 0x1D, 0x79, 0xB5, 0x81, 0x2F, 0x15 }
}, // 218

{
	0x80820010, 0x00000000,
	{0xDA, 0x7B, 0x31, 0x8D, 0xC9, 0x7F, 0x53, 0x0D, 0x6C, 0xDC, 0xD8, 0xD2, 0x0E, 0x0E, 0x9A, 0x4B }
}, // 219

{
	0x01040021, 0x00000001,
	{0xC3, 0x27, 0x8C, 0x96, 0x1B, 0xB0, 0x20, 0x32, 0x71, 0x56, 0x87, 0xD0, 0x3D, 0x0D, 0x48, 0x15 }
}, // 220

{
	0x02080042, 0x00000002,
	{0x22, 0xB0, 0x3E, 0xBF, 0x03, 0x94, 0x19, 0x95, 0x8C, 0x3E, 0xDC, 0x78, 0x3D, 0xFB, 0xDB, 0xB3 }
}, // 221

{
	0x04100084, 0x00000004,
	{0xA7, 0x1C, 0x75, 0xCC, 0xBE, 0x2D, 0xD9, 0xFE, 0x09, 0xE3, 0x43, 0xA3, 0xAE, 0xE6, 0x9E, 0x4F }
}, // 222

{
	0x08200108, 0x00000008,
	{0xDB, 0x48, 0x82, 0x56, 0xF1, 0xFE, 0xB0, 0x3A, 0x8D, 0x33, 0x5C, 0x3D, 0x7F, 0x2E, 0xEC, 0x09 }
}, // 223

{
	0x10400210, 0x00000010,
	{0x26, 0xC9, 0xB9, 0xF0, 0x96, 0x0E, 0xCD, 0x15, 0x04, 0xD5, 0x4A, 0x32, 0x90, 0xC6, 0x08, 0x59 }
}, // 224

{
	0x20800420, 0x00000020,
	{0xA1, 0x8D, 0x0B, 0xC8, 0xF6, 0x5C, 0x3F, 0x9B, 0xBD, 0x64, 0x50, 0xF2, 0x5E, 0x94, 0x60, 0x5E }
}, // 225

{
	0x41000840, 0x00000040,
	{0xCE, 0xA8, 0xB6, 0x8C, 0x7E, 0x34, 0x0E, 0xB0, 0x2C, 0x62, 0xE9, 0xB1, 0x05, 0xFC, 0x6B, 0x10 }
}, // 226

{
	0x82001081, 0x00000080,
	{0x2B, 0xCC, 0x5A, 0x9D, 0x04, 0x0B, 0xC1, 0x80, 0x25, 0xC3, 0x70, 0xD9, 0x1B, 0x9D, 0x29, 0x58 }
}, // 227

{
	0x04002103, 0x00000101,
	{0x4C, 0xD4, 0x6C, 0x88, 0xDB, 0x4E, 0x84, 0x5E, 0x22, 0x31, 0x48, 0x0D, 0xE1, 0x38, 0xBC, 0x07 }
}, // 228

{
	0x08004206, 0x00000202,
	{0xDA, 0x1B, 0x19, 0x17, 0xAD, 0x79, 0xA4, 0xDC, 0x93, 0xBB, 0xF7, 0x2E, 0x18, 0x01, 0x85, 0x41 }
}, // 229

{
	0x1000840D, 0x00000404,
	{0x33, 0xB5, 0x3C, 0x23, 0x26, 0xA3, 0x20, 0x89, 0x50, 0x16, 0x75, 0x1D, 0xC3, 0x92, 0x26, 0xD5 }
}, // 230

{
	0x2001081B, 0x00000808,
	{0xED, 0x2F, 0x66, 0x64, 0x91, 0x5D, 0xA5, 0x44, 0x9B, 0xA9, 0x42, 0xFF, 0xAF, 0x15, 0x8C, 0x3B }
}, // 231

{
	0x40021036, 0x00001010,
	{0xF1, 0xE2, 0x96, 0x42, 0x85, 0x6C, 0x62, 0x5F, 0x5F, 0x8B, 0xD3, 0xA5, 0x28, 0x4E, 0x2F, 0xBA }
}, // 232

{
	0x8004206C, 0x00002020,
	{0x46, 0x6A, 0xAB, 0x2C, 0x3E, 0x89, 0x0F, 0x6C, 0xE3, 0xC1, 0x8B, 0xF8, 0x8E, 0x73, 0x35, 0x48 }
}, // 233

{
	0x000840D9, 0x00004041,
	{0x3B, 0x4F, 0xE2, 0x99, 0xF4, 0x33, 0x77, 0x76, 0x9C, 0x36, 0x65, 0xCF, 0x63, 0x59, 0xCC, 0xAB }
}, // 234

{
	0x001081B3, 0x00008082,
	{0x46, 0xCB, 0x56, 0x2B, 0xC1, 0x1E, 0xF2, 0xAD, 0xCE, 0x77, 0x36, 0x10, 0xD5, 0xBA, 0xEA, 0x61 }
}, // 235

{
	0x00210367, 0x00010104,
	{0x67, 0x86, 0xCD, 0xB6, 0x13, 0xDC, 0x08, 0xBD, 0xC6, 0x8F, 0xCA, 0x26, 0x78, 0xFF, 0x94, 0x1F }
}, // 236

{
	0x004206CE, 0x00020208,
	{0x76, 0xBE, 0x55, 0x7E, 0x6C, 0x4F, 0xB0, 0x29, 0x03, 0xB7, 0x53, 0xD7, 0x0B, 0xF5, 0x56, 0xD5 }
}, // 237

{
	0x00840D9D, 0x00040410,
	{0xD9, 0xEC, 0x2F, 0xB9, 0xE7, 0x3D, 0x78, 0x81, 0x42, 0xB8, 0x59, 0x4D, 0xDD, 0xE9, 0x65, 0xD2 }
}, // 238

{
	0x01081B3B, 0x00080820,
	{0xBA, 0xEC, 0x17, 0xE9, 0xC3, 0xBA, 0xD9, 0xC7, 0x11, 0xA1, 0x18, 0x7F, 0x1E, 0xC6, 0x92, 0x6B }
}, // 239

{
	0x02103677, 0x00101040,
	{0xCF, 0x79, 0x39, 0x2B, 0x29, 0x78, 0x83, 0xC7, 0x89, 0x62, 0x41, 0x8F, 0x0E, 0xB2, 0xAF, 0x99 }
}, // 240

{
	0x04206CEE, 0x00202080,
	{0xB7, 0x7D, 0xA2, 0xD0, 0x06, 0x43, 0xAF, 0xCF, 0xD9, 0x1B, 0x63, 0x17, 0x95, 0xC3, 0x5F, 0xEB }
}, // 241

{
	0x0840D9DD, 0x00404100,
	{0xC4, 0x5E, 0xEE, 0x1E, 0xED, 0x0D, 0x69, 0x81, 0x74, 0x4E, 0x3E, 0xBC, 0x1E, 0x4C, 0xD9, 0xE1 }
}, // 242

{
	0x1081B3BA, 0x00808200,
	{0x13, 0x12, 0x16, 0xF1, 0xD5, 0x8F, 0x6B, 0x65, 0x77, 0x8A, 0x62, 0xB5, 0x4E, 0xEA, 0xCD, 0x43 }
}, // 243

{
	0x21036774, 0x01010400,
	{0xED, 0xDF, 0x55, 0x0E, 0x59, 0x1B, 0x59, 0x80, 0x8D, 0x71, 0x6D, 0xD2, 0x29, 0x56, 0xB2, 0x29 }
}, // 244

{
	0x4206CEE9, 0x02020800,
	{0xA9, 0x20, 0x90, 0xF4, 0xF8, 0x8C, 0x82, 0x7D, 0x9A, 0xA9, 0xF1, 0x30, 0xE6, 0xFE, 0xA6, 0x4E }
}, // 245

{
	0x840D9DD2, 0x04041000,
	{0x04, 0x8D, 0x47, 0x92, 0xF0, 0xB9, 0x9A, 0x3D, 0xBC, 0x97, 0x1D, 0x81, 0x51, 0x5E, 0x44, 0xFF }
}, // 246

{
	0x081B3BA5, 0x08082001,
	{0x40, 0xFD, 0xBE, 0x79, 0x29, 0x2A, 0x72, 0xD6, 0x1C, 0xB1, 0x21, 0x9C, 0xD3, 0xC6, 0x7A, 0xE3 }
}, // 247

{
	0x00000001, 0x80000000,
	{0xC4, 0x1E, 0xAA, 0xDB, 0x22, 0x92, 0xF3, 0x16, 0x95, 0xE6, 0xBC, 0x3B, 0x11, 0x1F, 0xC6, 0x24 }
},// 248

{
	0x00000001, 0x80000000,
	{0xC4, 0x1E, 0xAA, 0xDB, 0x22, 0x92, 0xF3, 0x16, 0x95, 0xE6, 0xBC, 0x3B, 0x11, 0x1F, 0xC6, 0x24 }
},// 249

{
	0x01040020, 0x00000001,
	{0x9E, 0x08, 0xE2, 0x2E, 0x3E, 0x17, 0x07, 0xCF, 0x80, 0x73, 0x03, 0x15, 0xE4, 0x5F, 0xFA, 0x5D }
}, // 250

{
	0x02080041, 0x00000002,
	{0x30, 0xBF, 0x3D, 0xA5, 0xE4, 0x68, 0xD8, 0x4F, 0x69, 0x2C, 0xFA, 0x04, 0x14, 0xF0, 0x10, 0x47 }
}, // 251

{
	0x04100082, 0x00000004,
	{0xFA, 0xBF, 0xF9, 0x12, 0x24, 0x34, 0x7F, 0x01, 0x83, 0x03, 0x6F, 0x1F, 0xCA, 0xA2, 0xE1, 0x8C }
}, // 252

{
	0x08200104, 0x00000008,
	{0xCE, 0x18, 0x0B, 0xB6, 0xFF, 0xBA, 0x56, 0xB3, 0xA4, 0xEF, 0x9A, 0x29, 0xF4, 0x56, 0x2E, 0x99 }
}, // 253

{
	0x10400208, 0x00000010,
	{0x05, 0xFD, 0xB2, 0x9B, 0x4E, 0xD6, 0x58, 0x8E, 0x33, 0x2F, 0x2B, 0xE1, 0xF4, 0x22, 0xDC, 0x3A }
}, // 254

{
	0x20800410, 0x00000020,
	{0x87, 0xB1, 0x5B, 0x0C, 0x58, 0x26, 0x3E, 0x04, 0x48, 0xC3, 0xC0, 0xFC, 0x37, 0xA3, 0x20, 0xE3 }
}, // 255

{
	0x41000821, 0x00000040,
	{0x72, 0x62, 0x87, 0x18, 0xC9, 0x97, 0xF1, 0x4A, 0x60, 0x87, 0x0C, 0xDB, 0xBE, 0xA7, 0x9D, 0xC3 }
}, // 256

{
	0x82001042, 0x00000080,
	{0xE1, 0x01, 0x3C, 0x92, 0x66, 0xBF, 0xCC, 0xE1, 0xF2, 0x00, 0xA6, 0xF0, 0xD1, 0x33, 0x1C, 0x9B }
}, // 257

{
	0x04002085, 0x00000101,
	{0xAC, 0x06, 0xB1, 0x7A, 0x1A, 0x24, 0xE6, 0xC8, 0x3B, 0x60, 0x6B, 0x1A, 0xD3, 0xBF, 0xA1, 0x39 }
}, // 258

{
	0x0800410B, 0x00000202,
	{0x54, 0x68, 0xBC, 0x4B, 0xA9, 0x73, 0xE6, 0xFC, 0x29, 0xEE, 0xBE, 0x4F, 0x27, 0x08, 0xDB, 0x7E }
}, // 259

{
	0x10008216, 0x00000404,
	{0x8A, 0xFF, 0x65, 0xA9, 0x3C, 0xED, 0x45, 0x63, 0x6F, 0xCB, 0xC4, 0x72, 0xB6, 0xD5, 0x5B, 0xAA }
}, // 260

{
	0x2001042D, 0x00000808,
	{0x27, 0x88, 0xA7, 0xE3, 0x30, 0x4D, 0x39, 0x34, 0x59, 0xE7, 0xAF, 0xDC, 0x5B, 0xCE, 0x75, 0x93 }
}, // 261

{
	0x4002085A, 0x00001010,
	{0xE5, 0x0B, 0x79, 0x93, 0x36, 0x8F, 0x7D, 0xF5, 0x4B, 0x01, 0x27, 0xF5, 0x8A, 0x8F, 0x0D, 0x7D }
}, // 262

{
	0x800410B4, 0x00002020,
	{0xA7, 0xB0, 0x58, 0x57, 0x04, 0x9A, 0xB2, 0xE9, 0xB0, 0x6F, 0xBA, 0x20, 0xAD, 0xD1, 0x4B, 0x6B }
}, // 263

{
	0x00082168, 0x00004041,
	{0x79, 0xAD, 0x27, 0xBA, 0xF0, 0xB1, 0x85, 0x16, 0xF2, 0x21, 0x72, 0xFB, 0xFF, 0xE7, 0x67, 0x24 }
}, // 264

{
	0x001042D1, 0x00008082,
	{0xDA, 0xCB, 0xAF, 0x24, 0x52, 0xE5, 0x93, 0x21, 0x76, 0x56, 0xF6, 0x84, 0xDA, 0xA8, 0xC8, 0xBB }
}, // 265

{
	0x002085A2, 0x00010104,
	{0x67, 0x5C, 0x2E, 0xE0, 0xCA, 0x18, 0xFE, 0x8B, 0xE7, 0x0A, 0x2B, 0x59, 0x31, 0xFE, 0xB6, 0xF4 }
}, // 266

{
	0x00410B45, 0x00020208,
	{0xE7, 0x92, 0xD1, 0xE8, 0xD6, 0x43, 0xF0, 0x57, 0x03, 0x9C, 0xA0, 0x26, 0xC9, 0xB6, 0x20, 0x84 }
}, // 267

{
	0x0082168B, 0x00040410,
	{0x50, 0x1B, 0xB8, 0x6D, 0xA0, 0x9E, 0xB7, 0x5D, 0xB5, 0xB5, 0xDD, 0xB1, 0x19, 0xA3, 0x4E, 0xDE }
}, // 268

{
	0x01042D16, 0x00080820,
	{0x40, 0x0B, 0x75, 0x52, 0x9B, 0xE1, 0x87, 0x97, 0xCF, 0xA7, 0xC1, 0x4F, 0xAE, 0x52, 0x88, 0x11 }
}, // 269

{
	0x02085A2D, 0x00101040,
	{0x18, 0xD2, 0x20, 0x72, 0x19, 0x0A, 0x7A, 0xD6, 0x74, 0xFB, 0x80, 0x70, 0x63, 0x5C, 0x89, 0x8A }
}, // 270

{
	0x0410B45B, 0x00202080,
	{0x59, 0x17, 0x20, 0x0F, 0x2A, 0x29, 0x96, 0x09, 0x13, 0x34, 0xD9, 0x52, 0xFC, 0xC1, 0x31, 0x65 }
}, // 271

{
	0x082168B6, 0x00404100,
	{0x6A, 0x30, 0x8D, 0x5D, 0xBE, 0xA0, 0x9B, 0x82, 0xC0, 0xD7, 0xA4, 0x79, 0x37, 0xD3, 0xC0, 0x8E }
}, // 272

{
	0x1042D16C, 0x00808200,
	{0x25, 0x80, 0x3C, 0x3E, 0x8A, 0x26, 0x0D, 0xAF, 0xDD, 0x11, 0xBF, 0x16, 0x26, 0xE4, 0xD1, 0xA9 }
}, // 273

{
	0x2085A2D8, 0x01010400,
	{0x97, 0x42, 0x07, 0x88, 0x95, 0x78, 0xD1, 0xAB, 0x44, 0x31, 0xEA, 0xF4, 0x7B, 0x2B, 0x61, 0x8C }
}, // 274

{
	0x410B45B1, 0x02020800,
	{0xD3, 0x82, 0xAE, 0x99, 0x12, 0x90, 0xF7, 0x63, 0x62, 0x83, 0x99, 0x1C, 0x2F, 0x7F, 0x67, 0xAF }
}, // 275

{
	0x82168B63, 0x04041000,
	{0xCC, 0x4D, 0x18, 0xAA, 0xC7, 0x1F, 0x50, 0x9C, 0x7E, 0x5B, 0x2D, 0xF1, 0x63, 0xBF, 0xC5, 0x09 }
}, // 276

{
	0x042D16C7, 0x08082001,
	{0x59, 0x90, 0x38, 0xFF, 0x6D, 0xC1, 0x65, 0xB6, 0x8A, 0x63, 0xED, 0x14, 0x89, 0x45, 0x4F, 0xA3 }
}, // 277

{
	0x82001080, 0x00000080,
	{0x76, 0xB9, 0x47, 0xE7, 0x1C, 0x5C, 0x07, 0x96, 0x0E, 0x15, 0x20, 0x74, 0x24, 0xCD, 0x7C, 0x1A }
}, // 278

{
	0x04002101, 0x00000101,
	{0x97, 0x95, 0x09, 0x1D, 0x3D, 0x0F, 0x60, 0x62, 0xE8, 0x76, 0xB1, 0xE3, 0x32, 0xCA, 0xFD, 0xD5 }
}, // 279

{
	0x08004203, 0x00000202,
	{0x60, 0x52, 0x35, 0x6F, 0xCA, 0x57, 0x35, 0x6F, 0x0F, 0x0D, 0x9A, 0x3A, 0x73, 0x2F, 0x92, 0xE6 }
}, // 280

{
	0x10008407, 0x00000404,
	{0xDD, 0x4B, 0xAF, 0x42, 0xE2, 0x16, 0x0D, 0x57, 0xDD, 0x07, 0x68, 0xF0, 0x34, 0x5D, 0xDC, 0x88 }
}, // 281

{
	0x2001080F, 0x00000808,
	{0xFB, 0x4C, 0x48, 0xEF, 0x15, 0x20, 0x83, 0x0F, 0xFD, 0xE7, 0xDE, 0xA0, 0x3D, 0x70, 0xD6, 0xDE }
}, // 282

{
	0x4002101E, 0x00001010,
	{0x37, 0xCA, 0xEE, 0x43, 0x0B, 0xCA, 0x47, 0x81, 0xFF, 0xE5, 0xE6, 0x41, 0xF1, 0xBA, 0x5A, 0xA1 }
}, // 283

{
	0x8004203D, 0x00002020,
	{0x41, 0x36, 0x21, 0x3A, 0xE8, 0x3D, 0x12, 0x99, 0x49, 0x79, 0x63, 0xD7, 0x4F, 0xD8, 0xBA, 0xFB }
}, // 284

{
	0x0008407A, 0x00004041,
	{0xAF, 0x0E, 0x84, 0x9D, 0x39, 0xCF, 0x9C, 0x85, 0x7B, 0x7C, 0x5E, 0xB4, 0xE8, 0x8E, 0x67, 0x5A }
}, // 285

{
	0x001080F4, 0x00008082,
	{0xC4, 0x49, 0x78, 0x0E, 0x9E, 0x86, 0x24, 0x73, 0x4A, 0xB6, 0xCD, 0x69, 0x1A, 0x2A, 0xF1, 0xFD }
}, // 286

{
	0x002101E8, 0x00010104,
	{0xB1, 0x32, 0x5F, 0xE2, 0xA2, 0x49, 0x65, 0x80, 0x88, 0x12, 0x71, 0x62, 0x76, 0x3D, 0xCB, 0x67 }
}, // 287

{
	0x004203D0, 0x00020208,
	{0x13, 0x23, 0x06, 0x66, 0x78, 0xDF, 0x65, 0x4F, 0xBC, 0xB7, 0xD6, 0x6F, 0x3A, 0x35, 0x78, 0xB6 }
}, // 288

{
	0x008407A0, 0x00040410,
	{0xFB, 0x32, 0x09, 0xF8, 0x12, 0x0F, 0xCE, 0x4A, 0x4F, 0x34, 0xC6, 0x87, 0xBB, 0x35, 0xC0, 0x08 }
}, // 289

{
	0x01080F40, 0x00080820,
	{0xCE, 0x9E, 0x65, 0x42, 0x72, 0x94, 0xB6, 0xC1, 0x07, 0x19, 0xBD, 0x11, 0xB2, 0x58, 0x71, 0x29 }
}, // 290

{
	0x02101E80, 0x00101040,
	{0x62, 0xD3, 0x02, 0x01, 0x80, 0xC6, 0x16, 0xC1, 0x1F, 0x0C, 0x7B, 0xAF, 0x9C, 0x9B, 0x63, 0x19 }
}, // 291

{
	0x04203D00, 0x00202080,
	{0xC3, 0xFA, 0x52, 0xE9, 0x73, 0x4D, 0xA4, 0xD9, 0xC6, 0xF8, 0xE3, 0x31, 0x62, 0x59, 0xC8, 0x77 }
}, // 292

{
	0x08407A00, 0x00404100,
	{0x29, 0x76, 0xFD, 0x87, 0x13, 0x4C, 0xC2, 0x35, 0xA8, 0x28, 0x42, 0xC9, 0xD6, 0x9B, 0xA5, 0x9B }
}, // 293

{
	0x1080F400, 0x00808200,
	{0x1C, 0xF7, 0xFE, 0x2C, 0x5D, 0xC0, 0x8B, 0x8B, 0xA8, 0x03, 0x5E, 0x7F, 0x66, 0x34, 0x4B, 0xE6 }
}, // 294

{
	0x2101E801, 0x01010400,
	{0x4C, 0x9E, 0x98, 0x86, 0xDB, 0xE4, 0x5E, 0x52, 0x0F, 0x06, 0xF7, 0xED, 0xE5, 0xB2, 0xB9, 0x0A }
}, // 295

{
	0x4203D003, 0x02020800,
	{0xDF, 0x38, 0x9D, 0x57, 0xDF, 0xB9, 0x10, 0x93, 0x05, 0x6D, 0xAA, 0x1F, 0x06, 0x09, 0x89, 0x62 }
}, // 296

{
	0x8407A006, 0x04041000,
	{0x40, 0x4E, 0x8E, 0xCF, 0x53, 0x65, 0x05, 0x83, 0x7A, 0x00, 0x90, 0xED, 0x6C, 0x6F, 0xE5, 0x28 }
}, // 297

{
	0x080F400D, 0x08082001,
	{0x5E, 0xDC, 0xE3, 0x6B, 0x13, 0xE2, 0x54, 0xE1, 0xCD, 0x8C, 0xEF, 0xA1, 0xAC, 0x98, 0x40, 0x1A }
}, // 298

{
	0x04041400, 0x00000000,
	{0xCF, 0xAD, 0xCE, 0x41, 0xD1, 0x31, 0x82, 0x07, 0x5A, 0x61, 0x81, 0xCD, 0xDE, 0x3B, 0x19, 0x7A }
}, // 299

{
	0x08082801, 0x00000000,
	{0x0C, 0x37, 0x3A, 0xC4, 0xC8, 0xD5, 0x41, 0x45, 0xD3, 0xA0, 0x2B, 0xD3, 0x51, 0x66, 0x2A, 0x85 }
}, // 300

{
	0x10105002, 0x00000000,
	{0x77, 0x02, 0xFE, 0x14, 0x51, 0xA5, 0x20, 0x02, 0xC0, 0xF4, 0x59, 0x5A, 0xE9, 0xE6, 0x6B, 0xAF }
}, // 301

{
	0x2020A004, 0x00000000,
	{0x28, 0xAF, 0x2E, 0xA6, 0x32, 0x2A, 0xD3, 0x73, 0xCC, 0x66, 0x50, 0x08, 0x14, 0x47, 0x7F, 0x57 }
}, // 302

{
	0x40414008, 0x00000000,
	{0xBF, 0x68, 0xF5, 0x6C, 0xF0, 0xB6, 0xE1, 0xBF, 0xC0, 0xD1, 0xF7, 0xD1, 0x4A, 0x85, 0x87, 0x9B }
}, // 303

{
	0x80828010, 0x00000000,
	{0xAB, 0xFF, 0x46, 0x79, 0xD6, 0xC0, 0xC0, 0x7E, 0x2A, 0x4B, 0x08, 0x77, 0xDD, 0x11, 0x15, 0x62 }
}, // 304

{
	0x01050020, 0x00000001,
	{0xD3, 0x76, 0xC6, 0xA9, 0x1B, 0xD7, 0x1D, 0x5D, 0xD1, 0x5E, 0xE0, 0x30, 0x41, 0xEB, 0xF7, 0xE7 }
}, // 305

{
	0x020A0041, 0x00000002,
	{0x43, 0x98, 0x2B, 0xBE, 0x83, 0xCB, 0x68, 0x93, 0xF6, 0x07, 0x76, 0x45, 0x4B, 0xE6, 0x1B, 0xA1 }
}, // 306

{
	0x04140082, 0x00000004,
	{0x14, 0x4F, 0x25, 0xC9, 0xB2, 0x2D, 0xCC, 0x3A, 0xD5, 0xCF, 0xDF, 0x6E, 0xA8, 0xBC, 0x90, 0x5F }
}, // 307

{
	0x08280104, 0x00000008,
	{0xF4, 0x1E, 0x15, 0xEC, 0x4F, 0x4A, 0xC2, 0x4F, 0xD8, 0xAA, 0xD6, 0xC4, 0x4C, 0xAC, 0x61, 0x3C }
}, // 308

{
	0x10500208, 0x00000010,
	{0x20, 0x35, 0x54, 0xFC, 0x12, 0xCC, 0xFC, 0x93, 0x8F, 0x5F, 0xDE, 0xC5, 0xF4, 0x76, 0x02, 0x29 }
}, // 309

{
	0x20A00410, 0x00000020,
	{0x40, 0xCC, 0x88, 0x10, 0x85, 0xA6, 0x65, 0xD0, 0xBA, 0x30, 0xE5, 0xF1, 0xDE, 0x70, 0x9D, 0xF5 }
}, // 310

{
	0x41400820, 0x00000040,
	{0xBE, 0xBD, 0xD2, 0x78, 0xB1, 0x2F, 0xF3, 0x46, 0xEB, 0x9C, 0xEB, 0x81, 0xFA, 0xE2, 0x6F, 0x36 }
}, // 311

{
	0x82801040, 0x00000080,
	{0x51, 0x8A, 0xB3, 0x0F, 0xA4, 0xF2, 0x52, 0x96, 0xD6, 0x45, 0x93, 0x16, 0x51, 0x0D, 0xAE, 0x81 }
}, // 312

{
	0x05002081, 0x00000101,
	{0xC6, 0xDC, 0x51, 0x3F, 0x5B, 0x34, 0xFE, 0x61, 0xFF, 0xE1, 0xD4, 0x66, 0x2E, 0x65, 0x4E, 0x18 }
}, // 313

{
	0x0A004102, 0x00000202,
	{0xD9, 0xC5, 0xBF, 0x31, 0x46, 0x0B, 0x48, 0x6F, 0x78, 0x41, 0xD0, 0xB2, 0x20, 0x99, 0xD6, 0xFE }
}, // 314

{
	0x14008205, 0x00000404,
	{0x20, 0xFD, 0xD2, 0x29, 0x1B, 0xC8, 0xBB, 0x46, 0x36, 0xE3, 0x3E, 0x76, 0x5C, 0x43, 0xC0, 0xFE }
}, // 315

{
	0x2801040B, 0x00000808,
	{0x27, 0x85, 0x29, 0xE7, 0xC3, 0x98, 0x70, 0x00, 0x50, 0xFA, 0xA3, 0xFB, 0x76, 0xDD, 0x85, 0x6D }
}, // 316

{
	0x50020816, 0x00001010,
	{0x80, 0x77, 0x75, 0x93, 0x89, 0x9D, 0x06, 0x85, 0x09, 0x8A, 0x72, 0xD0, 0x5F, 0x5D, 0x4A, 0x6A }
}, // 317

{
	0xA004102C, 0x00002020,
	{0xDF, 0xBA, 0x00, 0x2C, 0xC4, 0xB0, 0x5C, 0x5B, 0x24, 0xBB, 0x4A, 0xFB, 0x76, 0x94, 0xC0, 0x69 }
}, // 318

{
	0x40082059, 0x00004041,
	{0xA3, 0xC4, 0x2C, 0x31, 0x95, 0x21, 0x5C, 0x0A, 0xFC, 0x69, 0xAA, 0x29, 0xEC, 0xED, 0x80, 0xCE }
}, // 319

{
	0x801040B2, 0x00008082,
	{0x3A, 0xBE, 0x5F, 0x0F, 0x7B, 0xF5, 0x34, 0xB4, 0x7A, 0xC3, 0x5D, 0x70, 0xB6, 0x69, 0xB5, 0x0D }
}, // 320

{
	0x00208165, 0x00010105,
	{0x0F, 0xAC, 0xF3, 0xEC, 0x9E, 0x7D, 0x22, 0xC1, 0xE4, 0x42, 0x02, 0x42, 0x8C, 0x44, 0xC6, 0xA5 }
}, // 321

{
	0x004102CB, 0x0002020A,
	{0x31, 0x68, 0xFD, 0x60, 0x0D, 0x93, 0x34, 0x23, 0xC8, 0xDE, 0xBD, 0xD4, 0xD9, 0x47, 0x0B, 0x05 }
}, // 322

{
	0x00820597, 0x00040414,
	{0x36, 0xEC, 0x18, 0xC0, 0x70, 0xB7, 0x37, 0x78, 0xBA, 0xFF, 0x72, 0xD3, 0x26, 0x74, 0xCB, 0xD2 }
}, // 323

{
	0x01040B2F, 0x00080828,
	{0xC4, 0x6D, 0x13, 0x12, 0xE9, 0x46, 0xCB, 0x3D, 0x33, 0xD3, 0xB2, 0x62, 0x59, 0xC3, 0x58, 0x6E }
}, // 324

{
	0x0208165F, 0x00101050,
	{0x7F, 0x35, 0x72, 0x6D, 0xDA, 0xAE, 0x67, 0xD8, 0xF1, 0x4A, 0x3C, 0xC0, 0xE6, 0xFA, 0xDC, 0x45 }
}, // 325

{
	0x04102CBE, 0x002020A0,
	{0xF7, 0xA5, 0xAA, 0x3B, 0x6E, 0xA0, 0x68, 0x52, 0x4B, 0xC5, 0xD6, 0xAC, 0x65, 0x32, 0x1B, 0x74 }
}, // 326

{
	0x0820597C, 0x00404140,
	{0xC8, 0xCF, 0x53, 0x9F, 0x12, 0x8E, 0xA0, 0xDB, 0x57, 0x15, 0x3F, 0xD8, 0xC3, 0x07, 0xBD, 0x80 }
}, // 327

{
	0x1040B2F8, 0x00808280,
	{0x33, 0xAB, 0xBF, 0xA0, 0xAA, 0xD8, 0x67, 0xBD, 0xEE, 0x4F, 0xC4, 0xF2, 0xC6, 0xA3, 0x89, 0x0A }
}, // 328

{
	0x208165F0, 0x01010500,
	{0x3B, 0x74, 0xFE, 0x8F, 0x5C, 0xA0, 0x20, 0xFC, 0x5C, 0x9F, 0xE6, 0xFA, 0x0C, 0xA3, 0x2C, 0xA2 }
}, // 329

{
	0x4102CBE1, 0x02020A00,
	{0x08, 0xCD, 0x28, 0x45, 0xCC, 0x3C, 0x8C, 0x4D, 0x19, 0x6C, 0xAA, 0xE6, 0x0B, 0x33, 0x9A, 0x86 }
}, // 330

{
	0x820597C2, 0x04041400,
	{0x86, 0x9C, 0xDE, 0x3D, 0x62, 0x3C, 0x0D, 0xB7, 0x64, 0x6C, 0x29, 0x37, 0x84, 0x5A, 0xBB, 0xF3 }
}, // 331

{
	0x040B2F84, 0x08082801,
	{0x64, 0xE3, 0x94, 0x79, 0x74, 0x50, 0x69, 0x92, 0x2B, 0xF4, 0x18, 0x9A, 0xB1, 0x6F, 0xD9, 0x18 }
}, // 332

{
	0x0C008209, 0x00000404,
	{0xF2, 0xB1, 0x87, 0xBD, 0x3B, 0xFF, 0x33, 0x27, 0x42, 0xD5, 0x73, 0xF4, 0xD7, 0x41, 0x43, 0x70 }
}, // 333

{
	0x18010413, 0x00000808,
	{0x99, 0xC8, 0x14, 0x46, 0xBD, 0x8E, 0x62, 0xFE, 0xCB, 0x89, 0x71, 0xEE, 0x8A, 0xF4, 0xC4, 0x24 }
}, // 334

{
	0x30020826, 0x00001010,
	{0x30, 0xA1, 0x14, 0xCB, 0x69, 0x32, 0x44, 0x20, 0x2B, 0x91, 0xEB, 0x32, 0xF6, 0xC1, 0x03, 0x99 }
}, // 335

{
	0x6004104D, 0x00002020,
	{0x26, 0xCB, 0x01, 0xFB, 0x5E, 0xFC, 0x23, 0xD0, 0xCF, 0xD5, 0x3C, 0x75, 0x7C, 0xB7, 0x0E, 0x55 }
}, // 336

{
	0xC008209B, 0x00004040,
	{0x2B, 0xF2, 0x87, 0xFE, 0xF0, 0x82, 0xD8, 0x99, 0x7D, 0xD6, 0x55, 0xBE, 0x26, 0xD0, 0x95, 0x05 }
}, // 337

{
	0x80104136, 0x00008081,
	{0x31, 0x52, 0xE2, 0x82, 0x66, 0x66, 0x9D, 0x14, 0x30, 0xDD, 0xD2, 0x59, 0xF8, 0x84, 0xBB, 0xA1 }
}, // 338

{
	0x0020826D, 0x00010103,
	{0xB3, 0x12, 0x1C, 0xB2, 0xBE, 0x4C, 0x7B, 0xC7, 0x25, 0x3A, 0xBC, 0xEB, 0x20, 0x17, 0xF5, 0x93 }
}, // 339

{
	0x004104DB, 0x00020206,
	{0xC5, 0xA7, 0x96, 0x60, 0x90, 0x3D, 0x2A, 0x62, 0x03, 0x40, 0xDD, 0x73, 0xC0, 0x66, 0x88, 0x8F }
}, // 340

{
	0x008209B7, 0x0004040C,
	{0xFF, 0x35, 0x15, 0x90, 0xF1, 0x7E, 0xA1, 0x8D, 0x5A, 0x1E, 0xF9, 0xBA, 0x4D, 0xD1, 0xB8, 0xFF }
}, // 341

{
	0x0104136E, 0x00080818,
	{0xFC, 0x32, 0xEA, 0x8D, 0x80, 0xDB, 0x8B, 0xDD, 0xA1, 0x61, 0xDF, 0x49, 0x66, 0x6A, 0xB9, 0x1E }
}, // 342

{
	0x020826DD, 0x00101030,
	{0xBF, 0xDC, 0x19, 0x77, 0x0C, 0x48, 0xC9, 0xCF, 0xDF, 0x58, 0xF5, 0x2C, 0xA2, 0xCD, 0x52, 0x18 }
}, // 343

{
	0x04104DBB, 0x00202060,
	{0xD6, 0xC1, 0xB8, 0x59, 0xED, 0x01, 0x3C, 0x86, 0x1C, 0x55, 0xFF, 0x4F, 0xFE, 0x45, 0xD8, 0xEF }
}, // 344

{
	0x08209B76, 0x004040C0,
	{0xB6, 0x8B, 0xDB, 0xE5, 0x39, 0xBD, 0x01, 0xD7, 0xEB, 0x1C, 0x23, 0x9F, 0xEB, 0xC8, 0xC8, 0x9F }
}, // 345

{
	0x104136ED, 0x00808180,
	{0xB2, 0x3E, 0x92, 0x76, 0x21, 0x5E, 0x70, 0x91, 0x85, 0x7F, 0xEA, 0x03, 0xF7, 0xF3, 0x3D, 0x2F }
}, // 346

{
	0x20826DDA, 0x01010300,
	{0xF2, 0xB2, 0x16, 0x08, 0x3F, 0x11, 0x09, 0xED, 0x75, 0x0F, 0x09, 0x9F, 0xC4, 0xF9, 0x1B, 0x4F }
}, // 347

{
	0x4104DBB4, 0x02020600,
	{0x56, 0xB5, 0x29, 0x02, 0x1B, 0x85, 0x4C, 0xCA, 0x8E, 0x18, 0x1F, 0xC9, 0xBC, 0x88, 0x9E, 0xAA }
}, // 348

{
	0x8209B769, 0x04040C00,
	{0x97, 0x37, 0xED, 0xC5, 0xDD, 0x0A, 0x76, 0xE3, 0x22, 0x0C, 0x15, 0x68, 0xB3, 0xC2, 0xEF, 0x85 }
}, // 349

{
	0x04136ED3, 0x08081801,
	{0xAB, 0xC2, 0x58, 0x1E, 0x6E, 0xDC, 0x47, 0x99, 0x60, 0x5C, 0x80, 0x25, 0x9F, 0x99, 0x17, 0xC0 }
}, // 350

{
	0x41000820, 0x00000040,
	{0x03, 0x59, 0x54, 0x5A, 0x35, 0xD8, 0x4B, 0xF2, 0x28, 0xEF, 0x50, 0x30, 0x80, 0x3B, 0x32, 0x49 }
}, // 351

{
	0x82001041, 0x00000080,
	{0x51, 0x52, 0x43, 0xC5, 0xB5, 0x66, 0xDB, 0x34, 0x79, 0x1A, 0xAF, 0x12, 0x67, 0x59, 0x26, 0xDA }
}, // 352

{
	0x04002083, 0x00000101,
	{0xA2, 0xE5, 0x23, 0x12, 0x40, 0xD6, 0xEC, 0xE0, 0x9E, 0x75, 0x0A, 0xA6, 0x43, 0xE0, 0x6E, 0xC8 }
}, // 353

{
	0x08004107, 0x00000202,
	{0xBE, 0x10, 0x51, 0x72, 0x9B, 0xC9, 0xE1, 0x27, 0x61, 0x0B, 0x3A, 0xAA, 0x82, 0x30, 0xC0, 0x2B }
}, // 354

{
	0x1000820E, 0x00000404,
	{0xC5, 0x25, 0xFC, 0x93, 0x56, 0x95, 0x71, 0x85, 0x4F, 0x3D, 0x67, 0x4B, 0x67, 0xD4, 0x22, 0x08 }
}, // 355

{
	0x2001041C, 0x00000808,
	{0x6B, 0x9E, 0x52, 0xE3, 0xB5, 0x06, 0x34, 0x37, 0x3C, 0xD5, 0xB4, 0x31, 0x0A, 0xCA, 0xC3, 0x43 }
}, // 356

{
	0x40020839, 0x00001010,
	{0x77, 0x5B, 0x9E, 0x62, 0xD1, 0x40, 0xB8, 0xD0, 0x34, 0x6D, 0xED, 0x95, 0xA0, 0xAA, 0x67, 0xC9 }
}, // 357

{
	0x80041072, 0x00002020,
	{0x36, 0xC5, 0x98, 0x94, 0x29, 0xC8, 0x9B, 0x01, 0x3A, 0x6F, 0x85, 0xD5, 0x88, 0x63, 0x22, 0xC1 }
}, // 358

{
	0x000820E5, 0x00004041,
	{0x19, 0x11, 0x85, 0xCC, 0xC6, 0xAA, 0x4D, 0x66, 0x1D, 0x36, 0x31, 0x2C, 0x04, 0x04, 0x0C, 0x12 }
}, // 359

{
	0x001041CB, 0x00008082,
	{0x2F, 0xF4, 0x68, 0x6B, 0x45, 0x3F, 0x80, 0x10, 0x87, 0x44, 0x74, 0xF8, 0x95, 0xAA, 0x8E, 0x86 }
}, // 360

{
	0x00208396, 0x00010104,
	{0xF8, 0xC6, 0x50, 0xB3, 0x9D, 0x9A, 0xDC, 0x6D, 0x36, 0x0B, 0x94, 0x6B, 0x05, 0x32, 0xCC, 0x0B }
}, // 361

{
	0x0041072C, 0x00020208,
	{0x8D, 0x0E, 0x4D, 0x56, 0x11, 0x70, 0x1D, 0x4B, 0x9B, 0x76, 0x4D, 0xF4, 0x7C, 0xC3, 0x7E, 0xA7 }
}, // 362

{
	0x00820E59, 0x00040410,
	{0x65, 0x89, 0x7B, 0x8B, 0x44, 0xDD, 0x81, 0xEC, 0x08, 0x69, 0x5E, 0x1F, 0x86, 0x93, 0x7F, 0x75 }
}, // 363

{
	0x01041CB2, 0x00080820,
	{0xEA, 0x41, 0x3A, 0x02, 0xA0, 0x9E, 0xDA, 0x9F, 0x78, 0xB7, 0x8C, 0x7E, 0x95, 0x53, 0x2B, 0xF4 }
}, // 364

{
	0x02083964, 0x00101040,
	{0x26, 0x95, 0x93, 0x44, 0x1F, 0xA0, 0x94, 0x9F, 0xC2, 0x05, 0xF5, 0x0D, 0xB5, 0x91, 0xE2, 0x58 }
}, // 365

{
	0x041072C8, 0x00202080,
	{0x57, 0xD8, 0xE4, 0x1D, 0xAD, 0xEE, 0x5E, 0x57, 0xDD, 0xCC, 0x5A, 0x2A, 0xE7, 0x22, 0xAB, 0x6D }
}, // 366

{
	0x0820E591, 0x00404100,
	{0x99, 0x24, 0x3D, 0xA1, 0x0C, 0xC3, 0xC3, 0x51, 0x61, 0xC1, 0x81, 0x12, 0x47, 0x5A, 0x5F, 0x80 }
}, // 367

{
	0x1041CB23, 0x00808200,
	{0xB4, 0x1E, 0xF5, 0xE9, 0xDC, 0xD1, 0x2C, 0xC9, 0xCA, 0x0B, 0xED, 0x1D, 0xEC, 0xDE, 0xD3, 0xDF }
}, // 368

{
	0x20839647, 0x01010400,
	{0x0D, 0xDD, 0x6C, 0xF1, 0x4B, 0xEE, 0x07, 0x76, 0x40, 0x32, 0x1A, 0x9B, 0x13, 0x2B, 0x7A, 0x96 }
}, // 369

{
	0x41072C8E, 0x02020800,
	{0x18, 0x5D, 0xBE, 0xD1, 0xCB, 0x93, 0x94, 0x39, 0x12, 0xD2, 0xEB, 0xD9, 0x64, 0x8D, 0x3C, 0x59 }
}, // 370

{
	0x820E591C, 0x04041000,
	{0xE3, 0xC1, 0x77, 0x6A, 0x87, 0x36, 0xF5, 0x36, 0x04, 0x29, 0x4D, 0xE8, 0xED, 0x1A, 0x5E, 0x92 }
}, // 371

{
	0x041CB239, 0x08082001,
	{0x34, 0x24, 0x9B, 0xFF, 0x54, 0x5D, 0xFC, 0xA0, 0xB3, 0x11, 0x90, 0x45, 0x91, 0x8F, 0xE0, 0xFA }
}, // 372

{
	0x01050021, 0x00000001,
	{0x56, 0x63, 0xCA, 0x4C, 0x96, 0xDB, 0x14, 0xFF, 0xC9, 0xB8, 0x4F, 0xC6, 0xE5, 0x45, 0x3F, 0x59 }
}, // 373

{
	0x020A0042, 0x00000002,
	{0x01, 0x4D, 0x28, 0x31, 0xEC, 0x69, 0x4F, 0x95, 0xDE, 0x3F, 0xC1, 0x48, 0x83, 0xBC, 0x2D, 0x9A }
}, // 374

{
	0x04140084, 0x00000004,
	{0xAD, 0xAF, 0xB7, 0x37, 0x32, 0x1C, 0x38, 0xE2, 0x61, 0x3D, 0xD8, 0xCA, 0x76, 0x0F, 0x76, 0x21 }
}, // 375

{
	0x08280108, 0x00000008,
	{0x87, 0x21, 0xC7, 0xF7, 0x2A, 0x1A, 0x19, 0x14, 0x5F, 0x88, 0xAA, 0xB2, 0x66, 0xF6, 0x63, 0x6D }
}, // 376

{
	0x10500210, 0x00000010,
	{0x40, 0x47, 0x1C, 0x07, 0x88, 0x55, 0x64, 0x49, 0x23, 0x2A, 0x34, 0x34, 0x08, 0x4F, 0xBD, 0xF7 }
}, // 377

{
	0x20A00420, 0x00000020,
	{0xDF, 0xD2, 0x6A, 0x95, 0x0D, 0x5F, 0x7C, 0x02, 0xBA, 0xA7, 0xED, 0xA9, 0x95, 0x01, 0x2D, 0xBB }
}, // 378

{
	0x41400841, 0x00000040,
	{0x5D, 0x07, 0xCF, 0xB4, 0xC2, 0x8F, 0x3C, 0x1E, 0x6A, 0x4B, 0x7A, 0x18, 0xDD, 0x43, 0xE0, 0xA3 }
}, // 379

{
	0x82801082, 0x00000080,
	{0x36, 0x3C, 0x17, 0xF0, 0x18, 0x3C, 0x73, 0x73, 0x0F, 0x97, 0x87, 0x09, 0x1C, 0xE6, 0xEC, 0xB9 }
}, // 380

{
	0x05002104, 0x00000101,
	{0xA5, 0x46, 0x7A, 0xC2, 0x51, 0xA9, 0xCD, 0xBB, 0xFE, 0x6C, 0x21, 0x1D, 0xD5, 0xBF, 0x50, 0x50 }
}, // 381

{
	0x0A004208, 0x00000202,
	{0xD9, 0xFE, 0x19, 0xAB, 0x1E, 0x67, 0xA2, 0xDA, 0x9B, 0x4C, 0x2F, 0x07, 0xC3, 0xF5, 0x4A, 0xAF }
}, // 382

{
	0x14008411, 0x00000404,
	{0xAB, 0x39, 0x39, 0x14, 0xBD, 0xA8, 0x3F, 0x0D, 0x4B, 0xE9, 0x7A, 0x9A, 0x77, 0x73, 0x69, 0xDF }
}, // 383

{
	0x28010822, 0x00000808,
	{0x9E, 0x41, 0xEA, 0x8D, 0x71, 0x26, 0xAD, 0xA1, 0x74, 0x3B, 0x4F, 0x0D, 0x84, 0xF4, 0xF7, 0x5C }
}, // 384

{
	0x50021044, 0x00001010,
	{0x27, 0x8F, 0x56, 0xA2, 0x13, 0x31, 0xBF, 0x47, 0xF6, 0x0C, 0x1D, 0x58, 0x9F, 0xCB, 0x49, 0x5D }
}, // 385

{
	0xA0042088, 0x00002020,
	{0x91, 0xF1, 0x0A, 0xE6, 0x31, 0x5D, 0x42, 0xA5, 0x46, 0x24, 0x7B, 0x6C, 0x97, 0x74, 0x89, 0x51 }
}, // 386

{
	0x40084111, 0x00004041,
	{0x8F, 0x32, 0x15, 0x44, 0x96, 0xE1, 0x29, 0x22, 0x3F, 0xFC, 0xAF, 0x7B, 0x25, 0x02, 0xC4, 0xEB }
}, // 387

{
	0x80108222, 0x00008082,
	{0x40, 0x42, 0xFC, 0x84, 0x95, 0x17, 0x6D, 0x05, 0x62, 0x5E, 0xBF, 0xB5, 0x21, 0x79, 0xB5, 0x4F }
}, // 388

{
	0x00210444, 0x00010105,
	{0x55, 0xDF, 0x2F, 0xA9, 0x3A, 0xE2, 0x07, 0x6D, 0x22, 0x23, 0xB1, 0x84, 0xF2, 0xFC, 0xF6, 0x7D }
}, // 389

{
	0x00420889, 0x0002020A,
	{0xF9, 0xC9, 0xBC, 0xF3, 0xD5, 0x6A, 0xFD, 0x21, 0x28, 0xFF, 0x8F, 0x8D, 0x9F, 0x3B, 0x54, 0xD6 }
}, // 390

{
	0x00841112, 0x00040414,
	{0x7A, 0xA1, 0xDA, 0x13, 0xE4, 0x77, 0x4A, 0xE7, 0x57, 0xC4, 0x17, 0x95, 0x41, 0xEA, 0xC1, 0xED }
}, // 391

{
	0x01082224, 0x00080828,
	{0xEA, 0x8C, 0xF9, 0x42, 0x6C, 0x80, 0xB1, 0x4D, 0xD1, 0x62, 0xD1, 0x2D, 0x8D, 0x4C, 0x2A, 0x31 }
}, // 392

{
	0x02104449, 0x00101050,
	{0x9C, 0x12, 0xAE, 0x5D, 0x27, 0xD6, 0xF2, 0x50, 0xBB, 0x60, 0xFF, 0xA7, 0x45, 0xAA, 0xA1, 0xA3 }
}, // 393

{
	0x04208893, 0x002020A0,
	{0x54, 0x2F, 0xAB, 0x20, 0x48, 0x58, 0xE1, 0x24, 0x0A, 0x5F, 0xC1, 0xE4, 0xAE, 0x7C, 0x3F, 0xFC }
}, // 394

{
	0x08411126, 0x00404140,
	{0x28, 0xED, 0xCD, 0x90, 0x8D, 0x19, 0x0A, 0x2F, 0xB0, 0x03, 0xAD, 0x75, 0x1F, 0xD1, 0x43, 0x7D }
}, // 395

{
	0x1082224C, 0x00808280,
	{0x7B, 0x66, 0xB1, 0x79, 0x32, 0x46, 0xA1, 0x21, 0x50, 0x98, 0xD9, 0xEA, 0xEA, 0x1F, 0xD1, 0x44 }
}, // 396

{
	0x21044499, 0x01010500,
	{0x2C, 0x93, 0x99, 0x0D, 0xD7, 0x25, 0x5C, 0xF3, 0x1E, 0x98, 0x5F, 0xBB, 0xDE, 0x71, 0x87, 0xEC }
}, // 397

{
	0x42088932, 0x02020A00,
	{0x4E, 0xDE, 0xF1, 0xCF, 0xAB, 0x13, 0x21, 0xF5, 0xF5, 0xFD, 0xCF, 0xB5, 0x14, 0xCF, 0x5B, 0x26 }
}, // 398

{
	0x84111264, 0x04041400,
	{0x83, 0xF7, 0x8D, 0xB3, 0xAE, 0x7A, 0xEC, 0x06, 0xBA, 0xDD, 0xCD, 0x25, 0x06, 0x30, 0x0C, 0xB0 }
}, // 399

{
	0x082224C8, 0x08082801,
	{0x60, 0x0D, 0x0E, 0x64, 0xB7, 0x46, 0xC8, 0x4C, 0xD9, 0x38, 0x0B, 0x59, 0x12, 0xC4, 0x01, 0x0F }
}, // 400

{
	0x00000006, 0x80000000,
	{0x9B, 0xB4, 0x67, 0xD7, 0x9D, 0x9B, 0x1C, 0xD1, 0xB4, 0xC5, 0x48, 0x18, 0x38, 0x1A, 0x79, 0xCE }
}, // 401

{
	0x00000016, 0x80000000,
	{0x3B, 0x6D, 0x05, 0x66, 0x74, 0x0B, 0xCE, 0x27, 0x6A, 0x61, 0x1E, 0xA1, 0xC2, 0xF8, 0x52, 0x79 }
}, // 402

{
	0x00000026, 0x80000000,
	{0x28, 0x1C, 0xDF, 0xD3, 0x7D, 0xF3, 0x8E, 0xE4, 0x4B, 0xFA, 0x04, 0x6A, 0xA2, 0xA2, 0xFF, 0x60 }
}, // 403

{
	0x00000036, 0x80000000,
	{0x92, 0x47, 0x43, 0xB1, 0xAB, 0xE6, 0xEB, 0xE1, 0xBC, 0x91, 0xAA, 0xC1, 0x5A, 0x9F, 0x6B, 0xE5 }
}, // 404

{
	0x00000046, 0x80000000,
	{0xE0, 0xF7, 0xB8, 0x15, 0x5E, 0x62, 0x54, 0x01, 0xA1, 0x66, 0x07, 0x39, 0x81, 0xBD, 0x05, 0xAB }
}, // 405

{
	0x00002021, 0x00000000,
	{0x3C, 0xB9, 0xB8, 0x00, 0x5A, 0xAE, 0xFA, 0xDB, 0x7E, 0x6A, 0xFE, 0x74, 0x79, 0xFF, 0x24, 0x6D }
}, // 406

{
	0x00004042, 0x00000000,
	{0xC2, 0xB7, 0x0C, 0xB9, 0x4C, 0x83, 0x90, 0x25, 0x3E, 0x0D, 0x86, 0xEE, 0xCD, 0x1E, 0x33, 0xEB }
}, // 407

{
	0x00008084, 0x00000000,
	{0xA2, 0xA0, 0xC4, 0xF0, 0x14, 0x23, 0x7B, 0xF3, 0xEA, 0x0E, 0xF3, 0x20, 0xB3, 0x6D, 0x83, 0xF2 }
}, // 408

{
	0x00010109, 0x00000000,
	{0x3E, 0x48, 0x7A, 0xDA, 0xDF, 0x0E, 0x6B, 0x24, 0x37, 0xE3, 0xEB, 0xBD, 0x5D, 0xA3, 0x7B, 0x97 }
}, // 409

{
	0x00020212, 0x00000000,
	{0x9F, 0xCB, 0x77, 0x29, 0x5D, 0x5D, 0x1B, 0x3D, 0xB2, 0x52, 0x0D, 0x09, 0x31, 0x5F, 0x5E, 0x54 }
}, // 410

{
	0x00040424, 0x00000000,
	{0xD5, 0xD4, 0x83, 0x34, 0xBF, 0x3B, 0x69, 0x2A, 0x11, 0xA5, 0x0B, 0x4B, 0x6A, 0xFE, 0xF9, 0x37 }
}, // 411

{
	0x00080848, 0x00000000,
	{0xBB, 0xF8, 0x04, 0xDE, 0x94, 0x0F, 0xD4, 0x3E, 0x4E, 0x9F, 0x11, 0x02, 0x3C, 0x51, 0x67, 0x79 }
}, // 412

{
	0x00101090, 0x00000000,
	{0xDB, 0x95, 0x1C, 0x82, 0x54, 0xE0, 0xC3, 0xC0, 0x26, 0x63, 0x8B, 0x22, 0xEB, 0xFB, 0xBC, 0x2A }
}, // 413

{
	0x00202120, 0x00000000,
	{0x6A, 0xD2, 0x52, 0xA1, 0x91, 0xF8, 0xFB, 0xA3, 0x3C, 0xE9, 0xA4, 0x87, 0xE7, 0xC7, 0x28, 0xD9 }
}, // 414

{
	0x00404240, 0x00000000,
	{0xB3, 0x3F, 0xEB, 0x07, 0x8B, 0x2C, 0xE7, 0xF1, 0x45, 0x47, 0x6F, 0x13, 0xEA, 0xFA, 0x20, 0xD8 }
}, // 415

{
	0x00808480, 0x00000000,
	{0x8C, 0x84, 0xD7, 0x18, 0xCF, 0x9D, 0x23, 0x42, 0xFE, 0x33, 0xB2, 0x94, 0x29, 0x0D, 0x9B, 0xCB }
}, // 416

{
	0x01010900, 0x00000000,
	{0x38, 0x44, 0xB6, 0x85, 0xC9, 0xFE, 0xA0, 0xC1, 0xDD, 0xAE, 0xCA, 0x84, 0x80, 0x2A, 0x5D, 0x7C }
}, // 417

{
	0x02021200, 0x00000000,
	{0x27, 0x4D, 0x00, 0x9C, 0xD3, 0x41, 0xF8, 0x24, 0xA3, 0x7A, 0xD2, 0x57, 0x61, 0xB9, 0xF3, 0xA3 }
}, // 418

{
	0x04042400, 0x00000000,
	{0x05, 0x04, 0x24, 0xC9, 0x95, 0x8B, 0x69, 0x8E, 0x5A, 0xBD, 0x15, 0x22, 0xBC, 0x66, 0xA9, 0xDF }
}, // 419

{
	0x08084801, 0x00000000,
	{0xC1, 0xAB, 0xAE, 0x84, 0x55, 0xCF, 0xB5, 0x8A, 0xBC, 0xC7, 0x44, 0x8E, 0xEC, 0x32, 0x21, 0x73 }
}, // 420

{
	0x10109002, 0x00000000,
	{0xD8, 0x3C, 0x1E, 0xE9, 0xB7, 0x75, 0x6C, 0x11, 0x7D, 0xE2, 0xD3, 0x3F, 0xDE, 0x89, 0x37, 0xBF }
}, // 421

{
	0x20212004, 0x00000000,
	{0xEB, 0xE9, 0x62, 0xEB, 0x7E, 0x31, 0x31, 0x08, 0xE3, 0xFD, 0x4D, 0xDE, 0x2C, 0x1A, 0x4D, 0xA7 }
}, // 422

{
	0x40424008, 0x00000000,
	{0xAC, 0x6C, 0x01, 0xCB, 0x64, 0x27, 0x17, 0x29, 0x56, 0x33, 0xE5, 0x9D, 0x9D, 0xAF, 0x47, 0x49 }
}, // 423

{
	0x80848010, 0x00000000,
	{0x8C, 0x09, 0x92, 0x68, 0x69, 0x76, 0xD6, 0xC6, 0xDE, 0xCE, 0x7C, 0x41, 0x5F, 0x08, 0xB1, 0xEC }
}, // 424

{
	0x01090021, 0x00000001,
	{0xFB, 0xDA, 0xCD, 0xF4, 0x97, 0xA0, 0x4D, 0x8D, 0x20, 0xF2, 0xA0, 0x7B, 0xB2, 0x5A, 0x46, 0x92 }
}, // 425

{
	0x02120042, 0x00000002,
	{0x85, 0x38, 0x1A, 0x9A, 0x07, 0x4D, 0xD9, 0x6E, 0xF0, 0xEC, 0x00, 0x3E, 0xD3, 0x3D, 0x91, 0x3F }
}, // 426

{
	0x04240084, 0x00000004,
	{0x0B, 0x14, 0xD8, 0xA1, 0x93, 0xB6, 0x5D, 0x8D, 0x75, 0x5F, 0x23, 0x4D, 0xD7, 0x2D, 0xEA, 0x4D }
}, // 427

{
	0x08480108, 0x00000008,
	{0x3B, 0x03, 0xB1, 0xC6, 0xCB, 0x0E, 0x51, 0x57, 0xFC, 0xB0, 0xB7, 0x11, 0x64, 0xF5, 0x48, 0xBD }
}, // 428

{
	0x10900210, 0x00000010,
	{0xDC, 0x0B, 0x7B, 0x76, 0x08, 0x5F, 0x94, 0x60, 0xAC, 0xA8, 0x4D, 0x94, 0x9B, 0x1D, 0x43, 0x9A }
}, // 429

{
	0x21200420, 0x00000020,
	{0x7E, 0xEE, 0xAD, 0xF8, 0x50, 0x0B, 0x75, 0x00, 0xD6, 0x20, 0x6A, 0x27, 0x1E, 0x61, 0x75, 0x2B }
}, // 430

{
	0x42400840, 0x00000040,
	{0x5B, 0xAB, 0xAD, 0x85, 0x9F, 0x18, 0x43, 0xDB, 0x4A, 0x1C, 0x3C, 0xF5, 0x90, 0x15, 0x6D, 0x4B }
}, // 431

{
	0x84801080, 0x00000080,
	{0x27, 0xD6, 0x73, 0x72, 0xDD, 0xFD, 0xB3, 0xB6, 0x90, 0x9F, 0x49, 0x7C, 0x7B, 0xFB, 0x6D, 0xE6 }
}, // 432

{
	0x09002100, 0x00000101,
	{0x8C, 0xA4, 0xCC, 0xA4, 0x44, 0xA7, 0x37, 0xCB, 0x2E, 0x30, 0xAE, 0xD2, 0x03, 0x68, 0x13, 0x21 }
}, // 433

{
	0x12004200, 0x00000202,
	{0xD0, 0x05, 0xC5, 0x43, 0x9D, 0x27, 0x77, 0x03, 0x99, 0x24, 0xFE, 0x03, 0x3D, 0x5B, 0x5A, 0xA4 }
}, // 434

{
	0x24008400, 0x00000404,
	{0x63, 0x8B, 0x3C, 0x69, 0x00, 0x08, 0xEF, 0x8D, 0x49, 0x66, 0x01, 0xB0, 0x3D, 0xE6, 0xE0, 0xBA }
}, // 435

{
	0x48010801, 0x00000808,
	{0xC7, 0xEC, 0x0B, 0x0B, 0xF5, 0x1C, 0x9B, 0xD0, 0xF2, 0x38, 0x3D, 0x0C, 0xA0, 0x93, 0x15, 0x3E }
}, // 436

{
	0x90021002, 0x00001010,
	{0x78, 0xCB, 0xBA, 0x19, 0x26, 0xFE, 0xD5, 0x82, 0x8F, 0xE4, 0x55, 0x8A, 0x60, 0xC1, 0x92, 0x7B }
}, // 437

{
	0x20042004, 0x00002021,
	{0x70, 0x98, 0x0E, 0x09, 0xC8, 0x24, 0x7B, 0x9F, 0x11, 0x71, 0x70, 0x84, 0x48, 0x4C, 0x78, 0x9B }
}, // 438

{
	0x40084009, 0x00004042,
	{0xA7, 0xE1, 0x46, 0x1B, 0x3D, 0xA7, 0xF9, 0xA5, 0x2C, 0x32, 0x2D, 0x03, 0x9A, 0x53, 0x40, 0x82 }
}, // 439

{
	0x80108012, 0x00008084,
	{0x4C, 0x77, 0x16, 0x67, 0xC4, 0xF2, 0xF3, 0xF4, 0x91, 0xE4, 0xD1, 0xE1, 0x02, 0xB0, 0x20, 0x6D }
}, // 440

{
	0x00210025, 0x00010109,
	{0xB9, 0xBB, 0xC7, 0xE9, 0xAB, 0x3F, 0x36, 0x25, 0x27, 0x1E, 0x5E, 0x41, 0x79, 0x44, 0x63, 0x8B }
}, // 441

{
	0x0042004A, 0x00020212,
	{0xFE, 0x7E, 0xE6, 0x33, 0x82, 0x81, 0xFE, 0x07, 0x7B, 0x60, 0xC0, 0x99, 0x08, 0x47, 0x16, 0x5E }
}, // 442

{
	0x00840095, 0x00040424,
	{0x20, 0x6F, 0xA0, 0xFA, 0x11, 0x0B, 0x5D, 0x00, 0xA3, 0x0A, 0xA4, 0xF3, 0x7E, 0xD7, 0x3F, 0x0B }
}, // 443

{
	0x0108012A, 0x00080848,
	{0x41, 0xA2, 0x25, 0x36, 0x08, 0x46, 0x44, 0xBB, 0x54, 0xE5, 0x71, 0x7F, 0xF8, 0x55, 0x4F, 0x03 }
}, // 444

{
	0x02100254, 0x00101090,
	{0x5A, 0x4D, 0x58, 0x75, 0x87, 0x63, 0x4C, 0x9F, 0x20, 0x16, 0xBE, 0xEC, 0x3E, 0x56, 0x2E, 0x45 }
}, // 445

{
	0x042004A9, 0x00202120,
	{0x02, 0xEA, 0x9A, 0x3C, 0x07, 0x22, 0x38, 0xA7, 0xE0, 0xDD, 0x76, 0xFE, 0x19, 0x1E, 0x28, 0x8F }
}, // 446

{
	0x08400953, 0x00404240,
	{0xCD, 0xA9, 0x8D, 0x44, 0x52, 0x6F, 0xF4, 0x40, 0xE0, 0xCE, 0x95, 0xE7, 0x71, 0xE3, 0x54, 0xE7 }
}, // 447

{
	0x108012A7, 0x00808480,
	{0xB9, 0x9F, 0xC8, 0x30, 0x12, 0x17, 0x1B, 0x78, 0xC7, 0x63, 0xA2, 0x85, 0xEF, 0xB3, 0xF5, 0xD1 }
}, // 448

{
	0x2100254E, 0x01010900,
	{0x16, 0x66, 0x4E, 0x83, 0x31, 0x24, 0xEB, 0x60, 0xED, 0xFF, 0x99, 0x80, 0xDA, 0x59, 0x93, 0x1E }
}, // 449

{
	0x42004A9C, 0x02021200,
	{0x70, 0x44, 0x87, 0x43, 0x75, 0x4C, 0x8B, 0x2A, 0x7A, 0x39, 0xA6, 0xE4, 0x87, 0x5C, 0x13, 0xA5 }
}, // 450

{
	0x84009539, 0x04042400,
	{0xF1, 0x47, 0xBB, 0x55, 0x27, 0x9A, 0xAF, 0x4D, 0x20, 0x79, 0x32, 0x28, 0xD8, 0x9E, 0xA7, 0x56 }
}, // 451

{
	0x08012A72, 0x08084801,
	{0x5B, 0x1A, 0xBE, 0xFE, 0xFF, 0x3C, 0xCC, 0xEF, 0xB4, 0x9F, 0x27, 0x14, 0x62, 0xFD, 0x15, 0x77 }
}, // 452

{
	0x00000006, 0x80000000,
	{0x9B, 0xB4, 0x67, 0xD7, 0x9D, 0x9B, 0x1C, 0xD1, 0xB4, 0xC5, 0x48, 0x18, 0x38, 0x1A, 0x79, 0xCE }
},// 453

{
	0x00000006, 0x80000000,
	{0x9B, 0xB4, 0x67, 0xD7, 0x9D, 0x9B, 0x1C, 0xD1, 0xB4, 0xC5, 0x48, 0x18, 0x38, 0x1A, 0x79, 0xCE }
},// 454

{
	0x05002080, 0x00000101,
	{0xB9, 0x3C, 0xB4, 0xF4, 0xAE, 0x60, 0xA0, 0x78, 0x40, 0x03, 0xF3, 0xB6, 0x1B, 0x65, 0xE9, 0x66 }
}, // 455

{
	0x0A004101, 0x00000202,
	{0x93, 0x68, 0x47, 0xAA, 0xBE, 0x0F, 0xA9, 0xAE, 0xDD, 0xF7, 0x1B, 0x92, 0xDA, 0x3D, 0xF8, 0x7C }
}, // 456

{
	0x14008202, 0x00000404,
	{0x68, 0xC5, 0x76, 0x63, 0x2F, 0x3B, 0xD8, 0x41, 0x27, 0x51, 0x0D, 0x45, 0xB2, 0x32, 0x87, 0x1F }
}, // 457

{
	0x28010404, 0x00000808,
	{0x57, 0x88, 0xAC, 0xB2, 0xE8, 0xB3, 0xB8, 0xFF, 0x21, 0xC5, 0x76, 0x7F, 0xA4, 0xB9, 0xFC, 0x33 }
}, // 458

{
	0x50020809, 0x00001010,
	{0x9D, 0xAA, 0x75, 0xDE, 0xF2, 0x11, 0xAA, 0x3F, 0xC2, 0x93, 0x54, 0x59, 0x0F, 0xBA, 0xC2, 0x45 }
}, // 459

{
	0xA0041013, 0x00002020,
	{0x1E, 0x04, 0xE8, 0x63, 0xFF, 0xBD, 0x3C, 0x87, 0x74, 0x66, 0xBA, 0xE6, 0x6D, 0xAC, 0x8B, 0x88 }
}, // 460

{
	0x40082027, 0x00004041,
	{0x4E, 0xD0, 0xE6, 0xC1, 0x34, 0x60, 0xDA, 0xC0, 0x88, 0x80, 0xB3, 0x05, 0x4A, 0x02, 0x7B, 0x32 }
}, // 461

{
	0x8010404E, 0x00008082,
	{0x33, 0xC2, 0xAD, 0x8A, 0x1F, 0x9A, 0x30, 0xFB, 0x54, 0x08, 0xB1, 0x0B, 0x03, 0x58, 0x38, 0x76 }
}, // 462

{
	0x0020809D, 0x00010105,
	{0x36, 0xC7, 0x3E, 0x8D, 0x07, 0x0B, 0x90, 0x8C, 0x4D, 0x96, 0x8D, 0x85, 0x5E, 0xB3, 0x55, 0xE7 }
}, // 463

{
	0x0041013A, 0x0002020A,
	{0x70, 0xCF, 0x78, 0xC5, 0xE4, 0xDA, 0xC0, 0x51, 0x93, 0xB2, 0x0B, 0xB2, 0xC6, 0x1C, 0x00, 0x3C }
}, // 464

{
	0x00820275, 0x00040414,
	{0x28, 0xBF, 0xA7, 0xC1, 0x50, 0x5B, 0xF3, 0x0B, 0x6E, 0x1D, 0xAD, 0xB5, 0x86, 0xFD, 0xA8, 0xF1 }
}, // 465

{
	0x010404EB, 0x00080828,
	{0x5C, 0x5B, 0xD7, 0x67, 0x12, 0x9C, 0xED, 0x6A, 0x96, 0xD6, 0x75, 0x08, 0x98, 0x52, 0xFB, 0x84 }
}, // 466

{
	0x020809D7, 0x00101050,
	{0xC2, 0x97, 0x98, 0xE8, 0xA5, 0x8C, 0xFC, 0xDC, 0x01, 0x31, 0x48, 0x0D, 0x6A, 0xDC, 0xB9, 0xA7 }
}, // 467

{
	0x041013AF, 0x002020A0,
	{0x0A, 0xC7, 0x20, 0x94, 0xB8, 0xE3, 0xDC, 0x1B, 0x14, 0x64, 0x39, 0x66, 0x1A, 0x61, 0x4E, 0xE1 }
}, // 468

{
	0x0820275E, 0x00404140,
	{0xD2, 0x8D, 0x74, 0xC8, 0x68, 0x5B, 0xB5, 0xC8, 0x5A, 0xFE, 0xC9, 0x76, 0xD5, 0x9C, 0xC0, 0x47 }
}, // 469

{
	0x10404EBC, 0x00808280,
	{0xF2, 0x5D, 0xF5, 0xDF, 0x0A, 0xE9, 0xD6, 0xF6, 0x29, 0xEC, 0x03, 0x2D, 0xB9, 0xF1, 0xEE, 0xD6 }
}, // 470

{
	0x20809D78, 0x01010500,
	{0x89, 0x92, 0x4B, 0xF6, 0xF9, 0x48, 0x76, 0x53, 0xE4, 0xCE, 0x5C, 0x47, 0xE8, 0xB9, 0xAB, 0xAF }
}, // 471

{
	0x41013AF1, 0x02020A00,
	{0xC8, 0xE8, 0x73, 0xC1, 0x04, 0xA7, 0xDD, 0x0A, 0x40, 0x9F, 0x43, 0xAB, 0xB9, 0x60, 0x3C, 0x68 }
}, // 472

{
	0x820275E3, 0x04041400,
	{0x86, 0x14, 0xB2, 0x94, 0x1B, 0x99, 0x32, 0x48, 0xD4, 0x64, 0xFD, 0xA4, 0x57, 0xA5, 0x3D, 0x1B }
}, // 473

{
	0x0404EBC7, 0x08082801,
	{0x09, 0x7C, 0x72, 0x3B, 0x48, 0xCA, 0xAD, 0x36, 0x78, 0xCB, 0x6F, 0x5C, 0x0F, 0x9E, 0x81, 0xC7 }
}, // 474

{
	0x82802080, 0x00000080,
	{0x34, 0x87, 0xA2, 0x01, 0x47, 0x82, 0x02, 0xD5, 0xC1, 0x82, 0x37, 0x50, 0x37, 0x59, 0xA6, 0x63 }
}, // 475

{
	0x05004100, 0x00000101,
	{0x5F, 0x18, 0xBF, 0xF6, 0x74, 0xD3, 0x14, 0x7B, 0xC2, 0xFE, 0xDB, 0x0F, 0xA2, 0x74, 0xB0, 0xD2 }
}, // 476

{
	0x0A008200, 0x00000202,
	{0x09, 0x9E, 0x40, 0x19, 0xD4, 0xE1, 0xE9, 0xBF, 0x5D, 0xCE, 0x9A, 0x94, 0x26, 0xE8, 0xC4, 0x48 }
}, // 477

{
	0x14010400, 0x00000404,
	{0x64, 0x2A, 0x4A, 0xFA, 0x4A, 0x5F, 0x96, 0x65, 0x20, 0xA0, 0x40, 0x04, 0x42, 0xE1, 0x7C, 0x2D }
}, // 478

{
	0x28020801, 0x00000808,
	{0xD2, 0x77, 0x05, 0xF3, 0xEA, 0x70, 0x03, 0x2E, 0x12, 0x56, 0x91, 0xCB, 0x87, 0x81, 0xD6, 0x65 }
}, // 479

{
	0x50041003, 0x00001010,
	{0xD4, 0x55, 0xDB, 0xCB, 0x9F, 0x04, 0xFA, 0x20, 0x92, 0xEB, 0x67, 0x0E, 0xC8, 0xC4, 0xA3, 0x8D }
}, // 480

{
	0xA0082006, 0x00002020,
	{0x2F, 0x41, 0x0E, 0x61, 0x3C, 0x38, 0xF8, 0xFC, 0xC8, 0xBC, 0x06, 0x22, 0x7E, 0x1E, 0x2D, 0x15 }
}, // 481

{
	0x4010400C, 0x00004041,
	{0xD3, 0xA5, 0x64, 0x7C, 0xC8, 0xC9, 0x90, 0xAC, 0xD1, 0xD0, 0x8E, 0x80, 0x43, 0x9F, 0x9B, 0xFC }
}, // 482

{
	0x80208018, 0x00008082,
	{0x46, 0x84, 0xBF, 0x5A, 0xFD, 0x76, 0xB6, 0x17, 0x41, 0xE3, 0x3B, 0xC2, 0x99, 0x81, 0x29, 0x57 }
}, // 483

{
	0x00410030, 0x00010105,
	{0xB8, 0x86, 0x5D, 0xB2, 0x59, 0x20, 0x3F, 0x3D, 0x83, 0x49, 0x16, 0x84, 0x53, 0x66, 0x22, 0xA3 }
}, // 484

{
	0x00820060, 0x0002020A,
	{0xC4, 0xC6, 0x40, 0xFB, 0xBB, 0x22, 0xBB, 0xE5, 0x3A, 0xF0, 0x0D, 0xC3, 0xB8, 0xBF, 0x21, 0xE9 }
}, // 485

{
	0x010400C0, 0x00040414,
	{0xDF, 0x94, 0x77, 0xBF, 0x87, 0x7D, 0x22, 0x95, 0x03, 0x70, 0x85, 0xF0, 0x27, 0xDB, 0x75, 0x8C }
}, // 486

{
	0x02080180, 0x00080828,
	{0x2D, 0x6F, 0xA5, 0x5E, 0xA9, 0xEE, 0x54, 0xD6, 0xC3, 0x34, 0x3F, 0xE2, 0x31, 0xAC, 0x53, 0xEA }
}, // 487

{
	0x04100300, 0x00101050,
	{0x5F, 0xDA, 0xEB, 0xA7, 0x7C, 0xD3, 0x88, 0xDE, 0x0B, 0xBB, 0xD7, 0x4A, 0x62, 0x3D, 0x50, 0x36 }
}, // 488

{
	0x08200600, 0x002020A0,
	{0x7A, 0x78, 0xDD, 0x16, 0xDA, 0x90, 0x99, 0xDA, 0x33, 0x0F, 0x65, 0xBE, 0xF2, 0x8A, 0xC6, 0x01 }
}, // 489

{
	0x10400C00, 0x00404140,
	{0x06, 0x5F, 0xA7, 0xF1, 0x29, 0x12, 0xC1, 0x80, 0x4B, 0xAC, 0xE7, 0x17, 0x5E, 0x6E, 0x36, 0x68 }
}, // 490

{
	0x20801800, 0x00808280,
	{0x7D, 0x1E, 0x97, 0xE2, 0x13, 0x6B, 0x07, 0xF4, 0x4C, 0x40, 0xEA, 0x48, 0x59, 0x85, 0x85, 0x0B }
}, // 491

{
	0x41003000, 0x01010500,
	{0x8E, 0xAF, 0xB3, 0x61, 0x3F, 0x87, 0xB3, 0xBD, 0xC8, 0x0A, 0xA9, 0xAC, 0x6D, 0x6E, 0x3E, 0x19 }
}, // 492

{
	0x82006000, 0x02020A00,
	{0x70, 0xFD, 0x36, 0xE9, 0x56, 0x14, 0x05, 0x65, 0x32, 0xD5, 0x5A, 0x08, 0xD6, 0xD4, 0xF5, 0x22 }
}, // 493

{
	0x0400C001, 0x04041401,
	{0xBE, 0x21, 0xDB, 0xC6, 0x12, 0xE7, 0x76, 0x34, 0x24, 0x3D, 0xD2, 0xA4, 0x56, 0x63, 0xCE, 0x50 }
}, // 494

{
	0x08018002, 0x08082802,
	{0xAB, 0x0A, 0x4C, 0x03, 0xA7, 0x7F, 0x3E, 0xD6, 0x75, 0xB3, 0x78, 0x10, 0x60, 0xDD, 0xC7, 0xD7 }
}, // 495

{
	0x04041001, 0x00000000,
	{0xA5, 0xA7, 0x3A, 0x0F, 0xD7, 0x22, 0x9A, 0x91, 0xFC, 0xFC, 0xD3, 0xA5, 0x6D, 0x90, 0xF1, 0xCF }
}, // 496

{
	0x08082003, 0x00000000,
	{0xFD, 0x71, 0x1D, 0xA8, 0xA7, 0x92, 0x93, 0x26, 0xD4, 0x2A, 0x1B, 0x68, 0x76, 0x86, 0xE8, 0x38 }
}, // 497

{
	0x10104006, 0x00000000,
	{0x5E, 0x9F, 0x35, 0x32, 0xED, 0x97, 0xDE, 0x9F, 0xA1, 0x9D, 0xB8, 0xF7, 0x9F, 0x74, 0xAB, 0xFA }
}, // 498

{
	0x2020800C, 0x00000000,
	{0xA9, 0x64, 0x4C, 0xA6, 0x68, 0x1B, 0x89, 0x9B, 0x93, 0xFF, 0x4D, 0xFF, 0x66, 0xEA, 0x9E, 0xC6 }
}, // 499

{
	0x40410018, 0x00000000,
	{0x85, 0x2B, 0x92, 0x8E, 0xE7, 0x9D, 0x3B, 0x14, 0x47, 0x33, 0x52, 0xE2, 0x87, 0x9D, 0xC5, 0xD7 }
}, // 500

{
	0x80820030, 0x00000000,
	{0x15, 0x96, 0xF2, 0xF6, 0xA9, 0x8B, 0xA8, 0xE4, 0x0A, 0x7B, 0x0C, 0x35, 0x73, 0x9F, 0xDE, 0x3C }
}, // 501

{
	0x01040061, 0x00000001,
	{0x27, 0xEC, 0xC3, 0xB7, 0x85, 0xAC, 0x6B, 0xF0, 0x03, 0xD2, 0x39, 0x0D, 0xE0, 0xDF, 0x38, 0x91 }
}, // 502

{
	0x020800C3, 0x00000002,
	{0xC7, 0x84, 0xA9, 0xC6, 0x62, 0x40, 0x5E, 0x15, 0xCB, 0x9D, 0xEA, 0xDF, 0x2A, 0xBB, 0x9F, 0x99 }
}, // 503

{
	0x04100186, 0x00000004,
	{0xEB, 0x8B, 0xA7, 0x5A, 0x74, 0x60, 0x25, 0x1C, 0xA3, 0xC5, 0x26, 0x29, 0xFF, 0xFE, 0x90, 0xA7 }
}, // 504

{
	0x0820030C, 0x00000008,
	{0x15, 0x41, 0x02, 0x0D, 0x88, 0xA6, 0x62, 0xB4, 0x95, 0xCE, 0xB0, 0x17, 0xBE, 0x3B, 0x04, 0xA9 }
}, // 505

{
	0x10400618, 0x00000010,
	{0x73, 0x86, 0x7D, 0x62, 0xC8, 0x81, 0xD0, 0x94, 0x32, 0x56, 0xB8, 0xE6, 0x4A, 0x76, 0x9F, 0xE2 }
}, // 506

{
	0x20800C30, 0x00000020,
	{0xE6, 0x2E, 0x20, 0x47, 0x69, 0xF9, 0x72, 0x47, 0x2D, 0x6F, 0xF9, 0xB5, 0x18, 0x58, 0x7A, 0xA4 }
}, // 507

{
	0x41001860, 0x00000040,
	{0x94, 0xF7, 0xEB, 0x0D, 0xFE, 0x7D, 0xF2, 0x35, 0x3E, 0xC6, 0xF8, 0xBB, 0x99, 0x7F, 0x56, 0xB3 }
}, // 508

{
	0x820030C1, 0x00000080,
	{0x8C, 0x51, 0xCE, 0x7F, 0x3C, 0xC2, 0xE8, 0xF6, 0x5E, 0xFE, 0x29, 0xEA, 0x04, 0xFB, 0x16, 0x92 }
}, // 509

{
	0x04006182, 0x00000101,
	{0xF7, 0x4A, 0xA9, 0x3C, 0xC9, 0x29, 0xE0, 0x8F, 0xE9, 0xC5, 0x5F, 0x9C, 0x84, 0xE5, 0x77, 0x43 }
}, // 510

{
	0x0800C304, 0x00000202,
	{0xE2, 0x04, 0xAF, 0xE0, 0x16, 0x21, 0x46, 0x8D, 0x1C, 0x9F, 0x66, 0xF3, 0x24, 0x4B, 0xE8, 0xCC }
}, // 511

{
	0x10018609, 0x00000404,
	{0xA5, 0xC5, 0x37, 0x52, 0xE2, 0xAE, 0xFB, 0x4A, 0xAF, 0x4F, 0x5B, 0x40, 0x78, 0x66, 0x04, 0x32 }
}, // 512

{
	0x20030C12, 0x00000808,
	{0x35, 0xDB, 0x41, 0x96, 0x46, 0xC3, 0x5A, 0xA1, 0x74, 0x8E, 0x64, 0x52, 0x9E, 0x90, 0x41, 0x49 }
}, // 513

{
	0x40061824, 0x00001010,
	{0xEA, 0xD8, 0xD2, 0x13, 0x68, 0x87, 0x84, 0xD5, 0x83, 0xC1, 0xBA, 0x84, 0x21, 0x39, 0xB0, 0x1A }
}, // 514

{
	0x800C3049, 0x00002020,
	{0xBF, 0xA3, 0xBF, 0xF4, 0x77, 0x5C, 0x9E, 0x66, 0xF2, 0xF4, 0x50, 0x89, 0xD6, 0xCF, 0xCB, 0xCC }
}, // 515

{
	0x00186093, 0x00004041,
	{0x8E, 0xBB, 0x73, 0x94, 0x7B, 0x91, 0xEA, 0x57, 0xE4, 0xA2, 0x45, 0x2E, 0x1D, 0x67, 0xB3, 0x4C }
}, // 516

{
	0x0030C127, 0x00008082,
	{0x22, 0x37, 0xE4, 0x0E, 0x4D, 0x22, 0x48, 0x09, 0x73, 0x5B, 0xF9, 0xD5, 0xE4, 0xA2, 0xCE, 0x15 }
}, // 517

{
	0x0061824F, 0x00010104,
	{0xA0, 0xF8, 0xDF, 0x8A, 0x66, 0x41, 0x65, 0xA2, 0x70, 0xFD, 0x06, 0xE8, 0xCE, 0x6B, 0x65, 0x40 }
}, // 518

{
	0x00C3049E, 0x00020208,
	{0x35, 0x5E, 0x13, 0x9D, 0x27, 0x71, 0x90, 0x58, 0x2C, 0x29, 0xDC, 0xED, 0x78, 0xC5, 0xD6, 0x95 }
}, // 519

{
	0x0186093C, 0x00040410,
	{0xDD, 0x00, 0x1E, 0x73, 0x6E, 0x0A, 0x95, 0xE7, 0xF7, 0xE0, 0x68, 0x24, 0x3B, 0x70, 0x29, 0x61 }
}, // 520

{
	0x030C1278, 0x00080820,
	{0xAA, 0xC5, 0xE8, 0x30, 0xCD, 0xA7, 0x8F, 0xED, 0x0D, 0x78, 0x32, 0xF7, 0x0A, 0x3E, 0x03, 0x88 }
}, // 521

{
	0x061824F1, 0x00101040,
	{0xD4, 0x2D, 0x76, 0x9C, 0xE7, 0x87, 0x9B, 0x85, 0x92, 0xDB, 0xBD, 0x10, 0x80, 0xF8, 0x93, 0x13 }
}, // 522

{
	0x0C3049E2, 0x00202080,
	{0xD3, 0xBF, 0xDA, 0xC8, 0xFE, 0x3B, 0xB1, 0xA3, 0x26, 0x72, 0xB5, 0xCA, 0xFD, 0xEF, 0xB2, 0xD3 }
}, // 523

{
	0x186093C4, 0x00404100,
	{0x3F, 0xE8, 0x9E, 0x4D, 0x6C, 0x2A, 0xEE, 0x4D, 0xF4, 0xBC, 0xD0, 0x0C, 0xA7, 0xCE, 0xF5, 0xF9 }
}, // 524

{
	0x30C12789, 0x00808200,
	{0xD4, 0xDF, 0x2B, 0x7E, 0xEB, 0xEE, 0x1A, 0xDC, 0x77, 0x8B, 0x61, 0xFF, 0x42, 0x1E, 0x2B, 0x8B }
}, // 525

{
	0x61824F13, 0x01010400,
	{0x3D, 0x66, 0x75, 0x8A, 0xBC, 0x13, 0x58, 0xC0, 0x0E, 0xD9, 0x02, 0x02, 0x24, 0x11, 0x2D, 0x72 }
}, // 526

{
	0xC3049E26, 0x02020800,
	{0x6D, 0x11, 0xF4, 0x19, 0xD4, 0x03, 0x8C, 0xF9, 0x3B, 0x82, 0x97, 0xE7, 0x5D, 0xDE, 0x71, 0x22 }
}, // 527

{
	0x86093C4C, 0x04041001,
	{0x73, 0xA3, 0x12, 0xA3, 0xDA, 0xFB, 0xD7, 0xD6, 0x6B, 0xD7, 0x8B, 0x30, 0xA1, 0x0E, 0xAE, 0x07 }
}, // 528

{
	0x0C127898, 0x08082003,
	{0x5B, 0x72, 0xFF, 0xA8, 0x3A, 0xDC, 0x32, 0x13, 0x49, 0x97, 0x07, 0xDF, 0x1C, 0xF6, 0xE3, 0xD5 }
}, // 529

{
	0x41001861, 0x00000040,
	{0x14, 0xE3, 0x88, 0x2F, 0xA5, 0x4E, 0xF4, 0xBA, 0x9B, 0xC2, 0x36, 0x55, 0x74, 0xBC, 0x84, 0x15 }
}, // 530

{
	0x820030C3, 0x00000080,
	{0x81, 0xDC, 0x65, 0xD6, 0xD5, 0xD8, 0x91, 0xFD, 0xE9, 0x00, 0x68, 0x7E, 0xFF, 0x41, 0x1D, 0x13 }
}, // 531

{
	0x04006187, 0x00000101,
	{0x53, 0x00, 0x35, 0xCF, 0x6A, 0x44, 0x0C, 0xF5, 0xCC, 0xC8, 0xDE, 0x6C, 0x49, 0x12, 0xC0, 0x18 }
}, // 532

{
	0x0800C30F, 0x00000202,
	{0x73, 0x74, 0x90, 0xA9, 0x8C, 0xFF, 0x75, 0x5B, 0x73, 0x18, 0x7F, 0x31, 0x78, 0x9C, 0xB0, 0xD8 }
}, // 533

{
	0x1001861F, 0x00000404,
	{0x73, 0x28, 0x2B, 0x6C, 0xF4, 0xBE, 0xC3, 0x42, 0xF7, 0x88, 0xFC, 0x10, 0x8C, 0xEA, 0x1E, 0x39 }
}, // 534

{
	0x20030C3E, 0x00000808,
	{0xB1, 0xAE, 0x98, 0x70, 0xAB, 0x6D, 0xEA, 0x50, 0x7D, 0x22, 0x66, 0x5D, 0x82, 0x42, 0x2D, 0xFC }
}, // 535

{
	0x4006187C, 0x00001010,
	{0x38, 0xBC, 0x6E, 0x2B, 0x0F, 0x03, 0x86, 0x47, 0xDB, 0x92, 0x95, 0xFD, 0x76, 0x2A, 0xBE, 0x7A }
}, // 536

{
	0x800C30F8, 0x00002020,
	{0x1C, 0x6C, 0x3B, 0x97, 0x3F, 0xBD, 0x63, 0xAB, 0xF5, 0xE2, 0xC9, 0x49, 0xC8, 0x9D, 0x3C, 0x05 }
}, // 537

{
	0x001861F0, 0x00004041,
	{0x32, 0x1B, 0xB0, 0xE2, 0x3A, 0x87, 0xD7, 0x0E, 0xCC, 0x76, 0x2A, 0x53, 0xFA, 0x6B, 0x98, 0xCB }
}, // 538

{
	0x0030C3E1, 0x00008082,
	{0xDF, 0x69, 0xD4, 0x06, 0x80, 0x28, 0xD2, 0xF9, 0x6F, 0x5C, 0x9F, 0x87, 0x5E, 0xCD, 0xA1, 0x5C }
}, // 539

{
	0x006187C3, 0x00010104,
	{0x61, 0x34, 0xC7, 0x98, 0x0A, 0x5A, 0xAA, 0x18, 0xF0, 0xDA, 0xF1, 0x20, 0x99, 0xB2, 0x21, 0xE7 }
}, // 540

{
	0x00C30F87, 0x00020208,
	{0x21, 0x64, 0x30, 0x78, 0x11, 0x8C, 0xD9, 0x3A, 0x48, 0xF0, 0x18, 0x55, 0x9D, 0x2C, 0x7B, 0x6A }
}, // 541

{
	0x01861F0E, 0x00040410,
	{0xCC, 0x6D, 0x0D, 0xA8, 0x6B, 0x1C, 0x53, 0x4E, 0xF6, 0x35, 0xFC, 0x6E, 0x68, 0x48, 0xB4, 0x4B }
}, // 542

{
	0x030C3E1D, 0x00080820,
	{0x54, 0x00, 0x92, 0xD6, 0xD5, 0xC5, 0x82, 0xB7, 0x61, 0xF1, 0x20, 0xE1, 0x7B, 0x13, 0x45, 0xB0 }
}, // 543

{
	0x06187C3A, 0x00101040,
	{0x87, 0x56, 0x29, 0x25, 0x8C, 0xC9, 0xA0, 0x78, 0x86, 0xD2, 0x46, 0x1E, 0x1C, 0x20, 0xF3, 0x95 }
}, // 544

{
	0x0C30F875, 0x00202080,
	{0x23, 0x8C, 0xB0, 0x2D, 0x93, 0x6C, 0xC3, 0xF6, 0x20, 0xA6, 0xBE, 0x71, 0xE3, 0xB3, 0x4E, 0xC4 }
}, // 545

{
	0x1861F0EA, 0x00404100,
	{0x6F, 0xF0, 0x67, 0x10, 0x75, 0x96, 0x87, 0x57, 0x84, 0x43, 0x31, 0xA3, 0x0C, 0x85, 0xC8, 0x98 }
}, // 546

{
	0x30C3E1D4, 0x00808200,
	{0xBE, 0xD3, 0x43, 0xA0, 0x64, 0x62, 0x9B, 0x24, 0x3C, 0xBE, 0x89, 0x37, 0x21, 0x0C, 0x97, 0xFD }
}, // 547

{
	0x6187C3A9, 0x01010400,
	{0x5C, 0x2B, 0x5E, 0x1D, 0x58, 0xA8, 0xF5, 0x0D, 0x14, 0x9B, 0x31, 0xB7, 0x81, 0x50, 0x66, 0x18 }
}, // 548

{
	0xC30F8753, 0x02020800,
	{0x78, 0x4C, 0xFC, 0xD9, 0x13, 0x25, 0xC2, 0x7E, 0xD2, 0x4B, 0xE1, 0x95, 0x18, 0x43, 0xFA, 0xF6 }
}, // 549

{
	0x861F0EA7, 0x04041001,
	{0x8A, 0x65, 0xCF, 0xA9, 0x38, 0x4E, 0x57, 0x74, 0x1C, 0x67, 0xDE, 0x6F, 0x33, 0xD6, 0x77, 0x41 }
}, // 550

{
	0x0C3E1D4E, 0x08082003,
	{0x45, 0x77, 0x40, 0xDA, 0x9B, 0xFA, 0xE2, 0x7E, 0x12, 0xD0, 0xF7, 0xEF, 0xE1, 0xDF, 0x6C, 0x75 }
}, // 551

{
	0x40C01861, 0x00000040,
	{0xDF, 0x74, 0x6E, 0xE9, 0xED, 0xA0, 0xC8, 0x2C, 0x1C, 0xE2, 0x67, 0x0A, 0x26, 0x47, 0x42, 0x14 }
}, // 552

{
	0x818030C3, 0x00000080,
	{0x9C, 0x93, 0xC1, 0xDE, 0x0C, 0x49, 0x71, 0x9B, 0xD6, 0x1A, 0x24, 0xCC, 0x98, 0x15, 0x16, 0x65 }
}, // 553

{
	0x03006186, 0x00000101,
	{0x2A, 0x6B, 0x90, 0xBC, 0xD3, 0xCD, 0x18, 0x1F, 0xF2, 0x2F, 0xF5, 0x72, 0x71, 0xC8, 0xB2, 0x02 }
}, // 554

{
	0x0600C30C, 0x00000202,
	{0x33, 0xDF, 0xC3, 0x2B, 0x61, 0x49, 0x8D, 0xE3, 0x18, 0xE3, 0x4D, 0x6B, 0x9B, 0x5C, 0xD8, 0xD6 }
}, // 555

{
	0x0C018618, 0x00000404,
	{0x91, 0x73, 0xB8, 0xF9, 0xD3, 0xB0, 0x35, 0xF7, 0x8E, 0x91, 0x57, 0x8B, 0x91, 0x82, 0x4F, 0x43 }
}, // 556

{
	0x18030C31, 0x00000808,
	{0xEE, 0x9D, 0x50, 0x20, 0x73, 0xD4, 0xAD, 0xCF, 0x7A, 0xAA, 0xF8, 0xA9, 0x70, 0x25, 0x11, 0xC6 }
}, // 557

{
	0x30061863, 0x00001010,
	{0x2D, 0x42, 0xF4, 0xDE, 0x2A, 0xA3, 0x27, 0x69, 0xDC, 0x68, 0x4D, 0xA8, 0xFA, 0x78, 0x20, 0x25 }
}, // 558

{
	0x600C30C6, 0x00002020,
	{0x13, 0xAA, 0x5F, 0x70, 0x7F, 0x70, 0xD1, 0x0C, 0x31, 0xF9, 0x76, 0x8C, 0x62, 0xC4, 0xE0, 0x00 }
}, // 559

{
	0xC018618D, 0x00004040,
	{0x6A, 0x6A, 0x32, 0xEA, 0xB9, 0x30, 0x42, 0x3F, 0xAF, 0x91, 0x8A, 0xEB, 0x8B, 0x8C, 0xD4, 0x14 }
}, // 560

{
	0x8030C31B, 0x00008081,
	{0x08, 0xFC, 0xC5, 0xE1, 0x4D, 0x23, 0x6E, 0x74, 0x64, 0x04, 0x9E, 0x1C, 0xBD, 0x2F, 0x05, 0xE9 }
}, // 561

{
	0x00618637, 0x00010103,
	{0x6A, 0x96, 0xF4, 0x4B, 0x99, 0xA2, 0xA8, 0xC7, 0xBC, 0x76, 0x21, 0xE7, 0x14, 0x50, 0x6A, 0xEC }
}, // 562

{
	0x00C30C6E, 0x00020206,
	{0x8F, 0x45, 0x3B, 0xA9, 0x03, 0x5C, 0x86, 0xF7, 0x55, 0x02, 0xA3, 0x08, 0xED, 0x5F, 0xF3, 0x3A }
}, // 563

{
	0x018618DC, 0x0004040C,
	{0x83, 0x40, 0x83, 0x1C, 0x45, 0x14, 0x5B, 0xB7, 0xF8, 0xEE, 0x00, 0x72, 0x00, 0xD7, 0x59, 0x35 }
}, // 564

{
	0x030C31B8, 0x00080818,
	{0xF9, 0xA7, 0x13, 0x23, 0x9C, 0x9E, 0x39, 0x0F, 0x1A, 0x0F, 0x4D, 0x10, 0x2F, 0x81, 0xFE, 0xA1 }
}, // 565

{
	0x06186371, 0x00101030,
	{0x06, 0x84, 0x8F, 0xC9, 0x52, 0xDB, 0x6A, 0x24, 0x66, 0x57, 0x5B, 0x5E, 0xA1, 0x4F, 0xBC, 0x66 }
}, // 566

{
	0x0C30C6E2, 0x00202060,
	{0x0C, 0xAF, 0x87, 0x0B, 0xC1, 0x12, 0x5B, 0x0A, 0x71, 0x8F, 0x33, 0xB0, 0x13, 0x06, 0x9E, 0xA2 }
}, // 567

{
	0x18618DC5, 0x004040C0,
	{0x0E, 0xC1, 0xA0, 0x82, 0x3D, 0x71, 0x1F, 0xB2, 0x32, 0x91, 0x5B, 0xC8, 0xF1, 0x41, 0x6B, 0x7A }
}, // 568

{
	0x30C31B8A, 0x00808180,
	{0x1E, 0x35, 0x32, 0x4D, 0x0C, 0xB3, 0xF5, 0xF2, 0xFD, 0x79, 0x52, 0x09, 0x9A, 0xA1, 0x15, 0xD3 }
}, // 569

{
	0x61863715, 0x01010300,
	{0xF8, 0xD6, 0xC0, 0x78, 0x19, 0x46, 0x8A, 0x54, 0x46, 0xCA, 0x58, 0x99, 0xAF, 0xD2, 0x90, 0xEE }
}, // 570

{
	0xC30C6E2A, 0x02020600,
	{0xA5, 0x39, 0x74, 0x2D, 0xA0, 0x65, 0x5D, 0xEF, 0x3B, 0x66, 0xE8, 0x8A, 0x80, 0xA4, 0x54, 0x4B }
}, // 571

{
	0x8618DC54, 0x04040C01,
	{0xB8, 0xCE, 0x26, 0xFB, 0x0D, 0x65, 0xDB, 0x1D, 0x5C, 0x85, 0x81, 0x28, 0x16, 0x5C, 0x6A, 0x00 }
}, // 572

{
	0x0C31B8A9, 0x08081803,
	{0xE3, 0x1E, 0xAC, 0x47, 0xBC, 0xC5, 0xB2, 0xF0, 0x04, 0x15, 0xC2, 0xAC, 0xE4, 0xDC, 0x7B, 0x6D }
}, // 573

{
	0x41401861, 0x00000040,
	{0x75, 0xDB, 0xEA, 0x93, 0x09, 0x9E, 0xB5, 0xEC, 0x18, 0xCA, 0x05, 0xF1, 0x79, 0xA4, 0xD8, 0xC9 }
}, // 574

{
	0x828030C3, 0x00000080,
	{0x7C, 0xCC, 0x40, 0x55, 0xE9, 0x61, 0x6F, 0xA9, 0xC9, 0xA0, 0x93, 0x1B, 0x03, 0x0A, 0x23, 0x46 }
}, // 575

{
	0x05006187, 0x00000101,
	{0xB0, 0x15, 0x45, 0x35, 0x8B, 0x86, 0x6B, 0x28, 0xB3, 0x07, 0xC8, 0xD7, 0x3D, 0xC7, 0x08, 0x6D }
}, // 576

{
	0x0A00C30E, 0x00000202,
	{0x8A, 0xC7, 0x00, 0x16, 0x38, 0x65, 0x9C, 0xCB, 0x5B, 0x29, 0xBF, 0x86, 0x27, 0xC0, 0x02, 0xE9 }
}, // 577

{
	0x1401861D, 0x00000404,
	{0xDA, 0xED, 0x39, 0xDF, 0x97, 0xFF, 0xEE, 0xFA, 0x61, 0xF8, 0xC6, 0x54, 0x7B, 0x92, 0x9A, 0xDF }
}, // 578

{
	0x28030C3A, 0x00000808,
	{0x3D, 0x08, 0xE4, 0xF7, 0x8F, 0x65, 0xF1, 0x40, 0xF5, 0x7F, 0x8A, 0x41, 0x5E, 0xC6, 0x12, 0xE8 }
}, // 579

{
	0x50061874, 0x00001010,
	{0xB8, 0x6C, 0x5E, 0xE2, 0xAD, 0x7A, 0x03, 0xDA, 0x64, 0x4B, 0x06, 0xA0, 0xA6, 0xF5, 0xC9, 0xCF }
}, // 580

{
	0xA00C30E8, 0x00002020,
	{0xE6, 0x65, 0x29, 0x32, 0x36, 0xEE, 0x86, 0xA5, 0x08, 0xD4, 0x04, 0xC2, 0xBF, 0x82, 0x53, 0x9A }
}, // 581

{
	0x401861D0, 0x00004041,
	{0x81, 0x11, 0x6B, 0xBD, 0xF3, 0x0D, 0xF9, 0xA4, 0x94, 0xAF, 0x8C, 0x26, 0x89, 0x7F, 0xDB, 0xD4 }
}, // 582

{
	0x8030C3A0, 0x00008082,
	{0x85, 0x66, 0xE4, 0x32, 0x85, 0x45, 0xE4, 0xA4, 0x09, 0x16, 0x07, 0xC0, 0x19, 0x75, 0x3A, 0x94 }
}, // 583

{
	0x00618741, 0x00010105,
	{0xE9, 0x3D, 0x5C, 0x0A, 0x19, 0x25, 0xDD, 0x2B, 0x2E, 0x98, 0xDD, 0xF0, 0xF6, 0x65, 0x8D, 0x42 }
}, // 584

{
	0x00C30E83, 0x0002020A,
	{0xBA, 0x3E, 0xE9, 0xF9, 0xD0, 0xAD, 0x8D, 0xC5, 0x13, 0xE9, 0x0A, 0x28, 0x38, 0x15, 0xB0, 0xC3 }
}, // 585

{
	0x01861D06, 0x00040414,
	{0xA8, 0x4D, 0x12, 0xF7, 0x18, 0x5C, 0x92, 0xE0, 0x10, 0x99, 0x3C, 0xA7, 0xAA, 0x43, 0xA3, 0xAA }
}, // 586

{
	0x030C3A0D, 0x00080828,
	{0xAD, 0xD0, 0xD6, 0x65, 0xA1, 0xFD, 0x5D, 0xB2, 0xD2, 0xC9, 0x5B, 0xE3, 0xD1, 0x1F, 0x1D, 0xA4 }
}, // 587

{
	0x0618741B, 0x00101050,
	{0x5E, 0xC6, 0x4F, 0x29, 0xD8, 0x9C, 0xE5, 0xD8, 0x5F, 0x69, 0xCC, 0xA4, 0x55, 0x69, 0xDF, 0x80 }
}, // 588

{
	0x0C30E837, 0x002020A0,
	{0x35, 0x8A, 0xB6, 0xA1, 0x8B, 0xD1, 0x56, 0x4F, 0x43, 0x34, 0xBD, 0xE9, 0xE0, 0x0E, 0x04, 0xFE }
}, // 589

{
	0x1861D06E, 0x00404140,
	{0x42, 0xDE, 0xEC, 0xBD, 0x13, 0xE5, 0xE2, 0xB8, 0x12, 0xA3, 0xCF, 0x55, 0xDC, 0x6D, 0xF5, 0x7E }
}, // 590

{
	0x30C3A0DC, 0x00808280,
	{0xE7, 0x0C, 0x76, 0xDD, 0x35, 0x90, 0x2C, 0xF5, 0xF0, 0x24, 0x0C, 0x98, 0x89, 0x92, 0x79, 0x0F }
}, // 591

{
	0x618741B8, 0x01010500,
	{0x78, 0x9C, 0x1C, 0xB4, 0x81, 0x0D, 0xD9, 0x11, 0x5D, 0xFE, 0xBA, 0xB2, 0x97, 0x67, 0x9F, 0xD2 }
}, // 592

{
	0xC30E8371, 0x02020A00,
	{0xAB, 0xB1, 0x8C, 0xD2, 0x39, 0x5C, 0x84, 0x64, 0xD7, 0xB4, 0x5C, 0x2E, 0xA3, 0x19, 0x82, 0xD9 }
}, // 593

{
	0x861D06E2, 0x04041401,
	{0xAE, 0xF3, 0x74, 0x13, 0x75, 0x0F, 0xC6, 0x82, 0x5C, 0x5B, 0xE5, 0x90, 0xD4, 0xB6, 0xB6, 0xBF }
}, // 594

{
	0x0C3A0DC4, 0x08082803,
	{0x91, 0xFD, 0x62, 0xC5, 0x04, 0xD8, 0x9B, 0x4B, 0x9C, 0x1D, 0xE4, 0x0A, 0x32, 0xE6, 0x4C, 0x40 }
}, // 595

{
	0x08081802, 0x00000000,
	{0x2C, 0x7E, 0x0E, 0x11, 0x4D, 0xAF, 0x3F, 0x2B, 0x2E, 0x4D, 0x02, 0x8D, 0x3A, 0x56, 0x75, 0xFB }
}, // 596

{
	0x10103004, 0x00000000,
	{0x08, 0xEA, 0x03, 0x83, 0xFE, 0xA9, 0xD6, 0x23, 0x09, 0x78, 0x3F, 0xC3, 0x03, 0x4C, 0xBD, 0x8C }
}, // 597

{
	0x20206008, 0x00000000,
	{0x86, 0x81, 0x8A, 0xE3, 0x57, 0xEC, 0x0D, 0x1F, 0xD3, 0x6A, 0x29, 0x07, 0x62, 0xA5, 0x6A, 0x7B }
}, // 598

{
	0x4040C010, 0x00000000,
	{0xD0, 0xCB, 0x40, 0xB2, 0x7D, 0xA4, 0xA7, 0x54, 0x7A, 0x44, 0xD1, 0xC0, 0x9B, 0xE9, 0x0D, 0x34 }
}, // 599

{
	0x80818020, 0x00000000,
	{0x16, 0x18, 0x85, 0x7F, 0x72, 0xE5, 0x60, 0xBF, 0x50, 0x8F, 0xC2, 0x93, 0xC9, 0x35, 0x69, 0xEB }
}, // 600

{
	0x01030041, 0x00000001,
	{0xA5, 0x11, 0xB2, 0x31, 0xB7, 0x33, 0x71, 0x61, 0x89, 0x1E, 0xE3, 0x50, 0xE7, 0x5E, 0x0F, 0x0C }
}, // 601

{
	0x02060082, 0x00000002,
	{0xBD, 0xF7, 0x17, 0x24, 0x1A, 0xD0, 0x3D, 0x58, 0x82, 0xB1, 0xD2, 0x61, 0xEF, 0xF6, 0xBE, 0xCA }
}, // 602

{
	0x040C0104, 0x00000004,
	{0x9E, 0xB8, 0x23, 0x62, 0x3A, 0x2D, 0x9D, 0xA4, 0x60, 0xB0, 0x66, 0xF3, 0x6F, 0xAB, 0xFF, 0xCB }
}, // 603

{
	0x08180208, 0x00000008,
	{0x23, 0x5B, 0xE1, 0xB9, 0xF5, 0xD0, 0x96, 0x9F, 0xE4, 0xB7, 0x04, 0x82, 0xA2, 0x92, 0x54, 0x40 }
}, // 604

{
	0x10300410, 0x00000010,
	{0x0A, 0x66, 0x70, 0xD4, 0x7C, 0x30, 0x4B, 0xB9, 0x6B, 0x46, 0xE8, 0xAC, 0x85, 0xBD, 0x70, 0x18 }
}, // 605

{
	0x20600820, 0x00000020,
	{0xC5, 0x86, 0x6E, 0x04, 0x59, 0x97, 0x5A, 0xC3, 0x5D, 0x61, 0x91, 0x73, 0x7B, 0xD6, 0x1F, 0x0F }
}, // 606

{
	0x40C01041, 0x00000040,
	{0x1C, 0x3F, 0xDF, 0xA2, 0x35, 0x44, 0x24, 0xA0, 0x97, 0x75, 0x45, 0x78, 0x63, 0xE7, 0xA5, 0x8F }
}, // 607

{
	0x81802082, 0x00000080,
	{0xAB, 0x8F, 0x63, 0x98, 0x65, 0xFB, 0x06, 0x6C, 0x24, 0x58, 0x16, 0xFA, 0x9C, 0xA6, 0x2C, 0xA9 }
}, // 608

{
	0x03004105, 0x00000101,
	{0x57, 0x28, 0x0E, 0x87, 0x7F, 0xE5, 0xC7, 0xBB, 0x9B, 0x94, 0x7F, 0x04, 0x36, 0x7A, 0xB9, 0x8C }
}, // 609

{
	0x0600820A, 0x00000202,
	{0xDD, 0x32, 0x9D, 0x58, 0xCB, 0x18, 0x8E, 0xD5, 0x9A, 0x0D, 0x79, 0x53, 0xA4, 0x8F, 0xD5, 0x64 }
}, // 610

{
	0x0C010414, 0x00000404,
	{0x8D, 0x4E, 0x94, 0x98, 0xBB, 0x08, 0x49, 0x84, 0x19, 0x74, 0x26, 0x0A, 0xCE, 0x1B, 0xAC, 0xF4 }
}, // 611

{
	0x18020829, 0x00000808,
	{0x23, 0xB9, 0x02, 0xFF, 0xAE, 0x14, 0xC1, 0xCC, 0xE3, 0x41, 0x94, 0xDA, 0x42, 0x03, 0x4E, 0xD3 }
}, // 612

{
	0x30041052, 0x00001010,
	{0x61, 0xA5, 0x08, 0xD6, 0x4C, 0xD5, 0x61, 0x56, 0x1F, 0x63, 0xD6, 0xA1, 0x1E, 0x81, 0x34, 0x53 }
}, // 613

{
	0x600820A4, 0x00002020,
	{0x9B, 0xB2, 0x73, 0xEF, 0x30, 0x00, 0xA0, 0x48, 0xD5, 0x4F, 0x1E, 0x9C, 0x02, 0xC2, 0x79, 0xF6 }
}, // 614

{
	0xC0104148, 0x00004040,
	{0x3B, 0xA2, 0xFC, 0x20, 0x4F, 0xE4, 0x08, 0x59, 0x4E, 0x8C, 0xF3, 0xE4, 0xDE, 0x11, 0xE7, 0xF4 }
}, // 615

{
	0x80208290, 0x00008081,
	{0xC5, 0xCD, 0xC5, 0xAC, 0xB0, 0x8D, 0x79, 0xAD, 0x5C, 0x3E, 0x3D, 0x0B, 0x96, 0x15, 0xC4, 0xCD }
}, // 616

{
	0x00410520, 0x00010103,
	{0x21, 0x52, 0x69, 0x4F, 0x1C, 0x81, 0x6C, 0xFD, 0x80, 0xF7, 0x6D, 0xED, 0x3E, 0x79, 0xA8, 0x3C }
}, // 617

{
	0x00820A41, 0x00020206,
	{0x13, 0xA5, 0xAE, 0x35, 0xAE, 0xD4, 0xAF, 0x35, 0x48, 0xF1, 0xF3, 0xAD, 0x91, 0xA7, 0xF6, 0x95 }
}, // 618

{
	0x01041482, 0x0004040C,
	{0x05, 0xF7, 0x59, 0x73, 0xF8, 0xB9, 0x23, 0xC4, 0x72, 0x05, 0xA3, 0xCD, 0x97, 0xED, 0xED, 0xB3 }
}, // 619

{
	0x02082905, 0x00080818,
	{0x6A, 0xFE, 0xE9, 0x69, 0xB1, 0x85, 0x06, 0x98, 0x10, 0x24, 0x4F, 0xAE, 0x7C, 0xB3, 0xA2, 0xF4 }
}, // 620

{
	0x0410520B, 0x00101030,
	{0xDF, 0x26, 0x36, 0x12, 0x9A, 0x04, 0x09, 0xD3, 0x7B, 0x82, 0x90, 0x62, 0xF7, 0xD6, 0x19, 0x3F }
}, // 621

{
	0x0820A417, 0x00202060,
	{0x8B, 0x73, 0x9C, 0x9A, 0xD5, 0x18, 0x1C, 0xB7, 0x4D, 0x77, 0xE5, 0x86, 0xE2, 0x1F, 0x94, 0xDA }
}, // 622

{
	0x1041482F, 0x004040C0,
	{0xE8, 0x86, 0x8B, 0x3C, 0x8E, 0x3F, 0x30, 0x6B, 0x17, 0xC1, 0xAC, 0xFF, 0xA2, 0x6F, 0x3B, 0x0D }
}, // 623

{
	0x2082905F, 0x00808180,
	{0x14, 0xFD, 0xE9, 0xFA, 0x28, 0xD5, 0x4A, 0x32, 0x3A, 0xEC, 0xF1, 0x1F, 0x21, 0x89, 0x0A, 0x6E }
}, // 624

{
	0x410520BF, 0x01010300,
	{0x92, 0x87, 0x78, 0x8B, 0x95, 0xB7, 0x3A, 0xF9, 0x9F, 0x56, 0x38, 0xDB, 0x2F, 0x0B, 0xF0, 0xD1 }
}, // 625

{
	0x820A417E, 0x02020600,
	{0x7D, 0x39, 0xF6, 0xB2, 0x8C, 0x85, 0x60, 0xF4, 0x0C, 0x4F, 0x23, 0x03, 0xA7, 0xA9, 0x69, 0x6C }
}, // 626

{
	0x041482FC, 0x04040C01,
	{0xF3, 0xF8, 0x6D, 0xE0, 0x62, 0x36, 0x13, 0x1A, 0x9B, 0xED, 0x67, 0xC1, 0x0B, 0x99, 0xB0, 0x46 }
}, // 627

{
	0x082905F8, 0x08081802,
	{0x74, 0x97, 0x9B, 0x08, 0xF5, 0x0C, 0x78, 0xF1, 0x1B, 0xCF, 0x01, 0x5D, 0x97, 0x7D, 0x10, 0xB2 }
}, // 628

{
	0x818030C2, 0x00000080,
	{0x22, 0x01, 0x87, 0x4F, 0xD4, 0x6F, 0x46, 0xCD, 0x9D, 0x0D, 0x16, 0x5C, 0xD3, 0x3E, 0xFB, 0x87 }
}, // 629

{
	0x03006184, 0x00000101,
	{0xC3, 0x07, 0x5F, 0xCB, 0x31, 0x42, 0x63, 0x19, 0x2C, 0x66, 0x77, 0xA9, 0x9D, 0xF9, 0xF6, 0x51 }
}, // 630

{
	0x0600C309, 0x00000202,
	{0x37, 0x07, 0x4E, 0x4D, 0x85, 0xBC, 0x18, 0xF6, 0x45, 0x19, 0x5F, 0x10, 0x0F, 0x6C, 0x7D, 0xBF }
}, // 631

{
	0x0C018613, 0x00000404,
	{0x32, 0xD9, 0x6F, 0xCD, 0x6D, 0x78, 0x2D, 0x68, 0x5D, 0x7F, 0xC3, 0xB7, 0x35, 0xD3, 0xAB, 0x51 }
}, // 632

{
	0x18030C26, 0x00000808,
	{0x87, 0x99, 0x20, 0x1D, 0xF6, 0x91, 0x01, 0xC5, 0x7F, 0x62, 0xC7, 0x2C, 0x53, 0x5E, 0xD4, 0xFF }
}, // 633

{
	0x3006184D, 0x00001010,
	{0xD0, 0x1A, 0xCC, 0xF3, 0x09, 0x01, 0xC1, 0xC2, 0x42, 0x57, 0xF1, 0x4B, 0xBC, 0x76, 0xBB, 0xC8 }
}, // 634

{
	0x600C309A, 0x00002020,
	{0xCD, 0xF5, 0x9E, 0x6F, 0x6C, 0x4A, 0xA6, 0x66, 0xD6, 0xDA, 0xEE, 0xE6, 0xC6, 0xC4, 0x3A, 0x69 }
}, // 635

{
	0xC0186135, 0x00004040,
	{0xBD, 0x2C, 0x80, 0xB7, 0x00, 0x48, 0x85, 0xF1, 0x32, 0xCC, 0xEB, 0x31, 0x1F, 0xDD, 0xCE, 0x61 }
}, // 636

{
	0x8030C26A, 0x00008081,
	{0xB3, 0xDE, 0x21, 0x0E, 0xE9, 0x10, 0x12, 0x06, 0xBE, 0x36, 0x3B, 0x29, 0x01, 0x3E, 0xD7, 0xD4 }
}, // 637

{
	0x006184D4, 0x00010103,
	{0xC6, 0xA9, 0x27, 0xA3, 0x3D, 0x36, 0x61, 0x56, 0xDF, 0x3D, 0x01, 0xB0, 0x19, 0xB9, 0x5C, 0x6D }
}, // 638

{
	0x00C309A8, 0x00020206,
	{0xB4, 0xB7, 0xF2, 0x18, 0xDE, 0xDE, 0x84, 0x7E, 0x28, 0xB6, 0x53, 0x82, 0xCB, 0x4E, 0xDC, 0xC8 }
}, // 639

{
	0x01861351, 0x0004040C,
	{0x56, 0x7E, 0x6E, 0x0B, 0x9A, 0x5C, 0xE7, 0xBB, 0xEE, 0xC7, 0x4D, 0xB2, 0x20, 0x22, 0xCA, 0xFF }
}, // 640

{
	0x030C26A2, 0x00080818,
	{0x82, 0xA4, 0x84, 0x4F, 0x71, 0x6D, 0x6F, 0xAF, 0xD9, 0xA4, 0x34, 0x1F, 0x9E, 0x2C, 0x38, 0xB3 }
}, // 641

{
	0x06184D45, 0x00101030,
	{0xE5, 0x7F, 0x7B, 0x8D, 0x8E, 0xD7, 0x95, 0xF3, 0x31, 0x21, 0x4F, 0x21, 0x4A, 0xF5, 0xBC, 0x87 }
}, // 642

{
	0x0C309A8A, 0x00202060,
	{0xAB, 0x6B, 0x22, 0x4A, 0x5B, 0xE2, 0xAE, 0xCA, 0x3D, 0x62, 0xA7, 0xA7, 0x98, 0xE4, 0x21, 0x35 }
}, // 643

{
	0x18613514, 0x004040C0,
	{0xFA, 0x2C, 0x27, 0xFE, 0x87, 0xAF, 0x45, 0x62, 0x56, 0x9A, 0xCE, 0x88, 0x21, 0x52, 0x9E, 0xC8 }
}, // 644

{
	0x30C26A29, 0x00808180,
	{0x4C, 0x01, 0xA1, 0xF8, 0xBF, 0xEB, 0x70, 0x59, 0xE4, 0xB6, 0x34, 0xC4, 0xB7, 0xE9, 0x8C, 0xCA }
}, // 645

{
	0x6184D453, 0x01010300,
	{0xE8, 0x5D, 0x98, 0x70, 0x27, 0xB1, 0x85, 0x58, 0x6C, 0x0E, 0x47, 0x8B, 0x8E, 0x89, 0x94, 0x55 }
}, // 646

{
	0xC309A8A7, 0x02020600,
	{0xBC, 0xE2, 0x67, 0x28, 0x22, 0x55, 0x11, 0xFF, 0x7B, 0xE8, 0x6E, 0x9B, 0x83, 0x85, 0x2B, 0xFB }
}, // 647

{
	0x8613514F, 0x04040C01,
	{0x2F, 0xB7, 0x7D, 0xF6, 0x21, 0xDB, 0x70, 0x49, 0x7C, 0x40, 0x11, 0x55, 0x85, 0x6E, 0x53, 0xE4 }
}, // 648

{
	0x0C26A29F, 0x08081803,
	{0x4F, 0x5D, 0xA6, 0x29, 0x13, 0xED, 0x41, 0x29, 0x97, 0xC8, 0x2B, 0x2B, 0x16, 0x10, 0x70, 0xA0 }
}, // 649

{
	0x08082002, 0x00000000,
	{0x1D, 0xB3, 0xAF, 0x14, 0x1A, 0x4A, 0x1A, 0xD2, 0xD4, 0xD5, 0x78, 0x9A, 0xB4, 0x39, 0x19, 0xBF }
}, // 650

{
	0x10104004, 0x00000000,
	{0xD2, 0xED, 0x53, 0x6A, 0xDE, 0xA7, 0x06, 0x42, 0x7D, 0x6D, 0x98, 0xC0, 0x70, 0x07, 0x66, 0xD4 }
}, // 651

{
	0x20208008, 0x00000000,
	{0xFB, 0xA3, 0x69, 0x94, 0xCA, 0xD6, 0x14, 0xE5, 0x67, 0x2D, 0xC0, 0xA0, 0xB6, 0x93, 0xB1, 0xCE }
}, // 652

{
	0x40410010, 0x00000000,
	{0x93, 0xF1, 0x65, 0xCD, 0x2A, 0xFC, 0x23, 0x6D, 0x31, 0xF8, 0x6C, 0xD3, 0x5E, 0x76, 0xB7, 0xD1 }
}, // 653

{
	0x80820020, 0x00000000,
	{0x5E, 0x70, 0x4C, 0x33, 0x21, 0xED, 0xD0, 0x5D, 0x1E, 0x27, 0x3D, 0x22, 0x30, 0xC8, 0x02, 0x1D }
}, // 654

{
	0x01040041, 0x00000001,
	{0x6B, 0x5F, 0x65, 0x80, 0x09, 0x11, 0x59, 0x4B, 0x3A, 0xEF, 0xBA, 0xCA, 0x7C, 0x7C, 0xD3, 0x6D }
}, // 655

{
	0x02080083, 0x00000002,
	{0xCD, 0x16, 0x63, 0x84, 0x6C, 0x52, 0x83, 0x49, 0x59, 0x15, 0xB3, 0xD9, 0x1A, 0xBD, 0xD2, 0xB7 }
}, // 656

{
	0x04100106, 0x00000004,
	{0x38, 0x04, 0x86, 0x89, 0xC3, 0x9E, 0x47, 0x96, 0xB2, 0x7C, 0xF0, 0x2F, 0x46, 0xF6, 0x6C, 0x29 }
}, // 657

{
	0x0820020C, 0x00000008,
	{0xC2, 0x6F, 0x59, 0x15, 0x92, 0xF2, 0x18, 0xE6, 0x40, 0xE0, 0xC2, 0xCD, 0x2F, 0xCA, 0xE9, 0x65 }
}, // 658

{
	0x10400418, 0x00000010,
	{0x69, 0x36, 0x24, 0x3F, 0x7A, 0x50, 0x3C, 0xA0, 0x4C, 0x66, 0x86, 0x95, 0xC8, 0xCE, 0xA6, 0xF1 }
}, // 659

{
	0x20800830, 0x00000020,
	{0xC0, 0x8F, 0xA3, 0x62, 0x3D, 0x96, 0x8A, 0xF6, 0x0B, 0x83, 0xC2, 0xC3, 0xB2, 0x82, 0xFF, 0x26 }
}, // 660

{
	0x41001060, 0x00000040,
	{0x0B, 0x30, 0x04, 0xEB, 0x62, 0xD5, 0xA2, 0x82, 0x1C, 0x07, 0xD9, 0xE2, 0x90, 0x4E, 0xE8, 0xEF }
}, // 661

{
	0x820020C0, 0x00000080,
	{0x55, 0x6E, 0xFE, 0xB1, 0x82, 0x02, 0xDC, 0xC8, 0x2C, 0x1E, 0x24, 0xC1, 0x6D, 0xF8, 0x0F, 0x70 }
}, // 662

{
	0x04004180, 0x00000101,
	{0xB2, 0x25, 0x31, 0x94, 0x6B, 0x3C, 0x23, 0x03, 0x9E, 0x0B, 0xDD, 0xB2, 0x52, 0x0A, 0xC3, 0xB1 }
}, // 663

{
	0x08008301, 0x00000202,
	{0x43, 0x83, 0xB6, 0x6E, 0xFC, 0x77, 0x3C, 0x2B, 0xA1, 0xA3, 0x55, 0x51, 0x88, 0x09, 0xF3, 0x2D }
}, // 664

{
	0x10010602, 0x00000404,
	{0xE6, 0x5F, 0x1E, 0x22, 0x87, 0xA4, 0x56, 0x78, 0x2F, 0x74, 0x9C, 0x05, 0x6B, 0x51, 0x5F, 0x05 }
}, // 665

{
	0x20020C04, 0x00000808,
	{0x35, 0xEC, 0x29, 0x2C, 0x4E, 0xA0, 0xB6, 0x55, 0xDC, 0xC8, 0xC9, 0xC8, 0x44, 0xB8, 0x00, 0xD6 }
}, // 666

{
	0x40041809, 0x00001010,
	{0xF8, 0xC2, 0xE7, 0x35, 0x40, 0x07, 0xDA, 0xBA, 0x94, 0xC8, 0x4A, 0x17, 0xDE, 0xC9, 0x25, 0x44 }
}, // 667

{
	0x80083013, 0x00002020,
	{0xE9, 0xE3, 0x1A, 0x03, 0x04, 0x16, 0xE4, 0x96, 0xBB, 0xE2, 0x30, 0x74, 0xD1, 0x28, 0xB9, 0xC1 }
}, // 668

{
	0x00106027, 0x00004041,
	{0x93, 0xBA, 0x39, 0x5C, 0x09, 0x81, 0xA4, 0x44, 0x43, 0xC3, 0x69, 0x9F, 0xC7, 0xF0, 0x66, 0xC5 }
}, // 669

{
	0x0020C04E, 0x00008082,
	{0x7B, 0xB3, 0x86, 0x18, 0x3B, 0x05, 0x94, 0xC6, 0x23, 0x1D, 0x39, 0xD6, 0x2C, 0x1E, 0xA0, 0x03 }
}, // 670

{
	0x0041809C, 0x00010104,
	{0xC1, 0xA3, 0x8A, 0x90, 0xB1, 0x88, 0x6C, 0xF2, 0x9B, 0x68, 0x32, 0xE3, 0x93, 0xF2, 0xF1, 0x19 }
}, // 671

{
	0x00830138, 0x00020208,
	{0xFF, 0x03, 0xB6, 0xD4, 0x05, 0x8A, 0x5F, 0x7C, 0x43, 0x03, 0x0D, 0x2E, 0xBC, 0x73, 0xB1, 0x4F }
}, // 672

{
	0x01060270, 0x00040410,
	{0x31, 0x5D, 0x22, 0xCA, 0x3A, 0xF1, 0xAF, 0xDF, 0x8A, 0xC9, 0x67, 0xED, 0x11, 0x91, 0x21, 0xBA }
}, // 673

{
	0x020C04E1, 0x00080820,
	{0x8D, 0x10, 0x0A, 0x6C, 0xB7, 0xE3, 0xCA, 0x95, 0x76, 0x82, 0x66, 0x6A, 0x43, 0xFA, 0x9E, 0x30 }
}, // 674

{
	0x041809C2, 0x00101040,
	{0x2B, 0x90, 0xA8, 0x2A, 0xDE, 0xB8, 0x5C, 0x43, 0x00, 0xBC, 0xFA, 0xC1, 0x8F, 0x70, 0x0A, 0x7B }
}, // 675

{
	0x08301384, 0x00202080,
	{0x7E, 0xAF, 0x28, 0x79, 0x07, 0xAD, 0xC9, 0xA9, 0xC6, 0xBB, 0x55, 0x87, 0x7B, 0xA4, 0xEC, 0xF9 }
}, // 676

{
	0x10602708, 0x00404100,
	{0x14, 0xDD, 0x8E, 0xAF, 0x71, 0x04, 0x5C, 0xE6, 0x42, 0x2E, 0xD6, 0xAF, 0x8B, 0x60, 0x3F, 0xD1 }
}, // 677

{
	0x20C04E10, 0x00808200,
	{0xAD, 0xFE, 0x0B, 0x6F, 0x58, 0xC3, 0x51, 0x17, 0xB7, 0x70, 0xD2, 0xA2, 0x69, 0xB7, 0x9A, 0xBC }
}, // 678

{
	0x41809C20, 0x01010400,
	{0x94, 0x5D, 0xC1, 0x42, 0x0F, 0xD2, 0x40, 0xAA, 0x6A, 0x2D, 0x0A, 0xCC, 0x74, 0x7C, 0xC7, 0x6E }
}, // 679

{
	0x83013841, 0x02020800,
	{0x4C, 0x07, 0x84, 0x33, 0xA2, 0xA8, 0x08, 0x7C, 0x2F, 0x21, 0xA9, 0x1E, 0xD4, 0xF5, 0x52, 0xE6 }
}, // 680

{
	0x06027083, 0x04041001,
	{0x10, 0xEF, 0x54, 0x77, 0x76, 0x28, 0x5F, 0xD3, 0xB5, 0x94, 0xD2, 0x47, 0x13, 0x8D, 0xBD, 0x0E }
}, // 681

{
	0x0C04E106, 0x08082002,
	{0x1D, 0xDD, 0x1B, 0x45, 0x89, 0x7E, 0x9C, 0x2C, 0xE1, 0x32, 0x58, 0x0E, 0xAB, 0x81, 0x02, 0xC7 }
}, // 682

{
	0x40C01040, 0x00000040,
	{0xD1, 0x45, 0xF0, 0x77, 0xD3, 0x37, 0xFE, 0x72, 0xB6, 0x89, 0xDF, 0xF1, 0xC1, 0xE1, 0x68, 0x9C }
}, // 683

{
	0x81802080, 0x00000080,
	{0x90, 0x60, 0x5D, 0xBD, 0xFF, 0x59, 0x25, 0xDE, 0x0D, 0xB1, 0x61, 0xE5, 0x1C, 0xE4, 0xBB, 0x87 }
}, // 684

{
	0x03004101, 0x00000101,
	{0xA9, 0x3E, 0x49, 0xBD, 0x59, 0x72, 0x33, 0x91, 0x34, 0x98, 0x6D, 0xBB, 0xC8, 0x67, 0x3C, 0x3D }
}, // 685

{
	0x06008203, 0x00000202,
	{0x8C, 0xFD, 0xA3, 0xB7, 0xBE, 0x11, 0x8D, 0xBE, 0x5C, 0x0A, 0xA0, 0x55, 0xBA, 0x26, 0x30, 0x41 }
}, // 686

{
	0x0C010407, 0x00000404,
	{0x33, 0xC4, 0x3C, 0xA5, 0xA7, 0xF3, 0x05, 0x73, 0x72, 0x91, 0x4C, 0xBC, 0x10, 0xC2, 0xA9, 0x02 }
}, // 687

{
	0x1802080E, 0x00000808,
	{0x3F, 0x0C, 0xCB, 0x8A, 0xAA, 0xC6, 0xA4, 0x51, 0x47, 0xEB, 0x65, 0x04, 0x24, 0x9F, 0x62, 0xB4 }
}, // 688

{
	0x3004101C, 0x00001010,
	{0x7F, 0x33, 0xDE, 0x6A, 0x93, 0x99, 0x30, 0xBD, 0xF8, 0xEB, 0xF6, 0x87, 0xEF, 0xC8, 0x1E, 0x61 }
}, // 689

{
	0x60082038, 0x00002020,
	{0xC6, 0x22, 0xC2, 0xF4, 0x27, 0xFC, 0xE3, 0x4C, 0xE1, 0x8B, 0x44, 0x0C, 0x3B, 0x2C, 0x77, 0xE2 }
}, // 690

{
	0xC0104071, 0x00004040,
	{0x6A, 0x7E, 0xF8, 0x1D, 0xA3, 0xEC, 0x5E, 0xC3, 0xE1, 0x5D, 0x21, 0x7F, 0xBF, 0xD5, 0xD5, 0x64 }
}, // 691

{
	0x802080E3, 0x00008081,
	{0xDE, 0x50, 0xAB, 0x5D, 0x91, 0x2C, 0x13, 0xB6, 0x13, 0x0F, 0xBD, 0x98, 0x39, 0xDA, 0xB6, 0x63 }
}, // 692

{
	0x004101C6, 0x00010103,
	{0x01, 0xED, 0xDD, 0x41, 0x2B, 0x70, 0x7E, 0xC2, 0xB8, 0x9D, 0x3B, 0x8C, 0xD9, 0xAB, 0x16, 0xA3 }
}, // 693

{
	0x0082038C, 0x00020206,
	{0x30, 0xD0, 0x9C, 0x32, 0x5E, 0x4E, 0x46, 0x1A, 0xAC, 0x59, 0x2B, 0x29, 0xE8, 0xCA, 0x6B, 0xD6 }
}, // 694

{
	0x01040718, 0x0004040C,
	{0x75, 0xA8, 0x20, 0xB9, 0xAE, 0x33, 0xF0, 0x7F, 0x37, 0x7C, 0x46, 0x17, 0xF6, 0xE0, 0x24, 0x26 }
}, // 695

{
	0x02080E31, 0x00080818,
	{0xC8, 0x22, 0x6C, 0x32, 0x1F, 0x4A, 0xF7, 0xF8, 0x6A, 0xFA, 0x90, 0x25, 0x9B, 0x42, 0x91, 0xB9 }
}, // 696

{
	0x04101C63, 0x00101030,
	{0x52, 0x2D, 0xF8, 0xDA, 0xBF, 0xF5, 0xC8, 0x01, 0x41, 0xB2, 0x6B, 0x5D, 0x0E, 0x98, 0x6C, 0xCE }
}, // 697

{
	0x082038C6, 0x00202060,
	{0x6F, 0x34, 0xB0, 0x97, 0xBE, 0x04, 0x45, 0x81, 0x0E, 0x02, 0x78, 0x51, 0x4B, 0xAB, 0x7C, 0xB9 }
}, // 698

{
	0x1040718C, 0x004040C0,
	{0x6A, 0x65, 0x36, 0x9F, 0x94, 0xE6, 0x68, 0x6E, 0xAA, 0x53, 0xAC, 0x55, 0x0C, 0x6A, 0x09, 0xA2 }
}, // 699

{
	0x2080E318, 0x00808180,
	{0xB4, 0x5E, 0xFD, 0xAD, 0x25, 0x34, 0x41, 0x61, 0xCD, 0xF9, 0x33, 0x4C, 0xBC, 0x39, 0x73, 0xBD }
}, // 700

{
	0x4101C631, 0x01010300,
	{0xDD, 0xAE, 0x7E, 0x9A, 0x8E, 0xAB, 0x3F, 0x0A, 0x1F, 0xAB, 0x66, 0xF7, 0x91, 0xF2, 0xE6, 0x05 }
}, // 701

{
	0x82038C63, 0x02020600,
	{0x4C, 0x9F, 0x9A, 0x7A, 0x0D, 0x14, 0xFB, 0x3B, 0x6C, 0xD0, 0xB6, 0xD5, 0xCA, 0x54, 0x72, 0x99 }
}, // 702

{
	0x040718C7, 0x04040C01,
	{0x8D, 0x4B, 0xCF, 0xF7, 0xC3, 0x6A, 0x30, 0xD9, 0x5D, 0x77, 0x53, 0x86, 0x88, 0x5A, 0x4A, 0xBE }
}, // 703

{
	0x080E318E, 0x08081802,
	{0x10, 0xF9, 0x85, 0x30, 0x6B, 0x95, 0x0D, 0xCF, 0x1D, 0x9B, 0x6D, 0xBF, 0xAA, 0x5F, 0x08, 0xAE }
}, // 704

{
	0x0800C30E, 0x00000202,
	{0x51, 0xBD, 0x9A, 0xA1, 0x74, 0xFC, 0x1F, 0x9B, 0x4B, 0xA4, 0xD4, 0x84, 0x6F, 0x72, 0x06, 0x77 }
}, // 705

{
	0x1001861C, 0x00000404,
	{0x0B, 0x24, 0x45, 0xAD, 0x2B, 0xC2, 0x5C, 0x43, 0xA1, 0xD9, 0xE9, 0xB6, 0x96, 0x95, 0x3D, 0x13 }
}, // 706

{
	0x20030C38, 0x00000808,
	{0x33, 0x03, 0xBF, 0x8B, 0x5A, 0xAA, 0xA2, 0xBD, 0x94, 0xE2, 0xAE, 0xA7, 0x44, 0x2F, 0x5F, 0xD9 }
}, // 707

{
	0x40061871, 0x00001010,
	{0x74, 0x10, 0xC2, 0x39, 0x7A, 0x82, 0x3D, 0x98, 0x4A, 0x39, 0x1A, 0x9A, 0xBB, 0xCF, 0xA1, 0xE7 }
}, // 708

{
	0x800C30E2, 0x00002020,
	{0x37, 0x00, 0xB9, 0xB7, 0x0D, 0xF0, 0xD7, 0xEC, 0x80, 0xE7, 0xD3, 0x15, 0x17, 0xBB, 0x09, 0xFC }
}, // 709

{
	0x001861C5, 0x00004041,
	{0xF0, 0x9F, 0xB7, 0x58, 0xE6, 0x9C, 0x2F, 0xDB, 0xF1, 0x48, 0x1F, 0x6E, 0x5C, 0xD7, 0x1B, 0x5E }
}, // 710

{
	0x0030C38A, 0x00008082,
	{0xF7, 0x2B, 0xBE, 0x20, 0x85, 0xDC, 0x4D, 0x56, 0xEE, 0x6D, 0xC8, 0x62, 0xDA, 0x91, 0xD1, 0xDB }
}, // 711

{
	0x00618715, 0x00010104,
	{0x76, 0x53, 0xDE, 0x20, 0xA4, 0xB7, 0x7E, 0xC1, 0xBE, 0xC5, 0x93, 0xAB, 0x50, 0x34, 0xA6, 0x98 }
}, // 712

{
	0x00C30E2B, 0x00020208,
	{0x6B, 0x5B, 0xBA, 0x88, 0xA5, 0xF9, 0x4E, 0xFA, 0xC4, 0x5E, 0x4A, 0xB1, 0x6B, 0xA5, 0x16, 0x8E }
}, // 713

{
	0x01861C57, 0x00040410,
	{0x40, 0x7D, 0x5A, 0xC9, 0x58, 0xC7, 0x25, 0xC3, 0x65, 0xFF, 0xBE, 0x0A, 0x49, 0x6D, 0xA6, 0x7F }
}, // 714

{
	0x030C38AE, 0x00080820,
	{0x6E, 0xB2, 0x94, 0xC3, 0xFA, 0xC3, 0xE5, 0xAE, 0x74, 0xB4, 0x03, 0x1B, 0x8B, 0xB4, 0x6F, 0xF7 }
}, // 715

{
	0x0618715D, 0x00101040,
	{0x76, 0x28, 0xD3, 0xE5, 0xAE, 0xC9, 0x8B, 0xCC, 0x37, 0xAF, 0xA4, 0x38, 0x9B, 0x03, 0xFE, 0xE2 }
}, // 716

{
	0x0C30E2BA, 0x00202080,
	{0xDC, 0xC3, 0x0E, 0x06, 0xF3, 0xAA, 0x2C, 0x1E, 0x70, 0xF7, 0xCF, 0x3C, 0x15, 0xA9, 0xB6, 0xA3 }
}, // 717

{
	0x1861C574, 0x00404100,
	{0x88, 0x27, 0x08, 0xC7, 0x8B, 0x2A, 0x7D, 0xF2, 0x93, 0x2C, 0xF2, 0x2A, 0x28, 0x22, 0x78, 0x01 }
}, // 718

{
	0x30C38AE8, 0x00808200,
	{0x79, 0xB4, 0xE1, 0xEB, 0x1A, 0xFE, 0xAA, 0xB3, 0x51, 0x48, 0xD5, 0x02, 0xB4, 0x8D, 0xF6, 0x7D }
}, // 719

{
	0x618715D0, 0x01010400,
	{0x2C, 0x62, 0x37, 0x11, 0x31, 0xA6, 0xB9, 0xFE, 0xBA, 0xB0, 0x84, 0x3D, 0x4D, 0xAF, 0x34, 0x3E }
}, // 720

{
	0xC30E2BA1, 0x02020800,
	{0xD9, 0x44, 0x2A, 0x89, 0xC4, 0x71, 0x92, 0xA8, 0x98, 0x0C, 0xB5, 0xEE, 0xF9, 0xAA, 0xB1, 0xCD }
}, // 721

{
	0x861C5742, 0x04041001,
	{0x64, 0x61, 0xD3, 0x5A, 0x05, 0x26, 0xE9, 0x6B, 0x1D, 0xA4, 0x6A, 0x0C, 0xD3, 0x6C, 0x4B, 0x53 }
}, // 722

{
	0x0C38AE85, 0x08082003,
	{0x08, 0x9A, 0x40, 0xC1, 0xCA, 0x30, 0xDD, 0x93, 0xAB, 0x21, 0x08, 0x36, 0x98, 0xE2, 0xA6, 0x09 }
}, // 723

{
	0x02080082, 0x00000002,
	{0x2A, 0x07, 0x3E, 0x6C, 0xBA, 0x7D, 0x98, 0x5F, 0x60, 0x86, 0xF5, 0xB4, 0x2D, 0x05, 0x61, 0xFD }
}, // 724

{
	0x04100104, 0x00000004,
	{0xED, 0xF7, 0x60, 0xD4, 0xEE, 0x17, 0xC8, 0x99, 0xBB, 0x2E, 0xC8, 0x83, 0xE4, 0x76, 0x5D, 0x6F }
}, // 725

{
	0x08200208, 0x00000008,
	{0x6F, 0x5F, 0x17, 0x81, 0x0D, 0x79, 0x56, 0xA4, 0x34, 0xEE, 0x0E, 0xED, 0x23, 0xB1, 0xD1, 0xBF }
}, // 726

{
	0x10400410, 0x00000010,
	{0xE9, 0xF8, 0xEC, 0xF0, 0x78, 0xDB, 0x18, 0x2B, 0x98, 0xCA, 0xDB, 0xF5, 0x4F, 0xEF, 0xE2, 0x84 }
}, // 727

{
	0x20800820, 0x00000020,
	{0x5B, 0xF9, 0xFD, 0x46, 0x62, 0x35, 0x39, 0xB7, 0xD1, 0x18, 0xB5, 0xF2, 0x3D, 0xA8, 0x23, 0x82 }
}, // 728

{
	0x41001040, 0x00000040,
	{0x27, 0x85, 0x0D, 0x8A, 0x81, 0xE7, 0xB4, 0xBB, 0xC9, 0xD4, 0x88, 0x7A, 0x2A, 0x29, 0x3E, 0xAB }
}, // 729

{
	0x82002080, 0x00000080,
	{0xAC, 0xAE, 0x58, 0x80, 0xA5, 0xC3, 0xC1, 0xDD, 0x0E, 0xAF, 0x29, 0x41, 0x7C, 0xC7, 0x01, 0xCB }
}, // 730

{
	0x04004100, 0x00000101,
	{0x80, 0x0B, 0xFE, 0xFE, 0x5F, 0x36, 0xAE, 0x79, 0x5A, 0xF3, 0x50, 0x8C, 0xCC, 0x2C, 0xB8, 0xE3 }
}, // 731

{
	0x08008201, 0x00000202,
	{0xC5, 0xBF, 0x03, 0x8C, 0xBF, 0x03, 0x3D, 0x8F, 0xE8, 0x75, 0x88, 0x28, 0x76, 0x15, 0x80, 0x22 }
}, // 732

{
	0x10010403, 0x00000404,
	{0x45, 0xB5, 0xAD, 0x06, 0x9A, 0x97, 0x20, 0x8B, 0xA4, 0x98, 0x67, 0x9B, 0x2A, 0x5D, 0xDA, 0xEA }
}, // 733

{
	0x20020807, 0x00000808,
	{0x00, 0xD2, 0x82, 0xBF, 0x6A, 0x5E, 0x59, 0x65, 0x73, 0x64, 0xC5, 0x2E, 0x62, 0xEF, 0xA5, 0x16 }
}, // 734

{
	0x4004100E, 0x00001010,
	{0xF1, 0xB4, 0x8D, 0xB8, 0x5A, 0x6C, 0xAA, 0x4B, 0xE4, 0x89, 0x8C, 0xA9, 0x2D, 0x61, 0xD5, 0xA0 }
}, // 735

{
	0x8008201D, 0x00002020,
	{0x65, 0xA7, 0x21, 0xF3, 0xF5, 0x7E, 0x45, 0x23, 0x65, 0x58, 0xB0, 0x3A, 0xEF, 0x5E, 0xC2, 0x44 }
}, // 736

{
	0x0010403A, 0x00004041,
	{0xC7, 0xEE, 0x32, 0xB8, 0xB2, 0x45, 0x1D, 0x2A, 0xDE, 0x6C, 0xD8, 0x55, 0xE0, 0x6B, 0xE3, 0x6F }
}, // 737

{
	0x00208075, 0x00008082,
	{0x1B, 0xC7, 0x7B, 0x50, 0xAB, 0xA2, 0xB4, 0x91, 0x03, 0xBB, 0x50, 0xEE, 0x98, 0x8A, 0xE2, 0xE4 }
}, // 738

{
	0x004100EA, 0x00010104,
	{0xE4, 0x36, 0xBF, 0xAA, 0x57, 0xFB, 0x98, 0xCE, 0xD1, 0xF1, 0xDC, 0xE5, 0xC0, 0x64, 0x85, 0x97 }
}, // 739

{
	0x008201D4, 0x00020208,
	{0x6D, 0xB0, 0x86, 0x80, 0xEA, 0x00, 0x58, 0x43, 0x8C, 0x88, 0x82, 0x89, 0x34, 0xE4, 0x94, 0xCE }
}, // 740

{
	0x010403A9, 0x00040410,
	{0xC0, 0x71, 0xCD, 0x84, 0xDD, 0x86, 0xC9, 0x15, 0x20, 0x30, 0x37, 0xC4, 0x16, 0xB4, 0xBF, 0xF8 }
}, // 741

{
	0x02080752, 0x00080820,
	{0xFB, 0xDE, 0xFE, 0xBD, 0x54, 0x98, 0xD5, 0x8A, 0xC6, 0x7E, 0xDF, 0x37, 0x9D, 0x5E, 0xFC, 0x55 }
}, // 742

{
	0x04100EA4, 0x00101040,
	{0xE1, 0x9C, 0x6C, 0x6C, 0xB3, 0x15, 0xDB, 0x36, 0x37, 0x10, 0x78, 0x24, 0x22, 0x8A, 0xD3, 0x81 }
}, // 743

{
	0x08201D48, 0x00202080,
	{0x5F, 0x9C, 0xE3, 0x90, 0x68, 0xF1, 0x74, 0xC7, 0xA8, 0x9A, 0xAF, 0xB7, 0xA0, 0x5B, 0x5F, 0x42 }
}, // 744

{
	0x10403A90, 0x00404100,
	{0xD6, 0x8F, 0x8E, 0xA2, 0xDD, 0x19, 0x7C, 0x64, 0xB6, 0xF5, 0x62, 0x95, 0xE0, 0xD7, 0x92, 0xA5 }
}, // 745

{
	0x20807521, 0x00808200,
	{0x37, 0xEE, 0x2F, 0xF3, 0x5A, 0x33, 0x61, 0xD0, 0x5B, 0x59, 0x98, 0x2B, 0xFE, 0xC4, 0x2B, 0x05 }
}, // 746

{
	0x4100EA42, 0x01010400,
	{0x17, 0xFD, 0x19, 0xB7, 0x85, 0xB4, 0x75, 0x1B, 0x80, 0xD6, 0x85, 0x5D, 0x6C, 0x7E, 0x39, 0x31 }
}, // 747

{
	0x8201D485, 0x02020800,
	{0xF0, 0xF9, 0xDD, 0xF8, 0xB4, 0x9B, 0x51, 0xFB, 0xF3, 0x9E, 0x6B, 0x0E, 0x0C, 0x8C, 0x7C, 0x70 }
}, // 748

{
	0x0403A90A, 0x04041001,
	{0xC7, 0x9D, 0xEF, 0xAE, 0x61, 0xBE, 0xFB, 0x6A, 0xFE, 0xA1, 0x39, 0xB4, 0x4F, 0x33, 0x93, 0xC8 }
}, // 749

{
	0x08075215, 0x08082002,
	{0x7A, 0x59, 0x13, 0x4F, 0x10, 0xD7, 0x33, 0x62, 0xA1, 0x25, 0x65, 0xF6, 0x9C, 0xBB, 0x71, 0xDE }
}, // 750

{
	0x828030C0, 0x00000080,
	{0xB4, 0x85, 0x1F, 0xBD, 0xB1, 0x84, 0x80, 0xD8, 0xE9, 0x53, 0x76, 0xA2, 0xCB, 0x5A, 0xF8, 0x79 }
}, // 751

{
	0x05006180, 0x00000101,
	{0xD3, 0xFC, 0x3F, 0xA6, 0x27, 0x2F, 0x8E, 0x22, 0x3C, 0x6E, 0x2F, 0x31, 0xF9, 0xB2, 0x6A, 0x42 }
}, // 752

{
	0x0A00C300, 0x00000202,
	{0xD2, 0xE2, 0x02, 0x2D, 0x3F, 0x5F, 0x42, 0xF1, 0xC1, 0x92, 0xA6, 0x01, 0x62, 0xEC, 0x31, 0xA3 }
}, // 753

{
	0x14018601, 0x00000404,
	{0xB3, 0xF3, 0xA8, 0x44, 0x1C, 0x0E, 0xFF, 0x59, 0x71, 0x62, 0x15, 0xC8, 0x68, 0xF1, 0xDB, 0xF8 }
}, // 754

{
	0x28030C02, 0x00000808,
	{0xC4, 0x9A, 0x9C, 0x23, 0x1E, 0x99, 0x82, 0xC8, 0xA6, 0x0D, 0x84, 0xE9, 0x86, 0x0D, 0x39, 0x27 }
}, // 755

{
	0x50061805, 0x00001010,
	{0x1A, 0xBB, 0x34, 0x77, 0x87, 0x17, 0x39, 0xC6, 0x27, 0x8F, 0xB2, 0xA1, 0x3C, 0xAB, 0xF5, 0xE1 }
}, // 756

{
	0xA00C300A, 0x00002020,
	{0xB6, 0xF2, 0xC5, 0x80, 0x3C, 0xD0, 0xAF, 0x7A, 0x91, 0x42, 0x8C, 0x92, 0x84, 0xA9, 0xCF, 0x44 }
}, // 757

{
	0x40186015, 0x00004041,
	{0x70, 0x56, 0x10, 0x17, 0x54, 0x3F, 0xC6, 0x37, 0xE9, 0x6E, 0x68, 0xAF, 0x64, 0x22, 0xE5, 0x48 }
}, // 758

{
	0x8030C02A, 0x00008082,
	{0x2B, 0xFE, 0xC5, 0x44, 0x90, 0x57, 0x42, 0x43, 0xD7, 0xBB, 0xF3, 0x0D, 0x00, 0x1F, 0x33, 0x99 }
}, // 759

{
	0x00618055, 0x00010105,
	{0x74, 0xAF, 0xB7, 0x99, 0x0F, 0x54, 0x88, 0xCC, 0x7A, 0x8E, 0xF6, 0x85, 0x76, 0x88, 0xB7, 0x96 }
}, // 760

{
	0x00C300AB, 0x0002020A,
	{0xD2, 0xC7, 0x2C, 0xF9, 0x6A, 0x08, 0xCB, 0x38, 0x5C, 0x45, 0xE0, 0xAD, 0xC2, 0x31, 0x7E, 0xF6 }
}, // 761

{
	0x01860157, 0x00040414,
	{0x44, 0xE9, 0x7C, 0x7F, 0xC8, 0xF6, 0xC8, 0xD5, 0x9E, 0x59, 0x6D, 0x99, 0x12, 0x8E, 0x6A, 0x81 }
}, // 762

{
	0x030C02AE, 0x00080828,
	{0x5B, 0xFD, 0x0F, 0x85, 0x9B, 0x02, 0xA7, 0xA0, 0xB1, 0x84, 0x48, 0x39, 0x93, 0x18, 0x8F, 0xFB }
}, // 763

{
	0x0618055C, 0x00101050,
	{0x23, 0xD1, 0x3F, 0x66, 0x87, 0xEE, 0x4C, 0xF2, 0xA4, 0x5D, 0x92, 0x7C, 0x2D, 0x9F, 0xA3, 0xE0 }
}, // 764

{
	0x0C300AB8, 0x002020A0,
	{0x25, 0xC1, 0xFD, 0x13, 0xF9, 0x23, 0xD9, 0x25, 0x01, 0x45, 0xDC, 0xD4, 0x0D, 0xDE, 0x7F, 0x32 }
}, // 765

{
	0x18601570, 0x00404140,
	{0xA4, 0xC1, 0xAE, 0x97, 0xE9, 0x43, 0x01, 0xE4, 0xAB, 0x36, 0x44, 0x94, 0x19, 0xD8, 0x70, 0x0E }
}, // 766

{
	0x30C02AE1, 0x00808280,
	{0x6F, 0x04, 0x99, 0xD5, 0x40, 0x26, 0x10, 0x56, 0x89, 0x45, 0x97, 0xCB, 0x75, 0x4A, 0x0D, 0xF8 }
}, // 767

{
	0x618055C3, 0x01010500,
	{0x4A, 0x97, 0x92, 0x86, 0x66, 0x92, 0x64, 0xA4, 0xB3, 0x07, 0xD4, 0x49, 0x6F, 0x3A, 0xAC, 0x49 }
}, // 768

{
	0xC300AB86, 0x02020A00,
	{0x40, 0x5C, 0xB6, 0xD5, 0x9E, 0x1E, 0xEF, 0x6E, 0x6D, 0x78, 0x0E, 0x6B, 0xD3, 0xB1, 0x41, 0xAE }
}, // 769

{
	0x8601570D, 0x04041401,
	{0xE3, 0x6E, 0x26, 0xC3, 0xCE, 0xB4, 0xD5, 0x2A, 0x52, 0xB6, 0x84, 0x60, 0x0D, 0xC3, 0x01, 0x24 }
}, // 770

{
	0x0C02AE1A, 0x08082803,
	{0x8A, 0x82, 0xE4, 0x81, 0xDF, 0x8F, 0x70, 0xE5, 0xA6, 0x27, 0x64, 0xBB, 0x37, 0x88, 0xDF, 0x4F }
}, // 771



};

int __sceFactoryGetUMDKey(unsigned int part1, unsigned int part2, u8* out)
{
	u8* key = NULL;
	int ret = 0, x, i;

	u8 key_00001708[16] = 
	{
		0x05, 0x8F, 0xE5, 0x35, 0xA0, 0x4C, 0x89, 0xB1, 0xC0, 0x12, 0xAD, 0xE7, 0xDA, 0x06, 0xD9, 0x22
	};

	u8 key_000016D0[16] = 
	{
		0x80, 0x8C, 0x8D, 0xA8, 0x66, 0x65, 0x5C, 0x11, 0x35, 0x12, 0xA6, 0xBF, 0x96, 0x6D, 0x3C, 0x46
	};

	u8 key_00001698[16] = 
	{
		0x77, 0xA5, 0x97, 0xDE, 0x0A, 0xF3, 0xEF, 0x40, 0x06, 0x96, 0x00, 0x9C, 0x03, 0x3C, 0xE3, 0xAE
	};

	u8 key_00001664[16] = 
	{
		0x9F, 0x26, 0xB3, 0x3D, 0x66, 0xAF, 0xFC, 0xF7, 0xC6, 0xC0, 0x1A, 0x35, 0xAB, 0x32, 0x24, 0x22
	};

	u8 key_00001614[16] = 
	{
		0xAA, 0x62, 0x42, 0x04, 0x9E, 0x40, 0xFB, 0xBF, 0x91, 0x16, 0xC3, 0x23, 0x6F, 0xEA, 0x17, 0xEF
	};

	u8 key_00001598[16] = 
	{
		0x1A, 0x4A, 0xA0, 0x55, 0x30, 0x8D, 0x2B, 0xBC, 0x63, 0x2B, 0x54, 0x68, 0xBB, 0x3A, 0x9D, 0x5B
	};

	u8 key_00001564[16] = 
	{
		0xF2, 0x2C, 0x77, 0xAA, 0x4B, 0x1C, 0x14, 0x01, 0x9D, 0xE5, 0x53, 0x23, 0xE3, 0x41, 0x3D, 0x21
	};

	u8 key_000014D0[16] = 
	{
		0x42, 0x01, 0x82, 0x19, 0x24, 0xFF, 0x0C, 0x34, 0xEB, 0xB4, 0x3F, 0x5D, 0xCA, 0x80, 0x7E, 0x8D
	};

	u8 key_00001498[16] = 
	{
		0x40, 0x78, 0x66, 0xEB, 0x9B, 0xA5, 0xE9, 0x29, 0x0B, 0x2F, 0x8E, 0x4D, 0x30, 0x11, 0x67, 0x54
	};

	u8 key_00001460[16] = 
	{
		0x27, 0xC7, 0x59, 0xF1, 0x48, 0xD8, 0x7E, 0xA8, 0xC0, 0xC2, 0x7F, 0x01, 0x3A, 0xE6, 0xEC, 0xD6
	};

	u8 key_000013F0[16] = 
	{
		0x3A, 0xBA, 0x2E, 0x14, 0xC1, 0x07, 0x37, 0x9F, 0x5F, 0x03, 0x3F, 0x79, 0x4D, 0xE0, 0xCE, 0x7E
	};

	u8 key_000013BC[16] = 
	{
		0x71, 0x16, 0x41, 0x21, 0xC3, 0xE0, 0x6A, 0xB5, 0x76, 0x57, 0x35, 0x90, 0xAB, 0x22, 0x3E, 0xFD
	};

	//printf("part1=0x%08X, part2=0x%08X\n", part1, part2);

	if (sub_00003E38(part1, part2, 0x1000004F, 0x80000000) == 1)
	{
		key = key_00001708;
	} 
	
	else if (sub_00003E38(part1, part2, 0x1000004F, 0x80000000) >= 2)
	{
		x = sub_00003E38(part1, part2, 0x3000001F, 0x80000000);

		if (x == 1)
		{
			key = key_000016D0;
		}

		else if (x >= 2)
		{
			x = sub_00003E38(part1, part2, 0x3000003F, 0x80000000);

			if (x == 1)
			{
				key = key_00001698;
			}
			
			else if (x >= 2)
			{
				x = sub_00003E38(part1, part2, 0x3000004F, 0x80000000);

				if (x == 1)
				{
					key = key_00001664;
				}
				else
				{
					ret = -1;
				}
			}
			else
			{
				if (sub_00003E38(part1, part2, 0x3000002F, 0x80000000) == 1)
				{
					key = key_00001614;
				}
				else
				{
					ret = -1;
				}
			}
		}
		else
		{
			x = sub_00003E38(part1, part2, 0x2000000F, 0x80000000);

			if (x == 1)
			{
				key = key_00001598;
			}

			else if (sub_00003E38(part1, part2, 0x3000000F, 0x80000000) == 1)
			{
				key = key_00001564;
			}

			else
			{
				ret = -1;
			}
		}
	}
	else
	{
		x = sub_00003E38(part1, part2, 0x1000001F, 0x80000000);

		if (x == 1)
		{
			key = key_000014D0;
		}

		else if (x >= 2)
		{
			x = sub_00003E38(part1, part2, 0x1000002F, 0x80000000);

			if (x == 1)
			{
				key = key_00001498;
			}

			else if (sub_00003E38(part1, part2, 0x1000003F, 0x80000000) == 1)
			{
				key = key_00001460;
			}

			else
			{
				ret = -1;
			}
		}
		else
		{
			x = sub_00003E38(part1, part2, 1, 0x70000000);
			
			if (x == 1)
			{
				key = key_000013F0;
			} 

			else if (sub_00003E38(part1, part2, 0x1000000F, 0x80000000) == 1)
			{
				key = key_000013BC;
			} 

			else
			{
				ret = -1;
			}
		}
	}	

	if (key == NULL)
	{
		int i;

		for (i = 0; i < N_KEYS; i++)
		{
			if (umdkeys[i].t == part1 && umdkeys[i].m == part2)
			{
				key = umdkeys[i].key;
				ret = 0;
				break;
			}
		}
	}
	
	if (key == 0) 
	{
		Kprintf("Key not found.\n");
		return ret;
	}

	for (i = 0; i < 0x10; i++)
	{
		out[i] = key[i];
	}

	return ret;
}

int UM2(SomeStructure *ss, u8 *predata, Param3 *out, UmdIndex *m, u32 n)
{
	u8 key[16] = 
	{
		0x9F, 0x46, 0xF9, 0xFC, 0xFA, 0xB2, 0xAD, 0x05, 0x69, 0xF6, 0x88, 0xD8, 0x79, 0x4B, 0x92, 0xBA
	};
	
	int unk[4];            //+0
	u8 tmp[0x40];     //+0x10
	AES_KEY ctx1; //+0x50
	AES_KEY ctx2; //+0x150
	u8 *key_end; //+0x25C
	int i, j;
	u8 *idx_end;
	int idx_cnt;
	int ret;

	unk[0] = 0; 
 
	AES_set_encrypt_key(key, 128, &ctx1);
	AES_set_decrypt_key(key, 128, &ctx2);

	for (i = 0; i < 0xC; i++)
	{
		out->buf1[i] = predata[i];
	}
 
	for (i = 0xC; i < 0x10; i++)
	{
		out->buf1[i] = 0;
	}
 
	for (i = 0; i < 0x10; i++)
	{
		out->buf2[i] = out->buf1[i];
	}
 
	for (i = 0; i < 0xB; i++)
	{
		AES_encrypt(out->buf1, out->buf1, &ctx1);
		AES_decrypt(out->buf2, out->buf2, &ctx2);
	}

	out->outbuf[0].t = -1;
	out->outbuf[0].m = 0x80000001;
	//idx_cnt = ((n+0x14)/20)*20;
	idx_cnt = (((n+20)/20)*4+((n+20)/20))*4;
	idx_end = (u8 *)(&out->outbuf[idx_cnt]); // * 8
	key_end = idx_end+idx_cnt*0x10;
	//out->unk1 = ((n+0x14)/20)*6;
	out->unk1 = (((idx_cnt/5)*2)+(idx_cnt/5))*2;


	// v0 = (n+20)/20
	// (((n+20)/20)*4+((n+20)/20))*4; // 20
	// t1 = s7/5
	// (((s7/5)*2)+(s7/5))*2 // 24 (0x18)

	//printf("unk1 = %d, idx_cnt=%d.\n", out->unk1, idx_cnt);

	for(i = 0; i < 0x10; i++)
	{
		idx_end[i] = 0;
	}

	//populate the index/key tables with raw unencrypted indexes/keys.
 
	for(i = 0; i < n; i++)
	{
		out->outbuf[i+1].t = m[i].t;
		out->outbuf[i+1].m = m[i].m;
		ret = __sceFactoryGetUMDKey(m[i].t, m[i].m, &idx_end[0x10+(0x10*i)]);
		if (ret < 0) return -2;
	}
 
	i++;
 
	for(j = i; j < idx_cnt; j++)
	{
		out->outbuf[j].t = m[0].t;
		out->outbuf[j].m = m[0].m;
		ret = __sceFactoryGetUMDKey(m[0].t, m[0].m, &idx_end[j*0x10]);
		if (ret < 0) return -3;
	}

	//UP TO HERE IS JUST TRANSFORMATION. WE NEED TO REVERSE THE PROCESS FROM THE END UNTIL HERE TO GET SOMETHING GOOD.
 
	//encrypt the umd keys.
	AES_set_encrypt_key(out->buf1, 128, &ctx1);
  
	for(i = 1; i < idx_cnt; i++)
	{
		AES_encrypt(&idx_end[0x10*i], &idx_end[0x10*i], &ctx1);
	}

	//dunno

	EncryptRegion(out->buf2, (u8 *)out->outbuf, idx_cnt*24, key_end);
 
	//get keys for below
	GenerateSigncheck(ss, unk, tmp);  
 
	//wohoo, encryption shit, only for the *header* part, the one before actual indexes/keys.
	//maybe some sort of signature? like certgen for region...
	AES_set_encrypt_key(&tmp[0x20], 128, &ctx1);
	for(i = 0; i < 3; i++)
	{
		AES_encrypt(&tmp[0x10], &tmp[0x10], &ctx1);
	}

	AES_set_encrypt_key(&tmp[0x10], 128, &ctx1);
	AES_encrypt(out->buf1, out->buf1, &ctx1);
	
	for(i = 0; i < 0x10; i++)
	{
		out->buf2[i] = out->buf2[i]^out->buf1[i];
	}

	AES_encrypt(out->buf2, out->buf2, &ctx1); 
 
	//flush keys buffers got from O1
	for(i = 0; i < 0x40; i++)
	{
		tmp[i] = 0;
	}

	memset(&ctx1, 0, sizeof(ctx1)); 
 
	//return total size of the data.
	return idx_cnt*24+0x34;	
}

int GenKeys(SomeStructure *ss, u8 *ids_predata, int size, u8 *out, void *m, int n)
{
	int b;
	u32 c[0x40/4]; // sp+0x210
	//u8  sha[0x14]; // sp+0x250
	AES_KEY ctx,  ctx2; // sp, sp+0x100
	int i, j, k;

	b = 0;

	memcpy(out, ids_predata+0x10, 0x478);
	GenerateSigncheck(ss, &b, (void *)c);

	AES_set_encrypt_key((void *)(c+8), 128, &ctx);
	
	for (i = 0; i < 2; i++)
	{
		AES_encrypt((void *)(c+4), (void *)(c+4), &ctx);
	}

	for (i = 0; i < 3; i++)
	{
		AES_encrypt((u8 *)c, (u8 *)c, &ctx);
	}

	u8 *buf = out;
	
	for (i = 0; i < 6; i++)
	{
		AES_set_decrypt_key((u8 *)"FlashROM PreCalc", 128, &ctx2);
		AES_set_encrypt_key((void *)(c+4), 128, &ctx);
			
		for (j = 0; j < 32; j += 16)
		{
			AES_decrypt(buf+0xC0+j, buf+0xC0+j, &ctx2);

			if (j > 0)
			{
				for (k = 0; k < 16; k++)
				{
					buf[0xC0+j+k] ^= buf[0xB0+j+k];
				}
			}
		
			AES_encrypt(buf+0xC0+j, buf+0xC0+j, &ctx);		
		}

		EncryptRegion((void *)c, buf+0x38, 0xA8, buf+0xE0);	
		buf += 0xB8;
	}

	memset(c, 0, 64);
	memset(&ctx, 0, sizeof(ctx));
	memset(&ctx2, 0, sizeof(ctx2));

	return UM2(ss, ids_predata, (Param3 *)(out+0x048C), (UmdIndex *)m, n);
}

int CreateSS(SomeStructure *ss)
{
	U64 fuse;
	int i;

	memcpy(&fuse, &g_fuseid, 8);
	
	memset(ss->buf1, 0, 8);
	memset(ss->buf2, 0xFF, 8);

	memcpy(ss->buf3, &fuse.high, 4);
	memcpy(ss->buf3+4, &fuse.low, 4);

	for (i = 0; i < 4; i++)
	{
		ss->buf1[3-i] = ss->buf3[i];
		ss->buf1[7-i] = ss->buf3[4+i];
	}

	return 0;
}

int CreateCertificatesAndUmd(void *buf, int size, void *predata, void *m, int n)
{
	SomeStructure ss;

	CreateSS(&ss);
	memset(buf, 0, size);

	return GenKeys(&ss, predata, 0xA0, buf, m, n);
}

SceUID gen_thid = -1;
u8 *g_buf; 
int g_result;

int generate_thread(SceSize args, void *argp)
{
	u8 *predata;
	u8 *m;
	int n;

	int k1 = pspSdkSetK1(0);

	n = 77;

	if (is79)
	{
		if (g_region == JAPAN)
		{
			predata = predata_JAP_1_54398E06;
			m = mi_JAP_1_54398E06;
		}
		else if (g_region == AMERICA)
		{
			predata = predata_AM_1_1135060C;
			m = mi_AM_1_1135060C;
		}
		else
		{
			predata = predata_EU_1_8AC48706;
			m = mi_EU_1_8AC48706;
		}
	}
	else if (g_mb == TA81)
	{
		if (g_region == JAPAN)
		{
			predata = predata_JAP_1_54398E06;
			m = mi_JAP_1_54398E06;
		}
		else if (g_region == AMERICA)
		{
			predata = predata_AM_1_D7C2958D;
			m = mi_AM_1_D7C2958D;
		}
		else
		{
			predata = predata_EU_1_4DA9840C;
			m = mi_EU_1_4DA9840C;
		}
	}
	else if (g_mb == TA82)
	{
		if (g_region == JAPAN)
		{
			predata = predata_JAP_2_90A8888E;
			m = mi_JAP_2_90A8888E;
		}
		else if (g_region == AMERICA)
		{
			predata = predata_AM_2_57B41293;
			m = mi_AM_2_57B41293;
		}
		else
		{
			predata = predata_EU_2_27A91195;
			m = mi_EU_2_27A91195;
		}
	}
	else if (g_mb == TA86)
	{
		if (g_region == JAPAN)
		{
			predata = predata_JAP_2_10A18814;
			m = mi_JAP_2_10A18814;
		}
		else if (g_region == AMERICA)
		{
			predata = predata_AM_2_AEA31495;
			m = mi_AM_2_AEA31495;
		}
		else
		{
			predata = predata_EU_2_PINK;
			m = mi_EU_2_PINK;
		}
	}
	else if (g_mb == TA85)
	{
		if (g_region == JAPAN)
		{
			predata = predata_JAP_3_A1C6818E;
			m = mi_JAP_3_A1C6818E;
		}
		else if (g_region == AMERICA)
		{
			predata = predata_AM_3_4B330609;
			m = mi_AM_3_4B330609;
		}
		else
		{
			predata = predata_EU_3_34D90F8D;
			m = mi_EU_3_34D90F8D;
		}
	}
	else if (g_mb == TA85v2)
	{
		if (g_region == JAPAN)
		{
			predata = predata_JAP_3_BAC69087;
			m = mi_JAP_3_BAC69087;
		}
		else if (g_region == AMERICA)
		{
			predata = predata_AM_3_25AB118E;
			m = mi_AM_3_25AB118E;
		}
		else
		{
			predata = predata_EU_3_PLATA;
			m = mi_EU_3_PLATA;
		}
	}
	else // TA88... let's hope :p
	{
		if (g_region == JAPAN)
		{
			predata = predata_JAP_3_0DA18D0F;
			m = mi_JAP_3_0DA18D0F;
		}
		else if (g_region == AMERICA)
		{
			predata = predata_AM_3_8D561422;
			m = mi_AM_3_8D561422;
		}
		else
		{
			predata = predata_EU_3_1B478E18;
			m = mi_EU_3_1B478E18;
		}
	}
	
	g_result = CreateCertificatesAndUmd(g_buf, 0x4000, predata, m, n);
	
	gen_thid = -1;
	pspSdkSetK1(k1);
	return sceKernelExitDeleteThread(0);
}

int idsRegenerationCreateCertificatesAndUMDKeys(u8 *buf)
{
	int k1 = pspSdkSetK1(0);

	if (gen_thid > 0)
	{
		pspSdkSetK1(k1);
		return -1;
	}

	g_buf = (u8 *)(0x80000000 | (u32)buf);
	gen_thid = sceKernelCreateThread("generate_thread", generate_thread, 0x6C, 0x020000, 0, NULL);

	if (gen_thid < 0)
	{
		pspSdkSetK1(k1);
		return gen_thid;
	}

	int res = sceKernelStartThread(gen_thid, 0, NULL);
	if (res < 0)
	{
		pspSdkSetK1(k1);
		return res;
	}

	sceKernelWaitThreadEnd(gen_thid, NULL);

	sceKernelDcacheWritebackAll();
	sceKernelIcacheClearAll();
		
	pspSdkSetK1(k1);
	return g_result;
}

int module_start(SceSize args, void *argp)
{
	Kprintf("IdStorage Regeneration Module, by Dark_AleX & Adrahil\n");
	
	return 0;
}

